/* Mach64 IO routines
 * $Log: Mach64_io.inc,v $
 * Revision 1.3  1997/11/08 15:12:21  becka
 * Namespace cleanup ... probably broke all and everything ...
 *
 * Revision 1.2  1997/04/23 03:52:43  ezrec
 * Minor: Working on the ATI Mach64...
 *
 * Revision 1.1	1997/03/21 23:25:54  ezrec
 * Major: ATI Mach64 driver. Doesn't work. Tries to detect the Mach64 card.
 *	  PCI only. Have fun!
 *
 * Revision 1.2	1997/01/20 04:16:00  ezrec
 * Minor: Now finds the frame buffer - still needs work... And it does look
 *	 like I'll have to dig up the $50 for the ATI technical info - the
 *	 MMIO registers don't seem to cover text mode (font setting, etc..)
 *	 If at all possible, I'd like to use _only_ the MMIO regs, to stay
 *	 away from the VGA nastyness.....
 *
 * Revision 1.1	1997/01/18 22:12:59	ezrec
 * Major:	Adding support for the ATI MAch64 chipset, currently only
 *		detects the Mach64 (PCI only!)
 *
 */
#ifndef _MACH64_INC_
#define _MACH64_INC_

#include "Mach64_strings.inc"

static int ioBASE[MAX_NR_ATI] = { 0x2ec,0x1cc,0x1c8,0x000};

#define REGr(x) (*(ati->reg_map+x))
#define REGw(x,y) {*(ati->reg_map+x)=y;}

/******** Register Manipulation *************/
void Mach64_save(struct ati_config_s *ati,ati_reg *regs)
{
	/* Gods, I love MMIO.... */
	if (ati==NULL) return;
	WaitIdleEmpty(schedule);
	memcpy(regs,ati->reg_map,sizeof(ati_reg)*NUM_CTL_REGS);
}

void Mach64_restore(struct ati_config_s *ati,ati_reg *regs)
{ 
	if (state==NULL) return;
	WaitIdleEmpty(schedule);
	memcpy(ati->reg_map,regs,sizeof(ati_reg)*NUM_CTL_REGS);
}

/******** Detection and Initializtion *************/

static int Mach64_init(int ggiindex) 
{
	int cnt,i,j,status;
	uint8 bus,devfn;
	uint32 rev;
	unsigned int cardno,start,size,aperture;
	mem_addr mem_phys;
	int reg_offset;
	int mem_size;

	if (!pcibios_present()) {
		printk(KERN_ERR "Mach64_detect: Sorry, this driver only supports PCI\n");
		return -E(CHIPSET,UNKNOWN);
	}

	cnt=ggiindex;
	/* First, figure out if there is a Mach64 on the PCI bus. */
	for (i=0;PciDevice[i].string!=NULL;i++) {
		j=0;
		do {
			status = pcibios_find_device(PCI_ATI_VENDOR_ID,
						     PciDevice[i].id,
						     j, &bus, &devfn);
			j++;
			if (status==PCIBIOS_SUCCESSFUL) 
				cnt--;
		} while (status == PCIBIOS_SUCCESSFUL && cnt >= 0);
	}
	if (status != PCIBIOS_SUCCESSFUL) {
		printk(KERN_ERR "Mach64_detect: Failed to find a Mach64 "
				"device on the PCI bus.\n");
		printk(KERN_ERR "Mach64_detect: (contact "MAINTAINER" with the \n"
		       KERN_ERR "Mach64_detect:	contents of /proc/pci)\n");
		printk(KERN_ERR "Mach64_detect: ");
		printk(pcibios_strerror(status));
		printk("\n");
		return -E(CHIPSET,UNKNOWN);
	}


	/* Get the revision number of the Mach64. */
	status = pcibios_read_config_dword(bus, devfn, 
					   PCI_REVISION_ID, &rev);
	if (status != 0) {
		printk(KERN_ERR "Mach64_detect: Error reading PCI class "
				"and revision register.\n");
		printk(KERN_ERR "Mach64_detect: ");
		printk(pcibios_strerror(status));
		printk("\n");
		return -E(CHIPSET,UNKNOWN);
	}

	rev &= 0x000000FF; /* got the revision code */

	/* do I know how to talk to this revision? */
	if (ati.rev != 1) { 
		printk(KERN_ERR "Mach64_detect: Unknown Mach64 "
				"revision code (0x%X) detected.\n", ati.rev);
		printk(KERN_ERR "Mach64_detect: Please contact "
				MAINTAINER" and send the contents of "
				"`/proc/pci'.\n");
		return -E(CHIPSET,UNKNOWN);
	} else {
		DPRINTK("Mach64_detect: Found a %s, rev %d\n",
			PciDevice[i].string,rev);
	}

	/* OK, found it.  
         * Now, read out the configuration information. 
	 */
	pcibios_read_config_dword(bus, devfn, 
				  PCI_BASE_ADDRESS_0, &mem_phys);
	cardno=mem_phys&0x3;
	mem_phys &= ~3;

	DPRINTK("Mach64_init: Memory base is 0x%.8x\n",ati.mem_phys);

	/* BADCODEBADCODE
	 * From now on, I'll just assume that status always is OK... 
	 */

	/* How many bits are we decoding? */
	pcibios_write_config_dword(bus,devfn,
				   PCI_BASE_ADDRESS_0, 0xffffffff);
	pcibios_read_config_dword(bus,devfn,
				  PCI_BASE_ADDRESS_0, &size);
	pcibios_write_config_dword(bus,devfn,
				   PCI_BASE_ADDRESS_0, mem_phys);
	size=~size+1;
	if (size >= 0x800000) 
		aperture=APERTURE_8M_ENABLE;
	else
		aperture=APERTURE_4M_ENABLE;
	reg_offset=size-0x400;
	mem_size=size;

	DPRINTK("Mach64_init: Detected aperture is %dMb (0x%.8x), type %d\n",size/1024/1024,size,aperture);
	DPRINTK("Mach64_init: Setting register addr at (0x%.8x)\n",ati.mem_phys+ati.reg_offset);

	pcibios_read_config_dword(bus,devfn,
				  PCI_BASE_ADDRESS_1, &size);
	if (! (size & 0xfffffffc)) {
		int iobase;

		printk(KERN_ERR "Mach64_init: Block IO not supported on this card.\n");

		iobase=ioBASE[cardno];

		ioCONFIG_CNTL=iobase+(sioCONFIG_CNTL<<10);
		ioMEM_CNTL=iobase+(sioMEM_CNTL<<10);

		old_CONFIG_CNTL=inw(ioCONFIG_CNTL);
		old_MEM_CNTL=inw(ioMEM_CNTL);

		DPRINTK("Mach64_init: CONFIG_CNTL: %.8x Old Aperature %d, "
			"(0x%.8x)\n", old_CONFIG_CNTL,old_CONFIG_CNTL&0xf,
			(old_CONFIG_CNTL&0x3FF0) << 18);
		DPRINTK("Mach64_init: MEM_CNTL: %.8x\n",old_MEM_CNTL);
		DPRINTK("Mach64_init: CONFIG_CNTL: %.8x\n",
				      ((mem_phys>>18)&0x3FF0) | aperture);
		outw(((ati.mem_phys>>18)&0x3FF0) | aperture,ioCONFIG_CNTL);
		outw(old_MEM_CNTL & ~(MEM_BNDRY | MEM_BNDRY_EN),ioMEM_CNTL);

	} else {
		reg_offset=size & (size & 0x01 ? 0xfffffffc : 0xfffffff0);
		DPRINTK("Mach64_init: Register address now: (0x%.8x)\n",
			mem_phys+reg_offset);
	}

	/*
	 * Remember, we can only vremap() at page boundaries 
	 */
	DPRINTK("Mach64_init: remap(0x%.8x,0x%.8x) (reg offset=0x%.8x)\n",
		mem_phys,size,reg_offset);

/* BUGBUGBUGBUG */
	if (ati.reg_buff==NULL) {
		printk(KERN_ERR "Mach64_init: could not map MMIO space\n");
		goto badend;
	}
	ati.reg_map=ati.reg_buff+ati.reg_offset;

	/* Determine the memroy suze
	 */
	size=REGr(MEM_CNTL)&MEM_SIZE_ALIAS;
	switch (size) {
		case MEM_SIZE_512K: chipinfo.ramsize=512*1024; break;
		case MEM_SIZE_1M:	 chipinfo.ramsize=1*1024*1024; break;
		case MEM_SIZE_2M:	 chipinfo.ramsize=2*1024*1024; break;
		case MEM_SIZE_4M:	 chipinfo.ramsize=4*1024*1024; break;
		case MEM_SIZE_6M:	 chipinfo.ramsize=6*1024*1024; break;
		case MEM_SIZE_8M:	 chipinfo.ramsize=8*1024*1024; break;
		default: chipinfo.ramsize=0;
	}
	if (chipinfo.ramsize==0) {
		printk(KERN_ERR "Mach64_init: Cannot determine ram size, assuming 1M\n");
		chipinfo.ramsize=1*1024*1024;
	} else {
		DPRINTK("Mach64_init: Reported ram size: %dk\n",chipinfo.ramsize);
	}

	DPRINTK("Mach64_init: vremap(0x%.8x,0x%.8x)\n",ati.mem_phys,chipinfo.ramsize);
	ati.frame_buff=vremap(ati.mem_phys,chipinfo.ramsize);
	if (ati.frame_buff==0) {
		printk(KERN_ERR "Mach64_init: Cannot remap physical [0x%.8x-0x%.8x]\n",
					 ati.mem_phys,ati.mem_phys+chipinfo.ramsize);
		vfree(ati.reg_buff);
		goto badend;
	}
	DPRINTK("Mach64_init: frame_buff=0x%.8x\n" KERN_DEBUG "Mach64_init:	 reg_buff=0x%.8x\n",ati.frame_buff,ati.reg_buff);

	Mach64_save(&atiInitial);

	ati[ggiindex].pci_bus = bus;
	ati[ggiindex].devfn = devfn;
	ati[ggiindex].rev = rev;
	ati[ggiindex].chipinfo.name=PciDevice[i].string;
			/* !!! More chipinfo fields here.. */
	ati[ggiindex].
	
	return EOK;

badend:
	if (ioCONFIG_CNTL!=0) {
		outw(ioCONFIG_CNTL,old_CONFIG_CNTL);
		outw(ioMEM_CNTL,old_MEM_CNTL);
	}
	vfree(ati.reg_buff);
	return -E(CHIPSET,UNKNOWN);
}

static void Mach64_release(void)
{
	Mach64_restore(&atiInitial);

	if (ioCONFIG_CNTL!=0) {
		outw(ioCONFIG_CNTL,old_CONFIG_CNTL);
		outw(ioMEM_CNTL,old_MEM_CNTL);
	}
	vfree(ati.frame_buff);
	vfree(ati.reg_buff);
}

void outbDAC(unsigned char value, unsigned char dac_reg)
{
}

unsigned char inbDAC(unsigned char dac_reg)
{
	return 0;
}

void setCLK(int value)
{
}

#endif /* _MACH64_INC_ */
