/* --------------------------------------------------------------------------
 *	Cyrix MediaGX chipset driver
 * --------------------------------------------------------------------------
 *
 *	Copyright (C) 1997-1998		Emmanuel Marty
 *	Copyright (C) 1997-1998		Cyrix Corporation
 *	Copyright (C) 1997-1998		The GGI project
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 * 
 *	You should have received a copy of the GNU General Public License
 *	along with this program; see the file COPYING.  If not, write to
 *	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * --------------------------------------------------------------------------
 *
 *	This code was derived from the following sources of information :
 *
 *	Cyrix MediaGX confidential databooks
 *	Cyrix MediaGX gxvideo.c example sourcecode
 *	Help and code from Sengan Baring-Gould of Cyrix Corporation
 *
 *	Many thanks to the Cyrix Corporation for providing me with
 *	documentation and official support about the GX chipset family.
 *
 * --------------------------------------------------------------------------
 *
 *	$Log: gx.c,v $
 *	Revision 1.8  1998/03/20 00:14:19  emarty
 *	MAJOR: Cx5520 support, GGI_AUTO support, sync problem removed, driver reworked
 *	
 *	Revision 1.7  1998/03/05 00:07:51  emarty
 *	text <-> dpp change
 *	
 *	Revision 1.6  1998/01/29 09:04:20  emarty
 *	added 15/16 bpp modes and acceleration
 *	
 *	Revision 1.5  1997/12/31 16:18:15  emarty
 *	fixed typo. final Dali release
 *	
 *	Revision 1.4  1997/12/31 03:13:42  willie
 *	Fix at line 650 in gx.c
 *	
 *	Revision 1.3  1997/12/25 02:18:10  emarty
 *	small bugfix. Dali release.
 *	
 *	Revision 1.2  1997/12/24 18:20:09  emarty
 *	added textmode emulation
 *	
 *	Revision 1.1  1997/12/20 13:00:56  emarty
 *	major: initial checkin
 *	
 */
 
#include <ggi/maintainers.h>
#define MAINTAINER	Emmanuel_Marty
#define DRIVER_NAME	"Cyrix MediaGX chipset driver"
#define DRIVER_REV	"$Revision: 1.8 $"
#define DEBUG_LEVEL	0

#include <ggi/module.h>
#include <Cyrix/gx.h>
#include <Cyrix/cx5520.h>
#include <IBM/vga_dac.h>
#include <ICS/5342.h>

/*	size of memory used for text16 display	*/

#define	TEXT_MEMORY_SIZE	0xF000

/*	size of memory used for pointer sprite	*/

#define	SPRITE_MEMORY_SIZE	0x1000

/*	PCI device IDs of companion chips	*/

static uint32 companion_id[] = {
	PCI_DEVICE_ID_CYRIX_5510,		/* CX5510 */
	PCI_DEVICE_ID_CYRIX_5520,		/* CX5520 */
	PCI_DEVICE_ID_CYRIX_5530_VIDEO,		/* CX5530 */
	-1
};

/*
**	default mode
*/

struct kgi_mode gx_preset = {
		{ 1,				/* need one frame	*/
		  { 80, 25 }, { 128, 25 },	/* visible/virtual	*/
		  { 0, 0 },			/* size in mm		*/
		  GT_TEXT16,			/* mode			*/
		  { 9, 16 }			/* dpp			*/
		},
 
		{ 640, 640, 680, 768, 784, 800,  0 },	/* horiz.timing	*/
		{ 400, 407, 412, 414, 442, 449,  1 },	/* vert. timing	*/
		{ 1, 1 },			/* no magnification	*/
		0,	/* !!! check */		/* special capabilities	*/
		25056815,			/* dot clock in Hz	*/
		{ 1, 1 },			/* clock rate to program*/
		{ 1, 1 },			/* transfer rate to dac	*/
};

/*
**	structure holding most display controller registers,
**	for fast hardware programming
*/

struct chipsetregs {
	uint32 general_cfg;		/* config registers... */
	uint32 timing_cfg;
	uint32 output_cfg;

	uint32 fb_st_offset;		/* framebuffer start offset */
	uint32 cb_st_offset;		/* compression start offset */
	uint32 cur_st_offset;		/* cursor start offset*/
	uint32 icon_st_offset;		/* icon start offset */
	uint32 vid_st_offset;		/* video start offset */
	uint32 dc_line_delta;		/* fb and cb skip counts */
	uint32 dc_buf_size;		/* fb and cb line size */

	uint32 h_timing_1;		/* horizontal timing...	*/
	uint32 h_timing_2;
	uint32 h_timing_3;

	uint32 v_timing_1;		/* vertical timing... */
	uint32 v_timing_2;
	uint32 v_timing_3;

	uint32 dc_cursor_x;		/* cursor x position */
	uint32 dc_cursor_y;		/* cursor y position */
	uint32 dc_cursor_color;		/* cursor color indexes */
	uint32 ss_line_compare;		/* line compare value */

	uint32 gp_blit_status;		/* blit status register	*/
};

/*
**	mmapped I/O functions
*/

#include "gx_mmio.inc"

/*
**	graphic framebuffer text management
*/

#include "gx_text16.inc"

/*
**	detect the presence of SoftVGA then the amount of memory allocated
**	to graphics by it
**
**	since VGA registers aren't going to be used by the driver ever
**	again, the driver doesn't reserve the region in order to
**	support multihead.
*/

static int GX_getgramsize (void) {
	uint8 oldindex, oldlock;
	uint32 ramsize = 0;

	/*	store SoftVGA current index and lock register value	*/

	oldindex = io_in8 (0x3D4);
	io_out8 (0x30, 0x3D4);
	oldlock = io_in8 (0x3D5);

	/*	attempt to lock SoftVGA regs	*/

	io_out8 (0, 0x3D5);

	if (io_in8 (0x3D5) == 0xFF) {
		/*	use two-byte sequence to unlock vga registers	*/

		io_out8 (0x57, 0x3D5);
		io_out8 (0x4C, 0x3D5);

		/*	if unlocking worked, SoftVGA is present */

		if (!io_in8 (0x3D5)) {
			if (!io_in8(0x3D5)) {
				/*	get amount of graphics ram	*/

				io_out8 (0x3E, 0x3D4);
				ramsize = (io_in8 (0x3D5) - 1) << 16;
			}
		}
	}

	/*	restore original lock and index */

	io_out8 (oldlock, 0x3D5);
	io_out8 (oldindex, 0x3D4);

	return ramsize;
}

/*
**	detect the MediaGX processor and companion chip and, if present,
**	allocate the chipset resources
*/

static int GX_detect (struct kgi_display *dpy) {
	struct	chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	struct	kgi_chipset *chipset;
	struct	mem_region *gx_control, *gx_companion, *gx_graphics;
	struct	kgi_mmio_region *lfb_mm;
	uint8	bus, dev_fn, gcr, cfg;
	uint32	id, i, physbase, gramsize, padsize;

	/*
	**	WARNING: Uses Linux PCI nonportable search functions for now.
	**
	**	This will be replaced by a call into KGI as soon as the
	**	function is implemented.
	*/

	if (!pcibios_present()) {
		ERROR ("PCI bios required for MediaGX\n");
		return -E(CHIPSET, UNKNOWN);
	}

	/*
	**	detect the GX main processor (acts as PCI bridge too)
	**	there can only be one per machine.
	*/

	if (pcibios_find_device (PCI_VENDOR_ID_CYRIX,
		PCI_DEVICE_ID_CYRIX_PCI_MASTER, 0, &bus, &dev_fn) !=
		PCIBIOS_SUCCESSFUL) {

		ERROR ("GX processor not detected\n");
		return -E(CHIPSET, UNKNOWN);
	}

	/*
	**	detect the companion chip (CX55x0)
	**	there can only be one per machine aswell.
	*/

	for (i=0; (id = (uint32) companion_id[i]) != -1; i++) {
		if (pcibios_find_device (PCI_VENDOR_ID_CYRIX,
			id, 0, &bus, &dev_fn) == PCIBIOS_SUCCESSFUL)
			break;
	}

	if (id == -1) {
		ERROR ("Companion chip unknown or not detected\n");
		return -E(CHIPSET, UNKNOWN);
	}

	private->companion = i;

	/*
	**	access the GCR register through mediaGX I/O port, in
	**	order to get the address of mmapped registers and
	**	framebuffer memory, and the size of the graphic scratchpad.
	*/

	io_out8 (GX_IOIDX_DIR0, GX_IOPORT_INDEX); /* doesn't work w/o that */
	io_out8 (GX_IOIDX_GCR, GX_IOPORT_INDEX);
	gcr = io_in8 (GX_IOPORT_DATA);

	physbase = (gcr & 3) << 30;
	padsize = (gcr & 12) ? ( (((gcr & 12) >> 2) + 1) << 10) : 0;
	private->padsize = padsize;

	/*	detect the amount of system memory allocated to graphics */

	if ( !(gramsize = GX_getgramsize ()) ) {
		ERROR ("Graphics subsystem not initialized by BIOS.\n");
		return -E(CHIPSET, UNKNOWN);
	}

	/*	now map control registers and graphics memory	*/

	gx_control  = &(private->gx_control);
	gx_control->baseptr = MEM_NULL;
	gx_control->base = physbase;
	gx_control->size = 0x10000;
	gx_control->decode = 0xFFFFFFFF;	/* decode all address bits */
	gx_control->name = "MediaGX control registers";

	gx_companion = &(private->gx_companion);

	if (i) {
		gx_companion->baseptr = MEM_NULL;
		pcibios_read_config_dword (bus, dev_fn, PCI_BASE_ADDRESS_3,
			&(gx_companion->base));
		gx_companion->size = 0x10000;
		gx_companion->decode = 0xFFFFFFFF;
		gx_companion->name = "MediaGX companion registers";
	}

	gx_graphics = &(private->gx_graphics);
	gx_graphics->baseptr = MEM_NULL;
	gx_graphics->base = physbase + 0x800000;
	gx_graphics->size = 0x800000;
	gx_graphics->decode = 0xFFFFFFFF;
	gx_graphics->name = "MediaGX graphics memory";

	if (mem_check_region (gx_control)) {
		ERROR ("GX control registers region already in use.\n");
		return -E(CHIPSET, UNKNOWN);
	}

	if ((i) && (mem_check_region (gx_companion))) {
		ERROR ("GX companion registers region already in use.\n");
		return -E(CHIPSET, UNKNOWN);
	}
	
	if (mem_check_region (gx_graphics)) {
		ERROR ("GX graphics memory region already in use.\n");
		return -E(CHIPSET, UNKNOWN);
	}

	mem_claim_region (gx_control);
	if (i) mem_claim_region (gx_companion);
	mem_claim_region (gx_graphics);

	/*
	**	reserve memory in the framebuffer for storing
	**	text data. this is not portable to another
	**	architecture than x86, but since MediaGX *is*
	**	an x86 CPU, it isn't a problem - could always be
	**	changed into a kgi_malloc_huge() if the GX video
	**	chip ever ran by itself on another architecture.
	**	using framebuffer memory has the advantage of
	**	not taking up any kernel memory, and is as fast,
	**	since framebuffer memory is in main memory thanks
	**	to the unified memory architecture.
	*/

	private->availgram = gramsize - (TEXT_MEMORY_SIZE+SPRITE_MEMORY_SIZE);
	private->textmem  = (mem_vaddr) ((char *)gx_graphics->baseptr) +
			     private->availgram;
	private->spritemem  = (mem_vaddr) ((char *)private->textmem) +
			     TEXT_MEMORY_SIZE;

	/*	map companion chip mmio space	*/

	if (i) {
		pcibios_read_config_byte (bus, dev_fn, 0x42, &cfg);
		pcibios_write_config_byte (bus, dev_fn, 0x42, cfg | 9);
	}

	/*	report chipset information	*/

	NOTICE (" MediaGX and CX55%d0 companion chip detected", i+1);
	NOTICE (" %u MB framebuffer and %u KB scratchpad detected",
			gramsize >> 20, padsize >> 10);

	/*	initialize LFB mmio		*/

	lfb_mm = &(private->lfb_mm);

	lfb_mm->name   = "MediaGX MMIO framebuffer";
	lfb_mm->type   = MMIO_FRAMEBUFFER;
	lfb_mm->prot   = MMIO_RDWR | MMIO_SHARED;
	lfb_mm->size   = gramsize;
	lfb_mm->psize  = gramsize;
	lfb_mm->pstart = physbase + 0x800000;
	lfb_mm->offset = 0;

	/*	initialize chipset structure	*/

	chipset = &(private->chipset);
	strcpy (chipset->manufact, "Cyrix");
	strcpy (chipset->model, "MediaGX");
	chipset->modes = GM_TEXT16|GM_8BIT|GM_15BIT|GM_16BIT;
	chipset->flags = 0;
	chipset->ramsize = gramsize;
	chipset->maxdots.x = 1280;
	chipset->maxdots.y = 1024;
	chipset->dclk.min = 0;
	chipset->dclk.max = 135000000;

	return EOK;
}

/*
**	standard timing adjustment function
*/

static void adjust_timing (struct kgi_mode *mode) {
	register int foo = (mode->request.graphtype == GT_TEXT16) 
		? mode->request.dpp.x : 8;
   
#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);

	if (mode->x.total % foo) {
		mode->x.total += foo - (mode->x.total % foo);
	}
   
	if (((mode->x.syncend / foo) & 0x3F) ==
		((mode->x.syncstart / foo) & 0x3F)) {
		mode->x.syncend -= foo;
	}

	if (((mode->x.blankend / foo) & 0x7F) ==
		((mode->x.blankstart / foo) & 0x7F)) {
		mode->x.blankend -= foo;
	}
   
	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F)) {
		mode->y.syncend--;
	}

	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF)) {
		mode->y.blankend--;
	}
   
#undef ADJUST

}

/*
**	KGI module functions
*/

/*	detect if chipset is present, and initialize the necessary data */

int kgim_chipset_init (struct kgi_card *card, struct kgi_display *dpy) {
	struct	chipset_private	*private;
	int	err;

	if ( (private = kgi_malloc_small
		(sizeof (struct chipset_private), GFP_KERNEL)) ) {

		memset (private, 0, sizeof (struct chipset_private));
		KGIM_PRIVATE (dpy, chipset ) = private;

		NOTICE (DRIVER_NAME " rev " DRIVER_REV);

		if ( (err = GX_detect (dpy)) == EOK ) {
			memcpy (&dpy->card->preset, &gx_preset,
				sizeof(struct kgi_mode));
			dpy->defmode = (ggi_mode *) &gx_preset;

			card->chipset = &(private->chipset);
			return EOK;
		}
		else {
			kgi_free_small(private);
			KGIM_PRIVATE (dpy, chipset) = NULL;

			return err;
		}
	}
	else {
		DEBUG2 ("Failed to allocate private data");
		return -ENOMEM;
	}
}

/*	prepare to unload the driver : deallocate everything	*/

void kgim_chipset_done (struct kgi_display *dpy) {
	struct chipset_private *private = KGIM_PRIVATE (dpy, chipset);

	if (private->gx_graphics.baseptr)
		mem_free_region (&(private->gx_graphics));

	if ((private->companion) && (private->gx_companion.baseptr))
		mem_free_region (&(private->gx_companion));

	if (private->gx_control.baseptr)
		mem_free_region (&(private->gx_control));

	kgi_free_small (private);
	KGIM_PRIVATE (dpy, chipset) = NULL;

	NOTICE (DRIVER_NAME " removed.");
}

/* check that the board can fullfill the request and adjust if necessary */

int kgim_chipset_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
				enum kgi_tm_cmd cmd) {
	struct chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	uint32 xvisible, yvisible, pxvisible, pyvisible, xvirt, yvirt;
	uint32 dppx, dppy, xmagnify, ymagnify, pixshift = 0;
	uint32 gramsize, textfrsize, lsize, fbufsize, cblinesize;
	ggi_graphtype gt;
	
	xvisible = mode->request.visible.x;
	yvisible = mode->request.visible.y;
	xvirt    = mode->request.virt.x;
	yvirt    = mode->request.virt.y;
	dppx     = mode->request.dpp.x;
	dppy     = mode->request.dpp.y;
	gramsize = private->availgram;

	switch (gt = mode->request.graphtype) {
		case GT_TEXT16:
			break;

		case GT_8BIT:
			pixshift = 0;
			break;

		case GT_15BIT:
		case GT_16BIT:
			pixshift = 1;
			break;

		default:
	                DEBUG0("graphtype %i not supported", gt); 					break;
	}
 
	switch (cmd) {
		case CMD_PROPOSE:
			if (gt == GT_TEXT16) {
				if (xvisible == GGI_AUTO) xvisible = 80;
				if (yvisible == GGI_AUTO) yvisible = 25;
				if (xvirt == GGI_AUTO) xvirt = xvisible;
				if (yvirt == GGI_AUTO) yvirt = yvisible;
				if (dppx == GGI_AUTO) dppx = 8;
				if (dppy == GGI_AUTO) dppy = 16;

				if ((dppx < 8) || (dppx > 9) || (!dppy))
					return -EINVAL;

				if (!(xvirt | yvirt))
					return -EINVAL;

				if (!xvirt)
					xvirt = TEXT_MEMORY_SIZE / (yvirt << 1);

				if ((xvirt*dppx) > 2048) xvirt = 2048 / dppx;

				pxvisible = xvisible * dppx;
				pyvisible = yvisible * dppy;

				if (!yvirt) {
					lsize = ((xvirt*dppx) + 1023) &
						0xFFFFFC00;
					
					if (pyvisible < 400)
						yvirt = gramsize / (lsize*dppy);
					else
					if (((gramsize-pyvisible*lsize-15) /
					    pyvisible) > 272)
						yvirt = gramsize / 
							((lsize+272)*dppy);
					else	yvirt = yvisible;
				}

				if ( ((xvirt << 1) * yvirt) >
					TEXT_MEMORY_SIZE) {
	                                NOTICE("Not enough text memory");
	                                return -ENOMEM;
	                        }
			}
			else {	/* if (gt == GT_TEXT16)	*/
				if (xvisible == GGI_AUTO) xvisible = 640;
				if (yvisible == GGI_AUTO) yvisible = 480;
				if (xvirt == GGI_AUTO) xvirt = xvisible;
				if (yvirt == GGI_AUTO) yvirt = yvisible;
				if (dppx == GGI_AUTO) dppx = 1;
				if (dppy == GGI_AUTO) dppy = 1;
				
				xvisible = (xvisible + 7) & 0x7FFFFFF8;
				yvisible = (yvisible + 7) & 0x7FFFFFF8;
				pxvisible = xvisible; pyvisible = yvisible;

				if (!(xvirt | yvirt))
					return -EINVAL;

				if (!xvirt)
					xvirt = gramsize / yvirt;

				if (xvirt > (2048 >> pixshift))
					xvirt = 2048 >> pixshift;
				else	xvirt = (xvirt+1023)&0xFFFFFC00;

				lsize = xvirt << pixshift;

				if (!yvirt) {
					if (yvisible < 400)
						yvirt = gramsize / lsize;
					else if (((gramsize-yvisible*lsize-15) /
						yvisible) > 272)
						yvirt = gramsize/(lsize+272);
					else	yvirt = yvisible;
				}

				if ((lsize * yvirt) > gramsize) {
					NOTICE ("Not enough graphic memory");
					return -ENOMEM;
				}
			}
			
			if ((pxvisible < 320) || (pyvisible < 175))
				return -EINVAL;

                        xmagnify = (pxvisible < 640) ? 2:1;
                        ymagnify = (pyvisible < 350) ? 2:1;

			mode->request.visible.x = xvisible;
			mode->request.visible.y = yvisible;
			mode->request.virt.x = xvirt;
			mode->request.virt.y = yvirt;
			mode->request.dpp.x = dppx;
			mode->request.dpp.y = dppy;
			mode->x.width = pxvisible * xmagnify;
			mode->y.width = pyvisible * ymagnify;
			mode->magnify.x = xmagnify;
			mode->magnify.y = ymagnify;
				
			mode->clk.mul = (pxvisible < 1024) ? 2 : 1;
			mode->clk.div = 1;

			return EOK;

		case CMD_LOWER:
			if (mode->dclk < private->chipset.dclk.min) {
				DEBUG1("DCLK too low (%i)", mode->dclk);
				return -E(CHIPSET,FAILED);
			}

			mode->dclk = (mode->dclk > private->chipset.dclk.max)
				? private->chipset.dclk.max : mode->dclk;
      
			adjust_timing(mode);
			return EOK;

		case CMD_RAISE:
			if (mode->dclk > private->chipset.dclk.max) {
				DEBUG1("DCLK too high (%i)", mode->dclk);
				return -E(CHIPSET,FAILED);
			}

			mode->dclk = (mode->dclk < private->chipset.dclk.min)
				? private->chipset.dclk.min : mode->dclk;
      
			adjust_timing(mode);
			return EOK;

		case CMD_CHECK:
			if ((mode->lclk.mul != 1) && (mode->lclk.mul !=2)) {
				DEBUG1("Invalid lclk multiplier %d",
					mode->lclk.mul);
				return -E(CHIPSET,FAILED);
			}

			if (mode->lclk.div != 1) {
				DEBUG1("Invalid CLK to LCLK divider %d",
					mode->lclk.div);
				return -E(CHIPSET,FAILED);
			}

			lsize = (((xvirt*dppx)+1023) & 0xFFFFFC00) << pixshift;
			fbufsize = lsize * yvirt * dppy;

			if ( (mode->magnify.x < 1) ||
				(mode->magnify.x > 2) ||
				(mode->magnify.y < 1) ||
				(mode->magnify.y > 2) ||
				(!xvirt) || (!yvirt) ||
				((xvisible*dppx) < 320) ||
				((yvisible*dppy) < 200) ||
				(xvirt < xvisible) || (yvirt < yvisible) ||
				(fbufsize > gramsize) ||
				(mode->dclk < private->chipset.dclk.min) ||
				(mode->dclk > private->chipset.dclk.max) ||
				(mode->x.width > private->chipset.maxdots.x)||
				(mode->y.width > private->chipset.maxdots.y)){

				DEBUG1 ("Out of range for hardware");
				return -E(CHIPSET, FAILED);
			}

			if (mode->magnify.y == 1) {
				cblinesize = ((gramsize-fbufsize-15)/(yvirt*dppy))
					     & 0xFFFFFFFC;

				if (cblinesize >= 20) {
					if (cblinesize > 272) cblinesize = 272;
				}
				else	cblinesize = 0;
			}
			else	cblinesize = 0;
			
			if (gt == GT_TEXT16) {
				textfrsize = xvirt * (yvirt << 1);

				if ( ((xvisible*dppx) > 1280) ||
					((yvisible*dppy) > 1024) ||
					(dppx < 8) || (dppx > 9) || (!dppy) ||
					(textfrsize > TEXT_MEMORY_SIZE) ) {

					DEBUG1 ("Out of range for textmode");
					return -E(CHIPSET, FAILED);
				}
				
				mode->flags  |= FL_CHIP_COLORTEXT;
				mode->fb      = private->textmem;
				mode->fb_size = TEXT_MEMORY_SIZE;
				mode->fsize   = textfrsize;
				mode->mem     = &fbtext_color_mem;
				mode->io      = &fbtext_color_io;
			}
			else {
				if (pixshift) {
					if ((xvisible > 1024) ||
					    (yvisible > 768) ||
					    (xvirt & 0x3FF) ||
					    (xvirt > 1024)) {

						DEBUG1 ("Out of range");
						return -E(CHIPSET, FAILED);
					}
				}
				else {
					if ((xvisible > 1280) ||
					    (yvisible > 1024) ||
					    (xvirt & 0x3FF) ||
					    (xvirt > 2048)) {

						DEBUG1 ("Out of range");
						return -E(CHIPSET, FAILED);
					}
				}

				mode->fb = (char *)
						private->gx_graphics.baseptr;
				mode->fb_size = gramsize - cblinesize * yvirt;
				mode->fsize   = fbufsize;
				mode->mem     = NULL;
				mode->io      = NULL;

				private->lfb_mm.size   = gramsize;
				private->lfb_mm.psize  = gramsize;
				private->lfb_mm.pstart =
					private->gx_graphics.base;
				private->lfb_mm.offset = 0;
				private->lfb_mm.next   = mode->mmio;

				mode->mmio = &(private->lfb_mm);
			}

			private->fblinesize = lsize;
			private->cblinesize = cblinesize;
			private->fbufsize   = fbufsize;
			return EOK;

		default:
			INTERNAL_ERROR;
			return -E(CHIPSET, FAILED);
	}
}

/*	switch the board to requested display mode	*/

void kgim_chipset_set_mode (struct kgi_display *dpy, struct kgi_mode *mode) {
	struct chipset_private *private = KGIM_PRIVATE (dpy, chipset);
	struct chipsetregs	cr;
	ggi_graphtype	gt;
	uint32		fetchsize, fblinesize, cblinesize, fbufsize;
	uint32		palmode;

	/*
	**	setup display configuration registers
	*/

	gt = mode->request.graphtype;
	palmode = (gt == GT_TEXT16)||(gt == GT_8BIT);

	cr.general_cfg =
		((mode->lclk.mul > 1) ? DC_GCFG_DDCK : 0) |
		((mode->lclk.div >= 2) ? DC_GCFG_DPCK : 0) |
		((mode->magnify.y >= 2) ? DC_GCFG_LDBL : 0) |
		0x6500 |		/* FIFO high-pri start and end */
		DC_GCFG_DECE | DC_GCFG_CMPE |
		DC_GCFG_DFLE | DC_GCFG_SSLC;
		
	if ((mode->request.visible.x * mode->request.dpp.x) < 1024)
		cr.general_cfg |= (mode->lclk.mul << DC_GCFG_DCLK_POS);
	else	cr.general_cfg |= ((mode->lclk.mul + 1) << DC_GCFG_DCLK_POS);  

	cr.timing_cfg =
		((mode->magnify.x >= 2) ? DC_TCFG_PXDB : 0) |
		DC_TCFG_VIEN | DC_TCFG_TGEN | DC_TCFG_BLKE |
		DC_TCFG_VSYE | DC_TCFG_HSYE | DC_TCFG_FPPE |
		DC_TCFG_CVSP | DC_TCFG_CHSP |
		((gt == GT_TEXT16) ? DC_TCFG_BLNK|DC_TCFG_BKRT : 0);

	cr.output_cfg =
		((mode->lclk.mul > 1) ? DC_OCFG_2XCK :
			(DC_OCFG_PDEH | DC_OCFG_PDEL | DC_OCFG_CKSL)) |
		DC_OCFG_PCKE | 
		((gt == GT_15BIT) ? DC_OCFG_555 : 0) |
		(palmode ? DC_OCFG_8BPP : 0);

	/*
	**	setup memory configuration registers
	*/

	cr.fb_st_offset = 0;
	cr.cur_st_offset = private->availgram + TEXT_MEMORY_SIZE;

	fblinesize = private->fblinesize;
	cblinesize = private->cblinesize;
	fbufsize   = private->fbufsize;
	fetchsize  = (((gt == GT_TEXT16)||(gt == GT_8BIT)) ?
		     mode->request.visible.x : (mode->request.visible.x<<1)) *
		     mode->request.dpp.x;

	if (cblinesize) {
		/*
		** enough room to put the compression buffer
		** after uncompressed framebuffer in
		** graphics memory
		*/
		
		cr.cb_st_offset = (fbufsize + 15) & 0xFFFFFFF0;
		cr.dc_line_delta = ((fblinesize>>2)&0x3FF) | (cblinesize<<10);
		cr.dc_buf_size = (((fetchsize+7)>>3) & 0x1FF) |
					((cblinesize - 12)<<7);

		DEBUG1 ("Compression buffer line size: %d", cblinesize);
	}
	else {
		/*
		** not enough memory for compression buffer
		** deactivate compression :(
		*/

		cr.cb_st_offset  = fbufsize;
		cr.dc_line_delta = (fblinesize >> 2) & 0x3FF;
		cr.dc_buf_size   = ((fetchsize+7) >> 3) & 0x1FF;
		cr.general_cfg  -= (DC_GCFG_DECE | DC_GCFG_CMPE);

		DEBUG1 ("Not enough memory for compression - deactivated.");
	}			

	/*	setup display properties for accelerator engine		*/

	cr.gp_blit_status =
		((!palmode) ? BC_16BPP : BC_8BPP) |
		((fblinesize > 1024) ? BC_FB_WIDTH_2048 : BC_FB_WIDTH_1024);

	/*	setup horizontal timing	registers	*/

	cr.h_timing_1 =
		((mode->x.total - 1) & 0x7FF) << 16 |	  /* htotal */
		((mode->x.width - 5) & 0x7FF);		  /* hactive */

	cr.h_timing_2 =
		((mode->x.blankend - 1) & 0x7FF) << 16 |  /* hblankend */
		((mode->x.blankstart - 1) & 0x7FF);	  /* hblankstart */

	cr.h_timing_3 =
		((mode->x.syncend - 1) & 0x7FF) << 16 |   /* hsyncend */
		((mode->x.syncstart - 1) & 0x7FF);	  /* hsyncstart */

	/*	setup vertical timing registers		*/

	cr.v_timing_1 =
		((mode->y.total - 1) & 0x7FF) << 16 |     /* vtotal */
		((mode->y.width - 1) & 0x7FF);		  /* vactive */

	cr.v_timing_2 =
		((mode->y.blankend - 1) & 0x7FF) << 16 |  /* vblankend */
		((mode->y.blankstart - 1) & 0x7FF);	  /* vblankstart */

	cr.v_timing_3 =
		((mode->y.syncend -  1) & 0x3FF) << 16 |  /* vsyncend */
		((mode->y.syncstart - 1) & 0x3FF);	  /* vsyncstart */

	/*	reset cursor position, colors and splitline		*/

	cr.dc_cursor_x = cr.dc_cursor_y = 0;
	cr.dc_cursor_color = 0x1110;
	cr.ss_line_compare = mode->request.visible.y * mode->request.dpp.y *
			     mode->magnify.y;

	/*	dump GX registers ready to be programmed		*/

	DEBUG1 ("Ready to program:");
	DEBUG1 ("general_cfg %08X, timing_cfg %08X, output_cfg %08X",
		cr.general_cfg, cr.timing_cfg, cr.output_cfg);
	DEBUG1 ("fb_st_offset %08X, cb_st_offset %08X, cur_st_offset %08X",
		cr.fb_st_offset, cr.cb_st_offset, cr.cur_st_offset);
	DEBUG1 ("dc_line_delta %08X, dc_buf_size %08X",
		cr.dc_line_delta, cr.dc_buf_size);
	DEBUG1 ("h_timing_1 %08X, h_timing_2 %08X, h_timing_3 %08X",
		cr.h_timing_1, cr.h_timing_2, cr.h_timing_3);
	DEBUG1 ("v_timing_1 %08X, v_timing_2 %08X, v_timing_3 %08X",
		cr.v_timing_1, cr.v_timing_2, cr.v_timing_3);
	DEBUG1 ("gp_blit_status %08X", cr.gp_blit_status);

	/*	ok. we'll do the actual hardware programming now.	*/

	program_chipset (dpy, mode, &cr);

	/*	setup display properties 				*/

	if (gt == GT_TEXT16) {
		dpy->set_font		= fbtext_set_font;
		dpy->set_origin		= gx_set_origin;
		dpy->set_split		= gx_set_split;
		dpy->set_frame		= NULL;

		dpy->show_pointer	= NULL;
		dpy->hide_pointer	= NULL;
		dpy->undo_pointer	= NULL;

		dpy->show_cursor	= gx_show_pointer;
		dpy->hide_cursor	= gx_hide_pointer;
		dpy->undo_cursor	= gx_undo_pointer;

		dpy->fb.dpy		= dpy;
		dpy->fb.size.x		= mode->request.virt.x;
		dpy->fb.size.y		= mode->request.virt.y;
		dpy->fb.bpp		= MAP_16_BPP;
		dpy->fb.pbuf		= private->textmem;
		dpy->fb.zbuf		= NULL;
		dpy->fb.ops		= &fbtext_color_mem;

		/*	handle textmode emulation			*/

		fbtext_graphmem		= private->gx_graphics.baseptr;
		fbtext_linesize		= fblinesize;
		fbtext_set_textclut (dpy);

	}
	else {	/* if (gt == GT_TEXT16) { */
		dpy->set_font		= NULL;
		dpy->set_origin		= gx_set_origin;
		dpy->set_split		= gx_set_split;
		dpy->set_frame		= NULL;

		dpy->show_pointer	= gx_show_pointer;
		dpy->hide_pointer	= gx_hide_pointer;
		dpy->undo_pointer	= gx_undo_pointer;

		dpy->show_cursor	= NULL;
		dpy->hide_cursor	= NULL;
		dpy->undo_cursor	= NULL;

		dpy->fb.dpy		= dpy;
		dpy->fb.size.x		= mode->request.virt.x;
		dpy->fb.size.y		= mode->request.virt.y;
		dpy->fb.bpp		= (gt == GT_8BIT) ?
					  MAP_8_BPP : MAP_16_BPP;
		dpy->fb.pbuf		= private->gx_graphics.baseptr;
		dpy->fb.zbuf		= NULL;
		dpy->fb.ops		= NULL;
	}

	DEBUG1 ("Mode set complete");
}

/* execute a chipset command */

int kgim_chipset_command (struct kgi_graphic *state, ggi_uint cmd,
				ggi_uint iobuf) {

	switch (cmd) {
		case CHIP_SETSPLITLINE:
			if (iobuf > (state->dev.mode.visible.y*state->dev.mode.dpp.y))
				return -E (CHIPSET,BOGUS);

			kgi_set_split(&(state->dev),
				iobuf * state->dev.dpy->mode.magnify.y);
			return EOK;

		default:
			return -E(CHIPSET, UNKNOWN);
	}

}
