/*\
|*| --------------------------------------------------------------------------
|*|	ET4000 MMIO
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Jason McMullan, [jmcc@cs.cmu.edu]
|*|     Copyright (C) 1997 by Andreas Vogler, [duff@allgaeu.org]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
|*|	This code was derived from the following source of information:
|*|	[1]	IBM VGA driver by Jason McMullan
|*|	[2]	VGADOCS by Finn Thoegersen
\*/

#define MM_PROT_RW     (VM_READ|VM_WRITE|VM_SHARED)
#define MM_PROT_W      (VM_WRITE|VM_SHARED)
#define MMAP_IBM_VGA_4 (MMAP_PRIVATE | 4)

static void select_segment(int segment)
{
	ggi_uint hi;

	if(chip>=ET4000_W32) {
		hi = segment & 0x30;
		io_out8(hi | (hi >> 4), EXT_BANK);
	}
	segment &= 0x0f;
	io_out8(segment | (segment << 4), SEG_SELECT);
}

static void mmap_setpage_graphics(struct kgi_mmio_region *mm, ggi_uint offset)
{
	int page=(offset>>16) & 0x3F;
	
	select_segment(page);	
	
	mm->offset=(page<<16);
}

static struct kgi_mmio_region mm_text16= {
	NULL,		/* Next */
	"text16",	/* Name */
	MMAP_FRAMEBUFFER,/* Type (Linear) */
	MM_PROT_RW,	/* Protection map */
	
	0x10000,	/* Total size of area */
	0x10000,	/* Physical size (64k) */
	0xA0000,	/* Physical address of the windows */
	0,		/* Logical offset */
	
	NULL		/* Set offset function */
};

static struct kgi_mmio_region mm_graphics= {
	NULL,		/* Next */
	"ET4000",		/* Name */
	MMAP_FRAMEBUFFER,/* Type (Linear) */
	MM_PROT_RW,	/* Protection map */
	
	0x0,		/* Total size of area (depends on mode)*/
	0x10000,	/* Physical size (64k) */
	0xA0000,	/* Physical address of the windows */
	0,		/* Logical offset */
	
	mmap_setpage_graphics	/* Set offset function */
};

static struct kgi_mmio_region mm_linear_graphics= {
	NULL,		/* Next */
	"ET4000 W32 linear", /* Name */
	MMAP_FRAMEBUFFER,/* Type (Linear) */
	MM_PROT_RW,	/* Protection map */
	
	0,		/* Total size of area (depends on mode)*/
	0,		/* Physical size (4 M) */
	0x0,		/* Physical address of the windows (autodetected)*/
	0,		/* Logical offset */
	
	NULL		/* Set offset function */
};



