/*    
|*| ---------------------------------------------------------------------------
|*|	wd90c00 (PVGA1B, wd90c90) IO functions and variables
|*| ---------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1997 by Brian S. Julin, <bri@calyx.com>
|*|
|*|	This header is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU Library General Public License as 
|*|	published by the Free Software Foundation; either version 2, or 
|*|     (at your option) any later version.
|*|
|*|	You should have received a copy of the GNU Library General Public 
|*|     License along with this program; see the file COPYING.LGPL.  
|*|	If not, write to the Free Software Foundation, 675 Mass Ave, 
|*|     Cambridge, MA 02139, USA.
|*| ---------------------------------------------------------------------------
\*/

#define _wd90c00_io_c

/* MAINTAINER	Brian S. Julin, <bri@calyx.com> */

#include <WD/wd90c00_io.h>
#include "pvga1a_io.c"

/**** CRT Controller Register Manipulation ****/

void save_CRT(uint8 * buf)   
{
  register uint8 i; 
  int lockval1;
  lockval1 = runlockCRT();
  for (i = 0; i < NoCRTRegs; i++) 
    {
      *(buf++) = inbCRT(i);
    }
  outbCRT(lockval1, CRT_UPCRT_LOCK); /* Restore CRT lock */
}

void restore_CRT(uint8 *buf)  /* wd90c00+ has more locks */
{
  register uint8 i;
  int lockval1, lockval2;

  lockval1 = rwunlockCRT();
  
  outbCRT(inbCRT(CRT_HBLANKEND) | CR03_IS_VGA,CRT_HBLANKEND);
  
  for (i = 0; i < NoCRTRegs; i++) 
    {
      switch (i) 
	{
	case CRT_HBLANKEND:
	  *buf |= CR03_IS_VGA;
	  break;
	case CRT_VSYNCEND:
	  *buf &= ~CR11_LOCKTIMING;
	  break;
	default:
	  break;
	}
      outbCRT(*(buf++), i);
    }

  outbCRT(((lockval1 & 0xff00) >> 8 ), CRT_UPCRT_LOCK); /* Restore CRT lock */
  lockval2 = wunlockGRC();                       /* Unlock locking register */
                                   /* Restore locks on CRT timing registers */
  outbGRC(lockval1 & 0xff, GRC_CRTC_CTL); 
  outbGRC(lockval2, GRC_UPGRC_LOCK);                    /* Restore GRC lock */

}

#include "pvga1a_io.c"








