/* ---------------------------------------------------------------------------
**	Mystique clock driver
** ---------------------------------------------------------------------------
**
**	Copyright	1997		Matthias Grimrath
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
** ---------------------------------------------------------------------------
**	This code was derived from the following sources of information:
**
**	MGA 1064SG Developer Specification
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Matthias_Grimrath
#define	DRIVER_NAME	"Mystique clock driver"
#define	DRIVER_REV	"$Revision: 1.8 $"

#include <ggi/module.h>
#include "pll.h"

#include <Matrox/mga_common_regs.h>
#include <Matrox/mga_common_mmio.h>
#include <Matrox/mga1064_dac.h>
/* Driver specific things are included by mga_common_*.h */

static struct kgi_clock clock =
{
	"Matrox Graphics Inc",	        /* manufacturer         */
	"Mystique",			/* model		*/
	GM_ALL,				/* supported modes 	*/
	0,				/* special capabilities	*/
	{ 0x7FFF, 0x7FFF },		/* maximum resolution	*/

	CLK_TYPE_PROG,			/* clock type		*/
	{{
		5648969, 220000000,	/* Clock range		*/
	}}
};

/* The output frequency is calculated as
** fout = (fref * factor.mul/factor.div * dividend/divisor) >> post-VCO;
*/

static const struct pll_limits dclk_pll = 
{
	14318180,			/* fref				*/
	{ 1, 1 },			/* factor		 	*/
	{ 5648969, 220000000 },	        /* fvco range			*/
	{ 0, 3 },			/* post-VCO divider range	*/
	{ 101, 128 },			/* dividend range		*/
	{ 2, 32 }			/* divisor range		*/
};


#define	INVALID_COMMAND	((ggi_uint) -1)

static ggi_uint pll_command(struct pll_state *s, struct kgi_mode *mode)
{
	ggi_uint x;
	ASSERT((s->mul - 1 < 128) && (s->mul - 1 > 99));
	ASSERT((s->div - 1 < 32) && (s->div - 1 > 0));
	ASSERT(s->p < 4);
	ASSERT((s->f >= 5648969) && (s->f <= 220000000));

	if (s->fvco < 100000000) {
	        x = 0x00;
	} else if (s->fvco < 140000000) {
	        x = 0x08;
	} else if (s->fvco < 180000000) {
	        x = 0x10;
	} else {
	        x = 0x18;
	}
	DEBUG2("f %d fvco %d s %d", s->f, s->fvco, x >> 3);

	return ((s->div-1) << 16) | ((s->mul-1) << 8) | x | ((1 << s->p)-1);
}

static void pll_set(struct kgi_display *dpy, ggi_uint cmd)
{
	DEBUG2("cmd 0x%x", cmd);

	/* We assume the display has been turned off by the chipset driver */

	EDAC_out8(dpy, EDAC_in8(dpy, XPIXCLKCTRL) | PIX_CLK_DIS, XPIXCLKCTRL);
	EDAC_out8(dpy, cmd >> 16, XPIXPLLCM);
	EDAC_out8(dpy, cmd >> 8, XPIXPLLCN);
	EDAC_out8(dpy, cmd, XPIXPLLCP);
	while ((EDAC_in8(dpy, XPIXPLLSTAT) & PIX_LOCK) == 0) ;
	EDAC_out8(dpy, EDAC_in8(dpy, XPIXCLKCTRL) & ~PIX_CLK_DIS, XPIXCLKCTRL);
}

/*
**	KGI module interface
*/

#define	NO_KGI_INIT_DONE_COMMAND
#include "pll.inc"

ggi_sint kgim_clock_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	card->clock = &clock;
	return EOK;
}

void kgim_clock_done(struct kgi_display *dpy)
{
	dpy->card->clock = NULL;
	NOTICE(DRIVER_NAME " removed");
}

ggi_sint kgim_clock_command(struct kgi_graphic *state, ggi_uint cmd, ggi_uint arg)
{
	return -E(CLOCK, UNKNOWN);
}
