/* 
** Cirrus Logic 542x framebuffer support
**
** Copyright (C) 1997    Martin Erhardsen [mee@daimi.aau.dk]
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
**-----------------------------------------------------------------------
*/

#include <ggi/maintainers.h>
#define MAINTAINER      Martin_Erhardsen
#define DRIVER_NAME     "CL-GD542x Graphics Driver"
#define DRIVER_REV      "$Revision: 1.9 $"

/*\
|*|     KGI includes
\*/
#include <asm/io.h>

#include <ggi/module.h>
#include <ggi/ggi_commands.h>
#include "Cirrus/542x.h"

#if 0
/*
**     Accelerated drawing functions
*/

volatile uint32 *cirrus_mmio;
/* FIXME: Set this somewhere */

ggi_drawinfo *drwinfo;
/* FIXME: What should this be */

static int pixwidths[GT_LAST] = {0,0,0,0,1,2,2,3,4}; 

static int cirrus_mmio_drawbox(struct kgi_graphic *gr, void *iobuf)
{	
	struct box *boxp = (struct box *)iobuf;
	int pixw;

	if (gr->dev.mode.graphtype < GT_8BIT)
		return -E(GRAPHICS,NOSUP_MODE_MMAP);

	pixw =  pixwidths[gr->dev.mode.graphtype];

	if (boxp->w && boxp->h) {	/* 0 width is not OK! */

		cirrus_mmio[1] = drwinfo->fg_color;
		((uint16 *)cirrus_mmio)[4] = boxp->w--;
		((uint16 *)cirrus_mmio)[5] = boxp->h--;
		cirrus_mmio[3] = gr->dev.mode.virt.x*pixw;
		cirrus_mmio[4] = gr->dev.mode.virt.x*pixw*boxp->y
			+boxp->x*pixw;
		cirrus_mmio[5] = 0;
		cirrus_mmio[6] = 0x040D00C0 | (pixw<<4);
		((uint8 *)cirrus_mmio)[32] = 0x02;
		while (((uint8 *)cirrus_mmio)[32]&0x01);
	}

	return(0);
}

static int cirrus_mmio_copybox(struct kgi_graphic *gr,void *iobuf)
{	
	struct cbox *boxp = (struct cbox *)iobuf;
	int dir = 0, pixw = 0;

	if (gr->dev.mode.graphtype < GT_8BIT) 
		return -E(GRAPHICS,NOSUP_MODE_MMAP);

	pixw =  pixwidths[gr->dev.mode.graphtype];

	if (boxp->w&&boxp->h) {		/* 0 width is not OK! */

		/* calculate direction */
		if (boxp->y > boxp->y2) {
			dir = 0;
		} else if (boxp->y < boxp->y2) {
			dir = 1;
		} else if (boxp->y == boxp->y2) {
			dir = (boxp->x < boxp->x2);
		}
		((uint16 *)cirrus_mmio)[4] = boxp->w--;
		((uint16 *)cirrus_mmio)[5] = boxp->h--;
		cirrus_mmio[3] = gr->dev.mode.virt.x*pixw*0x10001;
		cirrus_mmio[4] = gr->dev.mode.virt.x*pixw*boxp->y2
			+boxp->x2*pixw;
		cirrus_mmio[5] = gr->dev.mode.virt.x*pixw*boxp->y
			+boxp->x*pixw;
		cirrus_mmio[6] = 0x000D0000 | dir;
		((uint8 *)cirrus_mmio)[32] = 0x02;
		while (((uint8 *)cirrus_mmio)[32]&0x01);		
	}
	return 0;
}
#endif

/*
**	KGI module interface
*/

int kgim_graphics_init(struct kgi_card *card, struct kgi_display *dpy)
{
	/* !!! We should really do something here !!! */
	NOTICE(DRIVER_NAME " rev ." DRIVER_REV " loaded.\n");
	return EOK;
}

void kgim_graphics_done(struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " removed.\n");
}

int kgim_graphics_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			    enum kgi_tm_cmd cmd)
{
	return EOK;
}

void kgim_graphics_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{
}

int kgim_graphics_command (struct kgi_graphic *gr, ggi_uint cmd,ggi_uint arg)
{
	ggi_suggest *sugg;
	const char *name;
	
	switch (cmd) {
	case GRAPHICS_GETSUGGEST:
		sugg=(ggi_suggest *)(gr->io_buf);
		gr->io_res=gr->io_buf;
		
		if (sugg->id<0 || sugg->id > 1)
			return -E(GRAPHICS,BOGUS);

		if (sugg->id==1) { /* Generic RAMDAC support */
			if (gr->dev.mode.graphtype>GT_8BIT)
				return -E(GRAPHICS,BOGUS);
			strcpy(sugg->name,"generic-ramdac");
			sugg->args[0]=0;
			sugg->id=0;
			return 0;
		}
		switch (gr->dev.mode.graphtype) {
		case GT_TEXT16:
			name="generic-text-16";
			break;
		case GT_1BIT:
			name="generic-linear-1";
			break;
		case GT_4BIT:
			name="vendor-IBM-vga-4bpp";
			break;
		case GT_8BIT:
			name="generic-linear-8";
			break;
		case GT_15BIT:
			name="generic-linear-15";
			break;
		case GT_16BIT:
			name="generic-linear-16";
			break;
		case GT_24BIT:
			name="generic-linear-24";
			break;
		case GT_32BIT:
			name="generic-linear-32";
			break;
			
		default:
			return -E(GRAPHICS,BOGUS);
		}
		strcpy(sugg->name,name);
		sugg->args[0]=0;
		sugg->id=(gr->dev.mode.graphtype<=GT_8BIT);  
		/* Next is RAMDAC.. */
		return 0;
#if 0
	case GRAPHICS_DRAWBOX:
		if (gr->dev.mode.graphtype>=GT_8BIT) 
			switch(chip_version) 
			{
			case CLGD5436_ID:
			case CLGD5446_ID:
				return cirrus_mmio_drawbox(gr, gr->io_buf);
			case CLGD5434_ID:
			case CLGD5430_ID:
			case CLGD5429_ID:
				return cirrus_mmio_patfill(gr, gr->io_buf);
			case CLGD5428_ID:
			case CLGD5426_ID:
				return cirrus_io_patfill(gr, gr->io_buf);
			}
		return -E(GRAPHICS,NOSUP_ALWAYS_MMAP);
	case GRAPHICS_COPYBOX:
		if (gr->dev.mode.graphtype>=GT_8BIT) 
			switch(chip_version) 
			{
			case CLGD5436_ID:
			case CLGD5446_ID:
			case CLGD5434_ID:
			case CLGD5430_ID:
			case CLGD5429_ID:
				return cirrus_mmio_copybox(gr, gr->io_buf);
			case CLGD5428_ID:
			case CLGD5426_ID:
				return cirrus_io_copybox(gr, gr->io_buf);
			}
		return -E(GRAPHICS,NOSUP_ALWAYS_MMAP);
#endif
	default:
		break;
	}
			
	return -E(GRAPHICS,NOSUP_ALWAYS_MMAP);
}
