/* ---------------------------------------------------------------------------
**	Matrox generic acceleration driver
** ---------------------------------------------------------------------------
**
**	Copyright (C) 1997 Jonas Borgstrm
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
** 
** ---------------------------------------------------------------------------
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Jonas_Borgstrom
#define	DRIVER_NAME	"Matrox Millennium (I/II) accelerator driver"
#define	DRIVER_REV	"$Revision: 1.14 $"

#undef	DEBUG_IO

/*
**	KGI includes
*/

#include <ggi/module.h>
#include <ggi/types.h>
#include <ggi/ggi_commands.h>

#include <linux/delay.h>

#include "Matrox/mga_common.h"
#include "Matrox/mga2x64_main.h"
#include "Matrox/mga_common_regs.h"

/*#define NUMPINGPONGPAGES 2*/
#define MGA2X64WWAITFREE() while(mem_in32(STATUS + 2) & 0x01);

const char *sugg_default[GT_LAST][5] = {
	{ "generic-text-16","generic-ramdac",NULL,NULL,NULL },
	{ "generic-text-32","generic-ramdac",NULL,NULL,NULL },
	{ "generic-linear-1","generic-ramdac",NULL,NULL,NULL },
	{ "vendor-IBM-vga-4bpp",NULL,NULL,NULL,NULL},
	{ "generic-linear-8","generic-ioctl","generic-ramdac",NULL,NULL},
	{ "generic-linear-15",NULL,NULL,NULL,NULL},
	{ "generic-linear-16","generic-ioctl",NULL,NULL},
	{ "generic-linear-24","generic-ioctl","generic-gamma",NULL,NULL},
	{ "generic-linear-32","generic-ioctl","generic-gamma",NULL,NULL}
};

/*
**	KGI module interface
*/
int kgim_graphics_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev " DRIVER_REV " loaded");
	/* !!! Check for Chipset and Version here !!! */

	return EOK;
}

void kgim_graphics_done(struct kgi_display *dpy)
{
	NOTICE1(DRIVER_NAME " removed");
}
/*
 * Check for a valid graphic mode.
 */
int kgim_graphics_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
	enum kgi_tm_cmd cmd)
{
	ggi_uint chiprounding = 1, tmp = 0, i;
	ggi_uint interleave = (dpy->card->chipset->ramsize > (2 MB));

	/* 
	 * Valid virtual X-resolutions
	 */
	int width[] = { 640, 768, 800, 960, 1024, 1152, 1280,
			1600, 1920, 2048, 0 };

	switch (cmd) {

	case CMD_CHECK:
		/* 
		 * Check for a valid mode
		*/
		if (mode->request.graphtype == GT_TEXT16 ||
			mode->request.graphtype == GT_4BIT ||
			mode->request.graphtype == GT_15BIT) {

			return EOK;
		}

/*		if (mode->request.graphtype == GT_8BIT &&
			mode->request.virt.x == 320 &&
			mode->request.virt.y == 200) {

			return EOK;
		}*/

		switch (mode->request.graphtype) {

		case GT_8BIT:
			chiprounding = interleave ? 128 : 64;
			break;

		case GT_16BIT:
			chiprounding = interleave ? 64 : 32;
			break;

		case GT_24BIT:
			chiprounding = interleave ? 128 : 64;
			break;

		case GT_32BIT:
			chiprounding = interleave ? 32 : 16;
			break;

		default:
			INTERNAL_ERROR;
		}

		/*
		 * Search for a valid virtual x resolution
		 */
		i = 0;
		do
	  	{
			tmp = width[i];
			if(width[i] == 0) return -E(GRAPHICS,NOSUP_ALWAYS_CANT);
			i++;
	     	} while((tmp < mode->request.virt.x) || (tmp % chiprounding));
	   	mode->request.virt.x = tmp;

		return EOK;

	default:
		return EOK;
	}
}
/*
 * Set the accelerator's foreground color
 */
static void mga2x64_setfgcolor(struct kgi_display *dpy,
			       unsigned long fg)
{
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
	unsigned long pixelfg = fg;
   
	switch(dpy->mode.request.graphtype) {
		case GT_8BIT:
			pixelfg &=0xff;   
			pixelfg |= (pixelfg << 8) |
			  (pixelfg << 16) | (pixelfg << 24);
			break;
		case GT_16BIT:
			pixelfg &=0xffff;   
			pixelfg |= pixelfg << 16;
			break;
		case GT_24BIT:
			pixelfg &= 0xffffff;
			break;
		case GT_32BIT:
			pixelfg |= 0xff000000;
			break;
		default:
	   		return;
	}
	mem_out32(pixelfg, FCOL);
}
/*
 * Copy Box using the accelerator
 */
static int mga2x64_copybox(struct kgi_display *dpy,
			   struct kgi_graphic *gr,void *iobuf)
{	
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
     	struct cbox *boxp = (struct cbox *)iobuf;
   
	if(boxp->y2 > boxp->y)
	{
	   
		mem_out32((boxp->y + boxp->h - 1) *
			  dpy->mode.request.virt.x +
			  boxp->x + boxp->w - 1, AR0);
		mem_out32((boxp->y + boxp->h - 1) *
			  dpy->mode.request.virt.x + boxp->x, AR3);
		mem_out32(4, SGN);
		mem_out32( - dpy->mode.request.virt.x, AR5);
		mem_out32(((boxp->y2 + boxp->h - 1) << 16) | boxp->h, YDSTLEN);
	} else {
	   
		mem_out32(boxp->y * dpy->mode.request.virt.x
			  + boxp->x + boxp->w - 1, AR0);
		mem_out32(boxp->y * dpy->mode.request.virt.x + boxp->x, AR3);
		mem_out32(0, SGN);
		mem_out32(dpy->mode.request.virt.x, AR5);
		mem_out32((boxp->y2 << 16) | boxp->h, YDSTLEN);
	}
	mem_out32(((boxp->x2 + boxp->w - 1) << 16) | boxp->x2, FXBNDRY);
	mem_out32(0x040c4008, (void *)((ulong)DWGCTL + ACCEL_GO));
   
	MGA2X64WWAITFREE();
	return EOK;
}
/*
 * Draw a box with the accelerator
 */
static int mga2x64_drawbox(struct kgi_display *dpy,
			   struct kgi_graphic *gr, void *iobuf)
{
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
	struct box *boxp = (struct box *)iobuf;

	if (boxp->w && boxp->h) {	/* 0 width is not OK! */
		mga2x64_setfgcolor(dpy, gr->gc->fg_color);
   
		mem_out32(((boxp->x + boxp->w) << 16) | boxp->x, FXBNDRY);
		mem_out32((boxp->y << 16) | boxp->h, YDSTLEN);
		mem_out32(0x00C7804, DWGCTL + ACCEL_GO);
   
		MGA2X64WWAITFREE();
	}
	return EOK;
}
/* 
 * Draw a line using the accelerator
 */
static int mga2x64_drawline(struct kgi_display *dpy,
			    struct kgi_graphic *gr, void *iobuf)
{
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
	struct line *linp = (struct line *)iobuf;
   
	mga2x64_setfgcolor(dpy, gr->gc->fg_color);
   
	mem_out32(linp->x | (linp->y << 16), XYSTRT);
	mem_out32(linp->x2 | (linp->y2 << 16), XYEND);
	mem_out32(0x040C4803, DWGCTL + ACCEL_GO);
  
	MGA2X64WWAITFREE();
	return EOK;
}
/*
 * Setup the accelerator engine.
 */
void kgim_graphics_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset);
	uint32 shift = 0;
   
	switch (mode->request.graphtype) {

	case GT_8BIT:
		shift = 0;
		break;

	case GT_16BIT:
		shift = 1;
		break;

	case GT_24BIT:
		shift = 3;
		break;

	case GT_32BIT:
		shift = 2;
		break;

	default:
		return;
	}

	mem_out32(0xFFFF0000, CXBNDRY);
	mem_out32(0x00000000, YTOP);
	mem_out32(0x007FFFFF, YBOT);
	mem_out32(0x00000000, FCOL);
	mem_out32(0x00000000, SHIFT);
	mem_out32(0x01000000, OPMODE);
	mem_out32(shift     , MACCESS);
	mem_out32(0x00000000, YDSTORG);
	mem_out32(0xFFFFFFFF, PLNWT);
	mem_out32(mode->request.virt.x, PITCH);
}

int kgim_graphics_command (struct kgi_graphic *gr, ggi_uint cmd, ggi_uint arg)
{
	ggi_suggest *sugg;
	const char *name;
   
	if (!(gr->dev.state & KGI_MAPPED_ON_DISPLAY) ) {

		/* Don't accel inactive screen! */
	        return -E(GRAPHICS,NOSUP_MODE_MMAP);
	}
   
	switch (cmd) { 
      
	case GRAPHICS_GETSUGGEST:
		sugg = (ggi_suggest *)(gr->io_buf);
		gr->io_res = gr->io_buf;
      
		if (sugg->id < 0 || sugg->id > 4) {

			return -E(GRAPHICS,BOGUS);
		}
      
		name = sugg_default[gr->dev.mode.graphtype][sugg->id];
		if (!name) {

			return -E(GRAPHICS,BOGUS);
		}
/*		if (gr->dev.mode.graphtype == GT_8BIT &&
			gr->dev.mode.virt.x == 320 &&
			gr->dev.mode.virt.y == 200 &&
			strcmp("generic-ioctl", name) == 0) {
		   
			return -E(GRAPHICS,BOGUS);
		}*/
      
		strcpy(sugg->name, name);
		sugg->args[0] = 0;
		sugg->id++;	/* Next is RAMDAC */
		if (sugg->id >= 4 ||
		    sugg_default[gr->dev.mode.graphtype][sugg->id] == NULL) {

			sugg->id = 0;
		}
		return 0;

	case GRAPHICS_COPYBOX:
		return (mga2x64_copybox(gr->dev.dpy, gr, gr->io_buf));
	   
	case GRAPHICS_DRAWLINE:
		return(mga2x64_drawline(gr->dev.dpy, gr, gr->io_buf));
	   
	case GRAPHICS_DRAWBOX:
		return(mga2x64_drawbox(gr->dev.dpy, gr, gr->io_buf));

	default: 
	  
	        return -E(GRAPHICS,NOSUP_ALWAYS_MMAP); 
	}
}
