/* ---------------------------------------------------------------------------
**	S3 Vision964 acceleration driver
** ---------------------------------------------------------------------------
**
**	Copyright (C) 1995-97, Andreas Beck and Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
** 
** ---------------------------------------------------------------------------
** 
**	$Log: 86c964.c,v $
**	Revision 1.4  1998/03/24 21:32:27  jkneschke
**	changed acording to the new libggi.
**	cliptl -> clip.x1/.y1
**	clipbr -> clip.x2/.y2
**	
**	Revision 1.3  1998/03/11 21:06:39  jkneschke
**	32-bit mode is working now
**	clipping implented.
**	
**	Revision 1.2  1998/02/03 09:38:06  seeger
**	Well, this is the file supposed to be here.
**	
** 
** ---------------------------------------------------------------------------
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Jan_Kneschke
#define	DRIVER_REV	"$Revision: 1.4 $"
#define DRIVER_NAME	"S3 86c964 accel driver"

#define DEBUG_LEVEL	1

#undef	DEBUG_IO

/*\
|*| Kernel includes 
\*/

#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#define	request_io_region	request_region
#define	check_io_region		check_region

/*\
|*| KGI includes
\*/

#include <ggi/module.h>
#include <ggi/system.h>

#include <ggi/ggi_commands.h>
#include <ggi/graphic.h>
#include <ggi/kgi.h>

#include "suggest.inc"

#define	S3_86C964
#define S3_86C968
#include "S3/Vision.h"

#include "S3/accel.h"

#define	IO_VGA_BASE		0x03C0
#define	IO_VGA_SIZE		0x20
#define	IO_ACCEL_BASE		0x02E8
#define	IO_ACCEL_SIZE		4

enum s3_sid_design
{
	SID_32S,
	SID_32P,
	SID_64S,
	SID_64P,
	SID_128S,
	SID_SPEA_MERCURY_P64,

	SID_LAST
};

/*	We keep the driver specific state per chipset handled in a
**	chipset_private struct. 
*/
struct chipset_private
{
	/*	internal stuff
	*/
	mem_vaddr	mem_vga;
	mem_vaddr	mem_accel;	/* memory mapped accel ports	*/

	ggi_uint		frame;
	ggi_uint		split;
	ggi_uint		origin, org_xdot, org_ydot, org_offs;
	ggi_uint		fb_size;

	uint32		ptr_old, cur_old;
	uint16		ptr_and, cur_and;
	uint16		ptr_xor, cur_xor;

	/*	exported stuff
	*/
	struct kgi_mmio_region

	lram,		/* on-board RAM		little endian	*/
	lpdt,		/* pixel data transfer	little endian	*/
	lpfd,		/* pixel formatter data	little endian	*/
	lpfm,		/* pixel formatter data	little endian	*/
	
	bram,		/* on-board RAM		big endian	*/
	bpdt,		/* pixel data transfer	big endian	*/
	bpfd,		/* pixel formatter data	big endian	*/
	bpfm;		/* pixel formatter mask	big endian	*/

	/*	administrative stuff
	*/
	struct io_region	io_vga_region;
	struct io_region	io_accel_region;
	struct mem_region	mem_accel_region;
	struct mem_region	mem_law_region;

	struct kgi_chipset	chipset_info;

	enum s3_sid_design 	sid;
};

#define GET_PRIVATE \
	struct chipset_private *private = KGIM_PRIVATE(dpy, chipset)
	
#define MEM_LAW		(private->mem_law_region.baseptr)
#define MEM_VGA		(private->mem_vga)
#define IO_ACCEL	(private->io_accel_region.baseptr-IO_ACCEL_BASE)
#define MEM_ACCEL	(private->mem_accel)

/*
** internal GC-version counter
*/

static ggi_uint		gc_version;

/*
**	accelerator control
*/

static inline void AC_out16(struct kgi_display *dpy, uint16 val, uint16 reg)
{
	GET_PRIVATE;

	ASSERT((reg & 0xF3FF) == 0x42E8);

	/*
	**	if MEM_VGA is valid, MEM_LAW points to the new MMIO region.
	*/
	if (MEM_VGA) {
		switch (reg) {
		case 0x42e8:
			mem_out16(val, MEM_LAW + 0x1008504);
			return;
		case 0x46e8:
			mem_out16(val, MEM_LAW + 0x1008508);
			return;
		case 0x4ae8:
			mem_out16(val, MEM_LAW + 0x100850c);
			return;
		}
	} else {
		io_out16(val, IO_ACCEL + reg);
	}
}

static inline uint16 AC_in16(struct kgi_display *dpy, uint16 reg)
{
	GET_PRIVATE;

	ASSERT((reg & 0xF3FF) == 0x42E8);

	/*
	**	if MEM_VGA is valid, MEM_LAW will point to a new-MMIO region.
	*/
	if (MEM_VGA) {
		switch (reg) {
		case AC_SUBSYS_STAT:
			return mem_in16(MEM_LAW + 0x1008504);
		case 0x46e8:
			return mem_in16(MEM_LAW + 0x1008508);
		case AC_ADVFUNC_CNTL:
			return mem_in16(MEM_LAW + 0x100850c);
		}
		return 0xFFFF;
	} else {
		return io_in16(IO_ACCEL + reg);
	}
}

/*
**	accelerator setup
*/

static inline void ACCEL_out16(struct kgi_display *dpy, uint16 val, uint16 reg)
{
	GET_PRIVATE;

	if (MEM_ACCEL) {
		mem_out16(val, MEM_ACCEL + reg);
	} else {
		io_out16(val, IO_ACCEL + reg);
	}
}

static inline void ACCEL_out32(struct kgi_display *dpy, uint32 val, uint16 reg)
{
	GET_PRIVATE;

	if (MEM_ACCEL) {
		mem_out32(val, MEM_ACCEL + reg);
	} else {
		io_out32(val, IO_ACCEL + reg);
	}
}


static inline uint16 ACCEL_in16(struct kgi_display *dpy, uint16 reg)
{
	GET_PRIVATE;

	/*	if MEM_VGA is valid, we have a new MMIO region and can
	**	read from the MEM_ACCEL region.
	*/
	if (MEM_VGA) {
		ASSERT(MEM_ACCEL);
		return mem_in16(MEM_ACCEL + reg);
	} else {
		return io_in16(IO_ACCEL + reg);
	}
}


#define NUMPINGPONGPAGES 2

void *pingpong;
enum pingpongstatus 
{ 
	WAITING, RECEIVING, EXECUTING 
};

enum pingpongstatus ppstat[NUMPINGPONGPAGES];

/* Helper functions */
static unsigned short fifo_masks[13] =
{
	0x0080, 0x0040, 0x0020, 0x0010, 
	0x0008, 0x0004, 0x0002, 0x0001,
	0x8000, 0x4000, 0x2000, 0x1000,
	0x0800
};

int s3_wait_fifo(struct kgi_display *dpy, ggi_uint i, int maysched)
{
        register uint16 mask = fifo_masks[(i-1 < 13) ? i-1 : 12];
	int x = 1000000000;

        while ((mask & ACCEL_in16(dpy, ECR_GP_STATUS)) && --x)
		if (maysched) schedule();

	return(!x);
}

static int s3_wait_idle(struct kgi_display *dpy, int maysched)
{ 	
	int x = 1000000000;
	while((ACCEL_in16(dpy, ECR_GP_STATUS) & EC06_GE_BUSY) && --x) 
		if (maysched) schedule();
	return(!x);
}

/* Accelerated drawing functions follow */
static int s3_drawbox(struct kgi_graphic *graphic, void *iobuf, int maysched)
{	
	struct box *boxp = (struct box *)iobuf;
        struct kgi_display *dpy = graphic->dev.dpy;


	if (graphic->dev.mode.graphtype == GT_4BIT || 
	    graphic->dev.mode.graphtype == GT_TEXT16)
	{
		
		return -E(GRAPHICS,NOSUP_MODE_MMAP);
	}

	if (boxp->w && boxp->h) {	/* 0 width is not OK! */
		
		if (graphic->gc->version != gc_version) {
			gc_version = graphic->gc->version;
			
			if (s3_wait_fifo(dpy, 5, maysched))
			{
				return -E(GRAPHICS,NOSUP_NOW_MMAP);
			}
			
			ACCEL_out32(dpy, graphic->gc->fg_color, ECR_FOREGROUND);
			
			/* Clip virtual area  */
			
			ACCEL_out16(dpy, MF_T_CLIP | graphic->gc->clip.y1, ECR_MULTIFUNC);	
			ACCEL_out16(dpy, MF_L_CLIP | graphic->gc->clip.x1, ECR_MULTIFUNC);
			ACCEL_out16(dpy, MF_R_CLIP | (graphic->gc->clip.x2 - 1), ECR_MULTIFUNC);
			ACCEL_out16(dpy, MF_B_CLIP | (graphic->gc->clip.y2 - 1), ECR_MULTIFUNC);
			
		}

		if (s3_wait_fifo(dpy, 2, maysched))
		{
			return -E(GRAPHICS,NOSUP_NOW_MMAP);
		}
		
		ACCEL_out16(dpy, SRC_FOREGROUND | MIX_SRC, ECR_FOREGR_MIX);
		
		ACCEL_out16(dpy, MF_PIX_CONTROL, ECR_MULTIFUNC);

		if (s3_wait_fifo(dpy, 5, maysched))
		{
			return -E(GRAPHICS,NOSUP_NOW_MMAP);
		}
		
		ACCEL_out16(dpy, boxp->x, ECR_CUR_X);
		ACCEL_out16(dpy, boxp->y, ECR_CUR_Y);
		ACCEL_out16(dpy, boxp->w - 1, ECR_MAJAXIS_CNT);
		ACCEL_out16(dpy, (boxp->h - 1) | MF_MINAXIS_CNT, ECR_MULTIFUNC);
		ACCEL_out16(dpy, EC07_FILLRECT | EC07_LITTLE_ENDIAN |  EC07_32BIT_IO |
		     EC07_DIR(5) | EC07_DRAWPIXEL | EC07_ACROSS_PLANE |
	             0x0001, ECR_COMMAND);
		s3_wait_idle(dpy, maysched);
	}

	return(0);
}

static int s3_copybox(struct kgi_graphic *graphic,void *iobuf)
{	
	struct cbox *boxp = (struct cbox *)iobuf;
        struct kgi_display *dpy = graphic->dev.dpy;
	int dir = 0;

	if (graphic->dev.mode.graphtype == GT_4BIT || 
	    graphic->dev.mode.graphtype == GT_TEXT16)
		return -E(GRAPHICS,NOSUP_MODE_MMAP);

	if (boxp->w && boxp->h) {		/* 0 width is not OK! */

		if (s3_wait_fifo(dpy, 2, 1)) 
		{
			return -E(GRAPHICS,NOSUP_NOW_MMAP);
		}

		ACCEL_out16(dpy, 0x67, ECR_FOREGR_MIX);
		ACCEL_out16(dpy, MF_PIX_CONTROL, ECR_MULTIFUNC);

		if (boxp->x > boxp->x2) 
		{
			dir |= EC07_DIR(1);	
		}
		else 
		{
			boxp->x += boxp->w - 1;
			boxp->x2 += boxp->w - 1;
		}

		if (boxp->y > boxp->y2) 
		{
			dir |= EC07_DIR(4);
		}
		else 
		{ 
			boxp->y += boxp->h - 1;
			boxp->y2 += boxp->h - 1;
		}


		
		if (s3_wait_fifo(dpy, 7, 1)) 
		{
			return -E(GRAPHICS,NOSUP_NOW_MMAP);
		}
		
		ACCEL_out16(dpy, boxp->x, ECR_CUR_X);
		ACCEL_out16(dpy, boxp->y, ECR_CUR_Y);
		ACCEL_out16(dpy, boxp->x2, ECR_DIAG_STEP);
		ACCEL_out16(dpy, boxp->y2, ECR_AXIAL_STEP);
		ACCEL_out16(dpy, boxp->w - 1, ECR_MAJAXIS_CNT);
		ACCEL_out16(dpy, (boxp->h - 1) | MF_MINAXIS_CNT,ECR_MULTIFUNC);
		ACCEL_out16(dpy, EC07_BITBLT | EC07_LITTLE_ENDIAN | EC07_32BIT_IO |
		     dir | EC07_DRAWPIXEL | EC07_ACROSS_PLANE |
	             0x0001, ECR_COMMAND);
		s3_wait_idle(dpy, 1);
	}
	
	return(0);
}

static int s3_drawline(struct kgi_graphic *graphic, void *iobuf)
{ 
	int dx, dy, min, max;
	struct line *linp = (struct line *)iobuf;
        struct kgi_display *dpy = graphic->dev.dpy;

	if (graphic->dev.mode.graphtype == GT_4BIT || 
	    graphic->dev.mode.graphtype == GT_TEXT16)
	{
		return -E(GRAPHICS,NOSUP_MODE_MMAP);
	}

	dx = linp->x2 - linp->x;
	dy = linp->y2 - linp->y;

	if (abs(dx) > abs(dy))
	{
		max = abs(dx);
		min = abs(dy);
	}
	else
	{
		max = abs(dy);
		min = abs(dx);
	}

	if (max == 0)
	{
		return(0);	/* FIXME! */
	}
	
	if (graphic->gc->version != gc_version) {
		gc_version = graphic->gc->version;
	
		if (s3_wait_fifo(dpy, 5, 1))
		{
			return -E(GRAPHICS,NOSUP_NOW_MMAP);
		}
	
		ACCEL_out32(dpy, graphic->gc->fg_color, ECR_FOREGROUND);
	
		/* Clip virtual area  */
	
		ACCEL_out16(dpy, MF_T_CLIP | graphic->gc->clip.y1, ECR_MULTIFUNC);	
		ACCEL_out16(dpy, MF_L_CLIP | graphic->gc->clip.x1, ECR_MULTIFUNC);
		ACCEL_out16(dpy, MF_R_CLIP | (graphic->gc->clip.x2 - 1), ECR_MULTIFUNC);
		ACCEL_out16(dpy, MF_B_CLIP | (graphic->gc->clip.y2 - 1), ECR_MULTIFUNC);
	
	}

	if (s3_wait_fifo(dpy, 2, 1))
	{ 
		return -E(GRAPHICS,NOSUP_NOW_MMAP);
	}
	
	ACCEL_out16(dpy, SRC_FOREGROUND | MIX_SRC, ECR_FOREGR_MIX);

	ACCEL_out16(dpy, MF_PIX_CONTROL, ECR_MULTIFUNC);

	if (s3_wait_fifo(dpy, 8, 1))
	{
		return -E(GRAPHICS,NOSUP_NOW_MMAP);
	}
	
	ACCEL_out16(dpy, linp->x, ECR_CUR_X);
	ACCEL_out16(dpy, linp->y, ECR_CUR_Y);

	ACCEL_out16(dpy, (max - 1), ECR_MAJAXIS_CNT);
	ACCEL_out16(dpy, (min - max) << 1, ECR_DIAG_STEP);
	ACCEL_out16(dpy, (min << 1), ECR_AXIAL_STEP);

	ACCEL_out16(dpy, (min + min - max - (dx < 0)), ECR_ERROR_TERM);

	ACCEL_out16(dpy, EC07_LINE
	        | ((dy>0)            ? EC07_DIR(4) : 0)
		| ((dx>0)            ? EC07_DIR(1) : 0)
		| ((abs(dy)>abs(dx)) ? EC07_DIR(2) : 0)
		| EC07_DRAWPIXEL 
		| EC07_ACROSS_PLANE
		| 0x0001 ,ECR_COMMAND);

	s3_wait_idle(dpy, 1);

	return(0);
}

/* KGI module interface */

int kgim_graphics_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE0(DRIVER_NAME" loaded");
	/* !!! Check for Chipset and Version here !!! */

	return EOK;
}

void kgim_graphics_done(struct kgi_display *dpy)
{
	NOTICE1(DRIVER_NAME" removed");
}

int kgim_graphics_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			      enum kgi_tm_cmd cmd)
{
	/* !!! We should probably check for accel-friendly dimensions here */
	return EOK;
}

void kgim_graphics_set_mode (struct kgi_display *dpy, struct kgi_mode *mode)
{

	if (mode->request.graphtype != GT_TEXT16) {

		/* Reset the Accelerator Engine */
		AC_out16(dpy, ((mode->request.graphtype == GT_4BIT) 
		      ? EC01_4BPP : EC01_ENH_ENB) | EC01_LAW_ENB |
		      EC01_MMIO_ENB, ECR_ADVFUNC_CNTL);

		AC_out16(dpy, EC00_RESET, ECR_SUBSYS_CNTL);
		AC_out16(dpy, EC00_ENABLE, ECR_SUBSYS_CNTL);
		
		/* Now set some basic accelerator registers */
		/* This might run in interrupt context, so maysched is 0 */
		s3_wait_fifo(dpy, 7, 0);

		ACCEL_out16(dpy, MF_MULT_MISC | MFE_32BIT_CMD_REGS
			, ECR_MULTIFUNC); /* 32 bit I/O registers */

		ACCEL_out16(dpy, SRC_FOREGROUND | MIX_SRC, ECR_FOREGR_MIX);
		ACCEL_out16(dpy, SRC_BACKGROUND | MIX_SRC, ECR_BACKGR_MIX);/* Direct color*/

		ACCEL_out16(dpy, MF_T_CLIP | 0, ECR_MULTIFUNC);	/* Clip virtual area  */
		ACCEL_out16(dpy, MF_L_CLIP | 0, ECR_MULTIFUNC);
		ACCEL_out16(dpy, MF_R_CLIP | (mode->request.virt.x - 1), ECR_MULTIFUNC);
		ACCEL_out16(dpy, MF_B_CLIP | (mode->request.virt.y - 1), ECR_MULTIFUNC);

		s3_wait_fifo(dpy, 3, 0);
		ACCEL_out32(dpy, ~0, ECR_WRITE_MASK);	/* Enable all planes */
		ACCEL_out32(dpy, ~0, ECR_READ_MASK);	/* Enable all planes */

		ACCEL_out16(dpy, MF_PIX_CONTROL | MFA_SRC_FOREGR_MIX, ECR_MULTIFUNC);

	}
	return;
}

int kgim_graphics_command (struct kgi_graphic *gr, ggi_uint cmd, ggi_uint arg)
{ 
	ggi_suggest *sugg;
	const char *name;

	if (!(gr->dev.state & KGI_MAPPED_ON_DISPLAY) ) {

		return -E(GRAPHICS,NOSUP_MODE_MMAP);
		/* Don't accel inactive screen! */
	}

	switch (cmd) { 
		
	case GRAPHICS_GETSUGGEST:
		sugg = (ggi_suggest *)(gr->io_buf);
		gr->io_res = gr->io_buf;

		if (sugg->id < 0 || sugg->id > 4) 
		{
			
			return -E(GRAPHICS,BOGUS);
		}
		
		name = sugg_default[gr->dev.mode.graphtype].text[sugg->id];
		if (!name)
		{
			
			return -E(GRAPHICS,BOGUS);
		}
		
		strcpy(sugg->name, name);
		sugg->args[0] = 0;
		sugg->id++;	/* Next is RAMDAC */
		if (sugg->id >= 4 || sugg_default[gr->dev.mode.graphtype].text[sugg->id] == NULL) 
		{
			
			sugg->id = 0;
		}
		
		return 0;

	case GRAPHICS_DRAWPIXEL:
#if 0	/* 1=SLOW */
		{  
			struct ggi_coord *h = gr->io_buf;
			unsigned offset = h->x + h->y * current_mode.virt.x;
			int col = 0x12345678;

			switch(current_mode.graphtype) {
			case GT_8BIT:
			 	((char *)fb_graphics)[offset] = (char)col;
				return EOK;
			
			case GT_15BIT:
			case GT_16BIT:	
				((short *)fb_graphics)[offset] = (short)col;
				return EOK;

			case GT_32BIT:
				((int *)fb_graphics)[offset] = (int)col;
				return EOK;
			default:
				printk("DUNNO MYSELF ... ???");
			}
		} 
		break;
#else
		return -E(GRAPHICS,NOSUP_ALWAYS_MMAP);
#endif

		/* Really accelerated functions. */

	case GRAPHICS_DRAWBOX:
		return(s3_drawbox(gr, gr->io_buf, 1));
		
	case GRAPHICS_COPYBOX:
		return(s3_copybox(gr, gr->io_buf));
		
	case GRAPHICS_DRAWLINE:
		return(s3_drawline(gr, gr->io_buf));
		
//	case GRAPHICS_CROSSBLIT:
//		return(s3_strech(gr, gr->io_buf));

		/* Functions that will probably benefit 
		** from lower-lever accel 
		*/
		
	case GRAPHICS_FILLSCREEN:
		return -E(GRAPHICS,NOSUP_ALWAYS_LOWER);

		/* Functions that should probably use MMAP
		 */
	default: 
		return -E(GRAPHICS,NOSUP_ALWAYS_MMAP);

/* These functions cannot be supported via standard interfaces due to
 * pointers to userspace in the args, but are efficiently emulated via MMAP:
 * GRAPHICS_PUTHLINE	GRAPHICS_GETHLINE	GRAPHICS_PUTVLINE
 * GRAPHICS_GETVLINE	GRAPHICS_GETBOX		GRAPHICS_PUTBOX
 */

/* These are not implemented due to efficient support via MMAP.
 * GRAPHICS_PUTPIXEL	GRAPHICS_GETPIXEL	GRAPHICS_DRAWPIXEL (dummy for testing above)
 * GRAPHICS_DRAWCIRCLE	GRAPHICS_DRAWHLINE	GRAPHICS_DRAWVLINE
 */
	}
	return -E(GRAPHICS,NOSUP_ALWAYS_CANT);	/* The driver can't do this */
}
