/* ---------------------------------------------------------------------------
**	MGA 1064,2064,2164 (Mystique, Millennium I & II) register definitions
** ---------------------------------------------------------------------------
**	Copyright (C)	1997		Jonas Borgstrm
**			1997		Hartmut Niemann
**                      1997            Matthias Grimrath
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
**
**	MAINTAINER	Jonas_Borgstroem
**
*/
#ifndef _MGA_COMMON_REGS_H_
#define _MGA_COMMON_REGS_H_

#include <IBM/vga.h>

/*	PCI Configuration space
**
**	Standard registers (64 bytes): please use definitions in linux/pci.h
*/

/* class register: p. 4-3 */
#define PCI_CLASS_REVISION_REV_MASK 0xff
#define PCI_CLASS_REVISION_REV_SHIFT 0
#define PCI_CLASS_REVISION_REV 1           /* current revision of MGA2064W */
#define PCI_CLASS_REVISION_CLASS_MASK 0xffffff00
#define PCI_CLASS_REVISION_CLASS_SHIFT 8
#define PCI_CLASS_REVISION_VGAMODE_MASK 0x8000
#define PCI_CLASS_REVISION_VGA     0x3800000
#define PCI_CLASS_REVISION_NONVGA  0x3000000

#define MGA_PCI_INDEX  0x44
#define MGA_PCI_DATA   0x48

/*	Apertures
**
**	You have to define MGA_BASE1 (base address of control aperture) and
**	MGA_BASE2 (base address of direct frame buffer access aperture)
**	externally.
*/

/*
**	MGA Control Aperture Detailed Map
*/

#define MGA_DMAWIN (MGA_BASE1+0x0000)   /* - 1bff  7k pseudo-DMA window */
#define MGA_DWGREG (MGA_BASE1+0x1C00)   /* - 1dff  drawing registers    */
#define MGA_HSTREG (MGA_BASE1+0x1E00)   /* - 1eff  host registers       */
#define MGA_VGAREG (MGA_BASE1+0x1F00)   /* - 1fff  memory-mapped VGA    */
#define MGA_RAMDAC (MGA_BASE1+0x3C00)   /* - 3c1f  ramdac addresses     */
#define MGA_EXPDEV (MGA_BASE1+0x3E00)   /* - 3fff  for external devices */

/* register map */
#define DWGCTL   (MGA_BASE1+0x1c00)
        #define OPCOD_MASK     0xF
        #define LINE_OPEN      0x0
        #define AUTOLINE_OPEN  0x1
        #define LINE_CLOSE     0x2
        #define AUTOLINE_CLOSE 0x3

        #define TRAP           0x4
        #define TRAP_ILOAD     0x5

        #define BITBLT         0x8
        #define ILOAD          0x9
        #define ILOAD_SCALE    0xD
        #define ILOAD_FILTER   0xF
        #define IDUMP          0xA
        #define ILOAD_HIQH     0x7
        #define ILOAD_HIQHV    0xE

        #define ATYPE_MASK     0x70
        #define ATYPE_RPL      0x00
        #define ATYPE_RSTR     0x10
        #define ATYPE_ZI       0x30
        #define ATYPE_BLK      0x40
        #define ATYPE_I        0x70

        #define LINEAR         0x80

        #define ZMODE_MASK     0x700
        #define ZMODE_NOZCMP   0x000
        #define ZMODE_ZE       0x200
        #define ZMODE_ZNE      0x300
        #define ZMODE_ZLT      0x400
        #define ZMODE_ZLTE     0x500
        #define ZMODE_ZGT      0x600
        #define ZMODE_ZGTE     0x700

        #define SOLID          0x800
        #define ARZERO         0x1000
        #define SGNZERO        0x2000
        #define SHFTZERO       0x4000

        #define BOP_MASK       0xF0000
        /* All possible 2 arg bool ops are possible, below are selected */
        #define BOP_CLEAR      0x00000
        #define BOP_NOR        0x10000
        #define BOP_COPYINV    0x30000
        #define BOP_INVERT     0x50000
        #define BOP_XOR        0x60000
        #define BOP_NAND       0x70000
        #define BOP_AND        0x80000
        #define BOP_EQUIV      0x90000
        #define BOP_NOOP       0xA0000
        #define BOP_IMP        0xB0000
        #define BOP_COPY       0xC0000
        #define BOP_OR         0xE0000
        #define BOP_SET        0xF0000

        #define TRANS_MASK     0xF00000
        /* Look in the doc to find transparency values */

        #define BLTMOD_MASK     0x1E000000
        #define BLTMOD_BMONOLEF 0x00000000
        #define BLTMOD_BMONOWF  0x08000000
        #define BLTMOD_BPLAN    0x02000000
        #define BLTMOD_BFCOL    0x04000000
        #define BLTMOD_BUYUV    0x1C000000
        #define BLTMOD_BU32BGR  0x06000000
        #define BLTMOD_BU32RGB  0x0E000000
        #define BLTMOD_BU24BGR  0x16000000
        #define BLTMOD_BU24RGB  0x1E000000

        #define PATTERN         0x20000000
        #define TRANSC          0x40000000

#define MACCESS  (MGA_BASE1+0x1c04)
#define MCTLWTST (MGA_BASE1+0x1c08) /* marked: reserved */
#define ZORG     (MGA_BASE1+0x1c0c)
#define PAT0     (MGA_BASE1+0x1c10)
#define PAT1     (MGA_BASE1+0x1c14)
#define PLNWT    (MGA_BASE1+0x1c1c)
#define BCOL     (MGA_BASE1+0x1c20)
#define FCOL     (MGA_BASE1+0x1c24)
#define SRCBLT   (MGA_BASE1+0x1c2c) /* marked: reserved */
#define SRC0     (MGA_BASE1+0x1c30)
#define SRC1     (MGA_BASE1+0x1c34)
#define SRC2     (MGA_BASE1+0x1c38)
#define SRC3     (MGA_BASE1+0x1c3c)
#define XYSTRT   (MGA_BASE1+0x1c40)
#define XYEND    (MGA_BASE1+0x1c44)
#define SHIFT    (MGA_BASE1+0x1c50)
        #define FUNCNT_MASK   0x7F
        #define X_OFF_MASK    0x0F
        #define Y_OFF_MASK    0x70
        #define STYLELEN_MASK 0x7F0000
        #define FUNOFF_MASK   0x3F0000

#define SGN      (MGA_BASE1+0x1c58)
        #define SGN_SDYDXL   0x01
        #define SGN_SCANLEFT 0x01
        #define SGN_SDXL     0x02
        #define SGN_SDY      0x04
        #define SGN_SDXR     0x20

#define LEN      (MGA_BASE1+0x1c5c)
#define AR0      (MGA_BASE1+0x1c60)
#define AR1      (MGA_BASE1+0x1c64)
#define AR2      (MGA_BASE1+0x1c68)
#define AR3      (MGA_BASE1+0x1c6c)
#define AR4      (MGA_BASE1+0x1c70)
#define AR5      (MGA_BASE1+0x1c74)
#define AR6      (MGA_BASE1+0x1c78)
#define CXBNDRY  (MGA_BASE1+0x1c80)
#define FXBNDRY  (MGA_BASE1+0x1c84)
#define YDSTLEN  (MGA_BASE1+0x1c88)
#define PITCH    (MGA_BASE1+0x1c8c)
        #define PITCH_YLIN 0x00008000
#define YDST     (MGA_BASE1+0x1c90)
#define YDSTORG  (MGA_BASE1+0x1c94)
#define YTOP     (MGA_BASE1+0x1c98)
#define YBOT     (MGA_BASE1+0x1c9c)
#define CXLEFT   (MGA_BASE1+0x1ca0)
#define CXRIGHT  (MGA_BASE1+0x1ca4)
#define FXLEFT   (MGA_BASE1+0x1ca8)
#define FXRIGHT  (MGA_BASE1+0x1cac)
#define XDST     (MGA_BASE1+0x1cb0)
#define DR0      (MGA_BASE1+0x1cc0)
#define DR1      (MGA_BASE1+0x1cc4) /* marked: reserved */
#define DR2      (MGA_BASE1+0x1cc8)
#define DR3      (MGA_BASE1+0x1ccc)
#define DR4      (MGA_BASE1+0x1cd0)
#define DR5      (MGA_BASE1+0x1cd4) /* marked: reserved */
#define DR6      (MGA_BASE1+0x1cd8)
#define DR7      (MGA_BASE1+0x1cdc)
#define DR8      (MGA_BASE1+0x1ce0)
#define DR9      (MGA_BASE1+0x1ce4) /* marked: reserved */
#define DR10     (MGA_BASE1+0x1ce8)
#define DR11     (MGA_BASE1+0x1cec)
#define DR12     (MGA_BASE1+0x1cf0)
#define DR13     (MGA_BASE1+0x1cf4) /* marked: reserved */
#define DR14     (MGA_BASE1+0x1cf8)
#define DR15     (MGA_BASE1+0x1cfc)
/* 1d00-1dff : as 1c00 - 1cff, but including "start drawing" command. */
#define ACCEL_GO 0x100 /* To be or'ed to the drawing registers */

/* register map to be used within Pseudo-DMA */
#define DMA_DWGCTL   (0x00>>2)
#define DMA_MACCESS  (0x04>>2)
#define DMA_MCTLWTST (0x08>>2) /* marked: reserved */
#define DMA_ZORG     (0x0c>>2)
#define DMA_PAT0     (0x10>>2)
#define DMA_PAT1     (0x14>>2)
#define DMA_PLNWT    (0x1c>>2)
#define DMA_BCOL     (0x20>>2)
#define DMA_FCOL     (0x24>>2)
#define DMA_SRCBLT   (0x2c>>2) /* marked: reserved */
#define DMA_SRC0     (0x30>>2)
#define DMA_SRC1     (0x34>>2)
#define DMA_SRC2     (0x38>>2)
#define DMA_SRC3     (0x3c>>2)
#define DMA_XYSTRT   (0x40>>2)
#define DMA_XYEND    (0x44>>2)
#define DMA_SHIFT    (0x50>>2)
#define DMA_PAD      (0x54>>2)
#define DMA_SGN      (0x58>>2)
#define DMA_LEN      (0x5c>>2)
#define DMA_AR0      (0x60>>2)
#define DMA_AR1      (0x64>>2)
#define DMA_AR2      (0x68>>2)
#define DMA_AR3      (0x6c>>2)
#define DMA_AR4      (0x70>>2)
#define DMA_AR5      (0x74>>2)
#define DMA_AR6      (0x78>>2)
#define DMA_CXBNDRY  (0x80>>2)
#define DMA_FXBNDRY  (0x84>>2)
#define DMA_YDSTLEN  (0x88>>2)
#define DMA_PITCH    (0x8c>>2)
#define DMA_YDST     (0x90>>2)
#define DMA_YDSTORG  (0x94>>2)
#define DMA_YTOP     (0x98>>2)
#define DMA_YBOT     (0x9c>>2)
#define DMA_CXLEFT   (0xa0>>2)
#define DMA_CXRIGHT  (0xa4>>2)
#define DMA_FXLEFT   (0xa8>>2)
#define DMA_FXRIGHT  (0xac>>2)
#define DMA_XDST     (0xb0>>2)
#define DMA_DR0      (0xc0>>2)
#define DMA_DR1      (0xc4>>2) /* marked: reserved */
#define DMA_DR2      (0xc8>>2)
#define DMA_DR3      (0xcc>>2)
#define DMA_DR4      (0xd0>>2)
#define DMA_DR5      (0xd4>>2) /* marked: reserved */
#define DMA_DR6      (0xd8>>2)
#define DMA_DR7      (0xdc>>2)
#define DMA_DR8      (0xe0>>2)
#define DMA_DR9      (0xe4>>2) /* marked: reserved */
#define DMA_DR10     (0xe8>>2)
#define DMA_DR11     (0xec>>2)
#define DMA_DR12     (0xf0>>2)
#define DMA_DR13     (0xf4>>2) /* marked: reserved */
#define DMA_DR14     (0xf8>>2)
#define DMA_DR15     (0xfc>>2)

#define DMA_GO       0x40

/* host registers */
#define FIFOSTATUS  (MGA_BASE1+0x1e10)
        #define FIFO_COUNT_MASK 0x3F
        #define FIFO_BFULL      0x100
        #define FIFO_BEMPTY     0x200

#define STATUS      (MGA_BASE1+0x1e14)
        #define STAT_PICKPEN   0x4
        #define STAT_VSYNCSTS  0x8
        #define STAT_VSYNCPEN  0x10
        #define STAT_VLINEPEN  0x20
        #define STAT_EXTPEN    0x40
        #define STAT_DWGENGSTS 0x10000

#define ICLEAR      (MGA_BASE1+0x1e18)
        #define ICLEAR_PICK  0x04
        #define ICLEAR_VLINE 0x20

#define IEN         (MGA_BASE1+0x1e1c)
        #define IEN_SOFTRAP 0x01
        #define IEN_PICK    0x04
        #define IEN_VLINE   0x20
        #define IEN_EXT     0x40

#define VCOUNT      (MGA_BASE1+0x1e20)
#define RESET       (MGA_BASE1+0x1e40)
#define OPMODE      (MGA_BASE1+0x1e54)


/* VGA registers in memory */
#define MMIO_MISCr   (MGA_BASE1+0x1fcc)  /* R  Misc output register */
#define MMIO_MISCw   (MGA_BASE1+0x1fc2)  /*  W        "             */
#define MMIO_FCTRLr  (MGA_BASE1+0x1fca)  /* R  Feature control reg. */

/*
#define MMIO_FCTRLwm MMIO_FCTRLwc
#define MMIO_FCTRLwc (MGA_BASE1+0x1fda)
*/
#define MMIO_FCTRLw  (MGA_BASE1+0x1fda)  /*  W Feature control reg. */
#define MMIO_INS0r   (MGA_BASE1+0x1fc2)  /* R  Input status reg. 0  */
/*
#define MMIO_INS1rm  MMIO_INS1rc      
#define MMIO_INS1rc  (MGA_BASE1+0x1fda)
*/
#define MMIO_INS1r   (MGA_BASE1+0x1fda)  /* R  Input Status 1       */
#define MMIO_SEQI    (MGA_BASE1+0x1fc4)  /* RW Sequencer Index      */
#define MMIO_SEQD    (MGA_BASE1+0x1fc5)  /* RW Sequencer Data       */
/*
#define MMIO_CRTIm   (MGA_BASE1+0x1fd4)  
#define MMIO_CRTIc   (MGA_BASE1+0x1fd4)
*/
#define MMIO_CRTI    (MGA_BASE1+0x1fd4)  /* RW CRTC Index           */
/*
#define MMIO_CRTDm   (MGA_BASE1+0x1fd5)
#define MMIO_CRTDc   (MGA_BASE1+0x1fd5)
*/
#define MMIO_CRTD    (MGA_BASE1+0x1fd5)  /* RW CRTC Data            */
#define MMIO_GRCI    (MGA_BASE1+0x1fce)  /* RW Graphics Controller Index*/
#define MMIO_GRCD    (MGA_BASE1+0x1fcf)  /* RW Graphics Controller Data */
#define MMIO_ATCI    (MGA_BASE1+0x1fc0)  /* RW Attribute register index */
#define MMIO_ATCDr   (MGA_BASE1+0x1fc1)  /* R  Attribute register data  */
#define MMIO_ATCDw   (MGA_BASE1+0x1fc0)  /*  W     "        "       "   */

/* 
** DAC_BASE, _MASK, xx_ADDR, _DATA do not exist in mmio. 
** Use RAMDAC range instead? 
*/

#define MMIO_DAC_STATUS (MGA_BASE1+0x1fc7)/*R  RAMDAC state             */  

#define MMIO_CRTXI   (MGA_BASE1+0x1fde)  /* RW CRT extension register index  */
#define MMIO_CRTXD   (MGA_BASE1+0x1fdf)  /* RW CRT extension register (data) */

#define CRTX_EXT0 0x0
        #define CRTX_STARTADD_MASK 0x0F
        #define CRTX_OFFSET_MASK   0x30
        #define CRTX_INTERLACE     0x80
#define CRTX_EXT1 0x1
        #define CRTX_HTOTAL8   0x01
        #define CRTX_HBLKSTR8  0x02
        #define CRTX_HSYNCSTR8 0x04
        #define CRTX_HRSTEN    0x08
        #define CRTX_HSYNCOFF  0x10
        #define CRTX_VSYNCOFF  0x20
        #define CRTX_HBLKEND6  0x40
        #define CRTX_VRSTEN    0x80
#define CRTX_EXT3 0x3
        #define CRTX_SCALE_MASK 0x7
        #define CRTX_SLOW256 0x20
        #define CRTX_CSYNCEN 0x40
        #define CRTX_MGAMODE 0x80

/* non-vga io space registers */
#define CRTXI  0x03de                   /* RW CRT extension register index  */
#define CRTXD  0x03df                   /* RW CRT extension register (data) */

#define	INS1_HRETRACE   0x01
#define	INS1_LIGHTPEN_T	0x02
#define	INS1_LIGHTPEN_S	0x04
#define INS1_VRETRACE	0x08
#define	INS1_DIAGNOSTIC	0x30

/* Clock select in VGA MISC register */
#define MISC_CLOCK_1X 0x08

#define MGA_PCI_OPTION 0x40
        #define MGA_PCI_OPTION_SYSCLKSEL_MASK   0x00000003
        #define MGA_PCI_OPTION_SYSCLKSEL_PCI    0x00000000
        #define MGA_PCI_OPTION_SYSCLKSEL_SYSTEM 0x00000001
        #define MGA_PCI_OPTION_SYSCLKSEL_EXT    0x00000002
        #define MGA_PCI_OPTION_SYSCLKDIS        0x00000004
        #define MGA_PCI_OPTION_GCLKDIV          0x00000008
        #define MGA_PCI_OPTION_MCLKDIV          0x00000010
        #define MGA_PCI_OPTION_SYSPLLPDN        0x00000020
        #define MGA_PCI_OPTION_VGAIOEN          0x00000100
        #define MGA_PCI_OPTION_FBMASKN_MASK     0x00000E00
        #define MGA_PCI_OPTION_MEMCONFIG        0x00001000
        #define MGA_PCI_OPTION_SPLITMODE        0x00002000
        #define MGA_PCI_OPTION_HARDPWMSK        0x00004000
        #define MGA_PCI_OPTION_UNIMEM           0x00008000
        #define MGA_PCI_OPTION_RFHCNT           0x000F0000
        #define MGA_PCI_OPTION_EEPROMWT         0x00100000
        #define MGA_PCI_OPTION_PRODUCTID_MASK   0x1F000000
        #define MGA_PCI_OPTION_NORETRY          0x20000000
        #define MGA_PCI_OPTION_BIOSEN           0x40000000
        #define MGA_PCI_OPTION_POWERPC          0x80000000

/* Card specific includes */
#if defined(DRIVER_mga1064)
#include "Matrox/mga1064_regs.h"
#elif defined(DRIVER_mga2064) || defined(DRIVER_mga2164)
#else
#warning !!! include the headerfile for the new Matrox card here
#endif

#endif	/* #ifdef _MGA_COMMON_REGS_H_ */
