#ifndef	_ps2kbd_h
#define	_ps2kbd_h

/*
**	controller ports
*/
#define	AUX_IN			0x60	/* input (from dev) buffer	*/
#define	AUX_OUT			0x60	/* output (to dev) buffer	*/
#define	AUX_CMD			0x64	/* command buffer		*/
#define	AUX_STATUS		0x64	/* status reg			*/

/*
**	controller status bits
*/
#define AUX_STATUS_IBF		0x01	/* inbuffer (from device) full	*/
#define AUX_STATUS_OBF		0x02	/* outbuffer (to device) full	*/
#define	AUX_STATUS_OBT		0x20	/* output (to device) timeout	*/
#define	AUX_STATUS_IBT		0x40	/* input (from device) timeout	*/
#define	AUX_STATUS_PERR		0x80	/* transmisson error		*/

/*
**	mode register
*/
#define	AUX_MODE_KBD_INTS_ON	0x01	/* enable keyboard interrupts	*/
#define	AUX_MODE_PTR_INTS_ON	0x02	/* enable pointer interrupts	*/
#define	AUX_MODE_SYSFLAG	0x04	/* system flag of status reg	*/
#define	AUX_MODE_IGNORE_SWITCH	0x08	/* ignore keyboard switch	*/
#define	AUX_MODE_KBD_OFF	0x10	/* keyboard interface off	*/
#define	AUX_MODE_PTR_OFF	0x20	/* pointer interface off	*/
#define	AUX_MODE_AT_SCANCODES	0x40	/* IBM-Scancodes		*/

/*
**	psaux controller commands
*/
#define	AUX_CMD_READ_MODE	0x20	/* read command register	*/
#define	AUX_CMD_WRITE_MODE	0x60	/* write command register	*/
#define	AUX_CMD_SELF_TEST	0xAA	/* controller self test		*/

#define	AUX_CMD_KBD_DISABLE	0xAD	/* switch keyboard interf. off	*/
#define	AUX_CMD_KBD_ENABLE	0xAE	/* switch keyboard interf. on	*/
#define	AUX_CMD_KBD_TEST	0xAB	/* test keyboard interface	*/

#define	AUX_CMD_PTR_DISABLE	0xA7	/* switch auxiliary off		*/
#define	AUX_CMD_PTR_ENABLE	0xA8	/* switch auxiliary on		*/
#define	AUX_CMD_PTR_TEST	0xA9	/* test auxiliary device	*/
#define	AUX_CMD_PTR_INTS_ON	0x47	/* enable auxiliary interrupts	*/
#define	AUX_CMD_PTR_INTS_OFF	0x65	/* disable auxiliary interrupts	*/
#define	AUX_CMD_PTR_WRITE	0xD4	/* send data to aux. device	*/

/*
**	common device commands
*/
#define	DEV_CMD_RESET		0xFF	/* reset device			*/
#define	DEV_CMD_ENABLE		0xF4	/* enable device		*/
#define	DEV_CMD_DISABLE		0xF5	/* disable device		*/

/*
**	common device replies
*/
#define	DEV_ACKNOWLEDGE		0xFA	/* acknowledge			*/
#define	DEV_RESEND		0xFE	/* resend last byte		*/

/*
**	pointer device commands
*/
#define	PTR_CMD_SET_RES		0xe8	/* set resolution		*/
#define	PTR_CMD_SET_SCALE11	0xe6	/* set 1:1 scaling		*/
#define	PTR_CMD_SET_SCALE21	0xe7	/* set 2:1 scaling		*/
#define PTR_CMD_GET_SCALE	0xe9	/* get scaling factor		*/
#define PTR_CMD_SET_STREAM	0xea	/* set stream mode		*/
#define PTR_CMD_SET_SAMPLE	0xf3	/* set sample rate		*/

/*
**	keyboard device commands
*/
#define	KBD_CMD_SET_LED		0xED	/* set mode indicators		*/
#	define	KBD_LED_SCROLL	0x01	/* scroll-lock			*/
#	define	KBD_LED_NUM	0x02	/* num-lock			*/
#	define	KBD_LED_CAPS	0x04	/* caps-lock			*/
#define	KBD_CMD_ECHO		0xEE	/* ping to keyboard		*/
#define	KBD_CMD_SCANCODE	0xF0	/* get/set scancode set		*/
#	define	KBD_SC_GET	0x00	/* get current scancode set	*/
#	define	KBD_SC_1	0x01	/* set scancode set 1 (PC/XT)	*/
#	define	KBD_SC_2	0x02	/* set scancode set 2		*/
#	define	KBD_SC_3	0x03	/* set scancode set 3		*/
#define	KBD_CMD_ID		0xF2	/* read device ID		*/
#define	KBD_CMD_TYPEMATICRATE	0xF3	/* set typematic rate		*/
#define	KBD_CMD_RESEND		0xFE	/* resend last byte		*/

/*
**	keyboard device replies
*/
#define	KBD_OVERRUN1		0x00	/* scancode buf overrun (set1)	*/
#define	KBD_TEST_COMPLETE	0xAA	/* self test passed		*/
#define	KBD_ECHO		0xEE	/* diagnostic echo		*/
#define	KBD_BREAK_CODE		0xF0	/* break code prefix		*/
#define	KBD_TEST_ERROR		0xFC	/* self test failed		*/
#define	KBD_RESEND		0xFE	/* transmission error, resend	*/
#define	KBD_OVERRUN2		0xFF	/* scancode buf overrun (set2)	*/
#define	KBD_OVERRUN3		0xFF	/* scancode buf overrun (set3)	*/

#endif
