/* ----------------------------------------------------------------------------
**	CL-GD5452x ramdac driver
** ----------------------------------------------------------------------------
**
**	Copyright (C) 1995,1996 by Steffen Seeger
**		1997 by Jason McMullan
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ----------------------------------------------------------------------------
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Martin_Erhardsen
#define DRIVER_NAME	"Cirrus Logic 542x DAC driver"
#define DRIVER_REV	"$Revision: 1.9 $"

#ifdef DEBUG_LEVEL
#define	NEED_GRAPHTYPE_SYMS
#endif

#include <ggi/kgi.h>
#include <ggi/module.h>
#include <ggi/ggi_commands.h>


static void dac_set_clut(struct kgi_display *dpy, ggi_clut *clut,
			ggi_uint first, ggi_uint cnt)
{
	ushort *data = (ushort *) (clut->data + first);
	ASSERT(cnt);
	cnt *= 3;
	outb(first, 0x3C8);

	while (cnt--) outb(*(data++) >> 10, 0x3C9);
}

static void dac_get_clut(struct kgi_display *dpy,
			 ggi_clut *clut, ggi_uint first, int cnt)
{
	ushort *data = (ushort *) (clut->data + first);
	cnt *= 3;
	outb(first, 0x3C7);

	while (cnt--) *(data++) = inb(0x3C9)*1040;
}

/* This structure is initialized by the chipset driver */

struct kgi_ramdac 
ramdac = {	"Cirrus",	       		/* manufacturer		*/
	 	"542x",				/* model		*/
 		GM_ALL,				/* supported modes	*/
	 	0,			 	/* flags		*/
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/

		 DAC_16BIT_PB, 			/* device type		*/
		  {       0,  85226000 },	/* normal mode		*/
		  {       0,         0 },	/* 2xclk mode		*/

		 {{	  0,  32000000 },	/* 	GT_TEXT16	*/
		  {	  0,         0 },	/* 	GT_TEXT32 	*/
		  {       0,  85226000 },	/* 	GT_1BIT		*/
		  {	  0,  85226000 },	/* 	GT_4BIT		*/
		  {       0,  85226000 },	/* 	GT_8BIT		*/
		  {	  0,  45076000 },	/* 	GT_15BIT	*/
		  { 	  0,  45076000 },	/* 	GT_16BIT	*/
		  {       0,  30000000 },	/* 	GT_24BIT	*/
		  {	  0,  22538000 }}	/* 	GT_32BIT	*/
         };

/*\
|*|	GGI interface
\*/
int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev. " DRIVER_REV);
	card->ramdac=&ramdac;
	return EOK;
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " removed.");
}

int
kgim_ramdac_check_mode (struct kgi_display *dpy,
                        struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int newclock = mode->dclk;
  
  if (!((1<<mode->request.graphtype)&ramdac.modes))
   {
  	DEBUG2("couldn't handle graphtype %i",mode->request.graphtype);
 	return -E(RAMDAC,MODE_CHECK);
   }

  switch (cmd)
   {
     case CMD_PROPOSE:
	mode->lclk.mul = mode->clk.mul = 1;
        mode->lclk.div = mode->clk.div = 1;
	return EOK;

     case CMD_LOWER:
	if (   (newclock < ramdac.clock[mode->request.graphtype].min)
	    || (newclock < ramdac.normal.min))
	  {
	    DEBUG2("DCLK too low (%i Hz).", newclock);
	    return -E(RAMDAC,CLK_LOW_LIMIT);
	  }

	if (ramdac.clock[mode->request.graphtype].max < newclock)
    	  newclock = ramdac.clock[mode->request.graphtype].max;

	mode->dclk   = newclock;
	return EOK;

     case CMD_RAISE:
	if (   (newclock > ramdac.clock[mode->request.graphtype].max)
            || (newclock > ramdac.normal.max))
	  {
	    DEBUG2("DCLK too high (%i Hz).", newclock);
	    return -E(RAMDAC,CLK_HIGH_LIMIT);
	  }

	if (newclock < ramdac.clock[mode->request.graphtype].min)
	  newclock = ramdac.clock[mode->request.graphtype].min;

	if (newclock < ramdac.normal.min)
	  newclock = ramdac.normal.min;

	mode->dclk = newclock;
	return EOK;

     case CMD_CHECK:
        if ((mode->lclk.mul != 1) || (mode->lclk.div != 1)
	    /* || (mode->clk.mul  != 1) */ || (mode->clk.div  != 1))
	  return -E(RAMDAC,MODE_CHECK);

	if (   (mode->dclk < ramdac.normal.min)
	    || (mode->dclk < ramdac.clock[mode->request.graphtype].min))
	  return -E(RAMDAC,CLK_LOW_LIMIT);

	if (   (mode->dclk > ramdac.normal.max)
	    || (mode->dclk > ramdac.clock[mode->request.graphtype].max))
	  return -E(RAMDAC,CLK_HIGH_LIMIT);

	return EOK;

    default:
	INTERNALERROR;
	return -E(RAMDAC,BOGUS);
  }
}

void kgim_ramdac_set_mode(struct kgi_display *dpy,struct kgi_mode *mode)
{
	outb(0xFF, 0x3C6);
        dpy->set_clut = dac_set_clut;
}

int kgim_ramdac_command(struct kgi_graphic *graph, ggi_uint cmd, ggi_uint arg)
{
	/*	this may set 
	**	-  dac_params (take care if 504 or 505 was detected).
	**	-  hardware cursor
	**	-  CLUT (or gamma correction table).
	*/
	ggi_clut clut;

	switch(cmd) { 
		case RAMDAC_GETINFO:
			graph->io_res=&ramdac;
			return EOK;
		case RAMDAC_SETINFO:
			if (suser()) {
				memcpy(&ramdac,graph->io_buf,sizeof(ramdac));
				return EOK;
			} else  return -EPERM;
		case RAMDAC_SETCLUT:
			{  clut.size=256;
			   clut.data=(ggi_color *)(graph->io_buf);
			   dac_set_clut(graph->dev.dpy,&clut,0,256);
			} 
			return EOK;
		case RAMDAC_GETCLUT:
			{  clut.size=256;
			   clut.data=(ggi_color *)(graph->io_buf);
			   graph->io_res=graph->io_buf;
			   dac_get_clut(graph->dev.dpy,&clut,0,256);
			} 
			return EOK;
	}
  	return -E(RAMDAC,NOSUP_ALWAYS_CANT);  /* This driver can't do this. */
}






