/*\
|*| ----------------------------------------------------------------------------
|*|     GGI general definitons, data structures, etc.
|*| ----------------------------------------------------------------------------
|*|
|*|	Copyright (C)	by Andreas Beck [becka@hp.rz.uni-duesseldorf.de]
|*|	1995, 1996	and Steffen Seeger [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ----------------------------------------------------------------------------
|*|	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]
\*/

#ifndef	_GGI_TYPES_H
#define	_GGI_TYPES_H
#define	_GGI_H_VERSION	3

#include <ggi/debug.h>
#include <ggi/version.h>
#define	GGI_VERSION	(GGI_VERSION_MAJOR*0x10000 + GGI_VERSION_MINOR*0x100 +\
			 GGI_VERSION_PATCH)

/*\
|*| ----------------------------------------------------------------------------
|*|	basic data types
|*| ----------------------------------------------------------------------------
|*|
|*|	naming convention: (unsigned ? 'u' : ' ') ## <base type>
|*|	where <base type> is one of the following
|*|
|*|	FIXME: Use sized types for this !
|*|	int		depends on architecture (at least 32 bit)
|*|	char		8 bit integer
|*|	short		16 bit integer
|*|	long		32 bit integer
\*/

#ifdef __KERNEL__
#include <linux/types.h>
#else
#include <sys/types.h>
#endif

#if !defined (__osf__) && !defined (_AIX)
typedef	unsigned char		uchar;
#endif

#define	BITS_PER(x)	(sizeof(x)*8)

#include <ggi/system.h>

#include <ggi/errno.h>

/*\
|*|	generic data types
\*/
#define        GGI_YES         1
#define        GGI_NO          0
#define        GGI_ON          1
#define        GGI_OFF         0
#define        GGI_TRUE        1
#define        GGI_FALSE       0

#define        GGI_AUTO     (-1)	/* Can be given to use defaults. */
/* the current GRAPHTYPE definition depends on GGI_AUTO == -1!!! */

typedef int	ggi_bool;	/* This was "char", but int is faster and better for user->kernel */

typedef struct { sint32 min, max; }				ggi_range;
typedef struct { sint32 mul, div; }				ggi_ratio;
typedef struct { sint16   x, y; }				ggi_coord;
typedef	struct { uint16   x, y; }				ggi_size;
typedef struct { int	x1,y1,x2,y2; }				ggi_rect;
typedef ggi_rect 						ggi_line;
typedef	uint32							ggi_pixel;
typedef	uint32							ggi_attr;
	#define	ATTR_FGCOLOR	0x0000FF00	/* fgcolor clut index	*/
	#define	ATTR_BGCOLOR	0x000000FF	/* bgcolor clut index	*/
	#define	ATTR_NORMAL	0x00000000	/* normal style		*/
	#define	ATTR_HALF	0x00010000	/* half intensity	*/
	#define	ATTR_BRIGHT	0x00020000	/* high intensity	*/
	#define	ATTR_INTENSITY	0x00030000	/* mask to get intensity*/
	#define	ATTR_UNDERLINE	0x00040000	/* underline attribute	*/
	#define	ATTR_BOLD	0x00080000	/* bold style		*/
	#define	ATTR_ITALIC	0x00100000	/* italic style		*/
	#define	ATTR_REVERSE	0x00200000	/* reverse fg/bg	*/
	#define	ATTR_BLINK	0x00800000	/* enable blinking	*/
	#define	ATTR_COLOR(fg, bg)	((bg & 0xFF) | ((fg & 0xFF) << 8))

typedef	uint8							ggi_alpha;
	#define	GGI_MAX_ALPHA	255
typedef	uint16							ggi_zvalue;
typedef struct { uint16 r,g,b; }				ggi_color;
typedef struct { uint16	size; ggi_color	*data; }		ggi_clut;
#define	GGI_COLOR_PRECISION 16	/* 16 bit per R,G, B value	*/

/* FIXME ! Don't use shorts here ! Needs conversion ! See C argument expansion rules ! */
typedef struct { short x,y; unsigned short a,b; short a1,a2; }	ggi_arc_data;
typedef struct { short x,y; }					ggi_point_data;
typedef struct { short x1,y1; short x2,y2; }			ggi_line_data;
typedef struct { short x,y; unsigned short w,h; }		ggi_rect_data;

typedef enum
{	GT_TEXT16,		/* 16 bpp alphanumeric modes	*/
	GT_TEXT32,		/* 32 bpp alphanumeric modes	*/
	GT_1BIT,		/*  1 bpp graphics		*/
	GT_4BIT,		/*  4 bpp graphics		*/
	GT_8BIT,		/*  8 bpp graphics		*/
	GT_15BIT,		/* 15 bpp graphics 		*/
	GT_16BIT,		/* 16 bpp graphics		*/
	GT_24BIT,		/* 24 bpp graphics		*/
	GT_32BIT,		/* 24 bpp word aligned  	*/
	GT_LAST			/* last mode assigned   	*/
} ggi_graphtype;

#define GGI_IS_TEXTMODE(mode) ((mode)->graphtype<=GT_TEXT32)

#define MASK(X) GM_##X = (1 << GT_##X) 	
typedef enum
{	MASK(TEXT16),
	MASK(TEXT32),
	MASK(1BIT),
	MASK(4BIT),
	MASK(8BIT),
	MASK(15BIT),
	MASK(16BIT),
	MASK(24BIT),
	MASK(32BIT),
	GM_ALL	=  ((1 << GT_LAST) - 1)
} ggi_graphmask;
#undef MASK

#ifdef NEED_GRAPHTYPE_SYMS
  static char *ggi_graphmask_sym[GT_LAST+1] = 
  {
	"GT_AUTO",
	"GT_TEXT16",	"GT_TEXT32",	"GT_1BIT",	"GT_4BIT",
	"GT_8BIT",	"GT_15BIT",	"GT_16BIT",	"GT_24BIT",
	"GT_32BIT"
  };
  #define GRAPHTYPE(X) (ggi_graphmask_sym[ X+1 ])
  /* This depends on GGI_AUTO == -1 !!!! (hniemann) */
#endif


typedef struct		/* requested by user and changed by driver    */
{
	uint32		frames;		/* frames needed			*/
	ggi_coord	visible;	/* vis. pixels, may change slightly	*/
	ggi_coord	virt;		/* virtual pixels, may change		*/
	ggi_coord	size;		/* size of visible in mm		*/
	ggi_graphtype	graphtype;	/* which mode ?				*/
	ggi_coord	dpp;		/* dots per pixel			*/
} ggi_mode;

/*\
|*|	drawables and displays
\*/

#include <ggi/font.h>
#include <ggi/events.h>

typedef enum
{	MAP_NOTHING,			/* ignore all op's (must be 0)	*/
	MAP_HARDWARE,			/* use hardware accel only	*/
	MAP_HARD_MEM,			/* mmap'ed hardware or accel	*/
	MAP_MEMORY,			/* use only a memory buffer	*/
	MAP_DRIVER			/* use ioctl interface		*/
} ggi_mapping;

typedef enum
{	MAP_256_BPP,
	MAP_128_BPP,
	MAP_64_BPP,
	MAP_32_BPP,
	MAP_16_BPP,
	MAP_8_BPP,
	MAP_4_BPP,
	MAP_2_BPP,
	MAP_1_BPP
} ggi_maptype;

#ifndef __KERNEL__

struct ggi_gc;

#define	display	ggi_display
#define	pixmap	ggi_pixmap
#define	pixel	ggi_pixel
#define	attr	ggi_attr
#define	fb_ops	ggi_fb_ops
#define	gc	ggi_gc
#define	FBARGD	pixmap *fb, struct gc *c
#define	FBARGC	fb, c

#include <ggi/pixmap.h>

#undef	display
#undef	pixmap
#undef	pixel
#undef	attr
#undef	fb_ops
#undef	gc
#undef	FBARGD
#undef	FBARGC

typedef struct ggi_display
{	int		fd;		/* device file handle		*/
	ggi_mode	mode;		/* video hardware mode		*/
	ggi_clut	lut[8];		/* (color) look up table	*/
	ggi_pixmap	virt[8];	/* virtual screen		*/
	ggi_pixmap	visible[8];	/* visible screen		*/

	ggi_mapping	mapping;	/* current mapping state	*/
/* !!!	ggi_context_ops	gc_ops;		context setting operations	*/
	ggi_fb_ops	fb_ops;		/* drawing operations		*/
} ggi_display;

#endif

/*\
|*|	drawing context
\*/

typedef struct ggi_gc
{
	ggi_uint	version;
	ggi_pixel	fg_color;	/* foreground pixel value	*/
	ggi_pixel	bg_color;	/* background pixel value	*/
	ggi_coord	cliptl,	clipbr;	/* obsolete ! */
/*	ggi_rect	clip;		 clipping rectangle (for degas)	*/
        ggi_pixel	plane_mask;	/* plane mask bits		*/
} ggi_gc;

#endif	/* #ifdef _GGI_GGI_H */
