/* 
   Graphics library for GGI. Horizontal lines.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <string.h>

#include "ggi-dl.h"

/**********************************/
/* draw/get/put a horizontal line */
/**********************************/

int GGIdrawhline(ggi_visual *vis,int x,int y,int w)
{
	unsigned char *adr;
	int i,j,mask,color;

	/* Clipping */
	if (y<(VIS_GC(vis)->cliptl.y) || y>=(VIS_GC(vis)->clipbr.y)) return 0;
	if (x< (VIS_GC(vis)->cliptl.x)) {
		int diff=(VIS_GC(vis)->cliptl.x)-x;
		x+=diff;
		w-=diff;
	}
	if (x+w>(VIS_GC(vis)->clipbr.x)) {
		w=(VIS_GC(vis)->clipbr.x)-x;
	}
	if (w<=0) return 0;

	color=(VIS_GC_FGCOLOR(vis)&1)*0xFF;

	adr=((unsigned char *)(VIS_FB_LINEAR(vis)));
	adr+=(x+y*VIS_FB_WIDTH(vis))>>3;

	/* Draw 'front' pixels */ 
	j=w;
	if ((i=x&7)) {
		if ((j=j+i-8)<=0) {
			mask=(0xff>>i)&(0xff<<-j); 
			*adr   = (*adr & ~mask) | (color & mask);
			return EOK;
		} else {
			mask=(0xff>>i);
			*adr++ = (*adr & ~mask) | (color & mask);
		}
	}
	
	while ((j-=8)>=0) 
		*(adr++) = color;
    
	/* Draw `back` pixels if necessary */
	mask=~(0xff>>(j&7));
	*adr = (*adr & ~mask) | (color & mask);

	return EOK;
}

int _GGIdrawhline(ggi_visual *vis,int x,int y,int w)
{
	unsigned char *adr;
	int i,j,mask,color;

	color=(VIS_GC_FGCOLOR(vis)&1)*0xFF;

	adr=((unsigned char *)(VIS_FB_LINEAR(vis)));
	adr+=(x+y*VIS_FB_WIDTH(vis))>>3;

	/* Draw 'front' pixels */ 
	j=w;
	if ((i=x&7)) {
		if ((j=j+i-8)<=0) {
			mask=(0xff>>i)&(0xff<<-j); 
			*adr   = (*adr & ~mask) | (color & mask);
			return EOK;
		} else {
			mask=(0xff>>i);
			*adr++ = (*adr & ~mask) | (color & mask);
		}
	}
	
	while ((j-=8)>=0) 
		*(adr++) = color;
    
	/* Draw `back` pixels if necessary */
	mask=~(0xff>>(j&7));
	*adr = (*adr & ~mask) | (color & mask);

	return EOK;
}

int GGIputhline(ggi_visual *vis,int x,int y,int w,void *buffer)
{ 
	unsigned char *adr,*buff=(unsigned char *)buffer;
	int mask,i,j,diff=0;
	unsigned char foo,color;

	/* Clipping */
	if (y<(VIS_GC(vis)->cliptl.y) || y>=(VIS_GC(vis)->clipbr.y)) return 0;
	if (x< (VIS_GC(vis)->cliptl.x)) {
		diff=(VIS_GC(vis)->cliptl.x)-x;
		x+=diff;w-=diff;
		buff+=(diff>>3);diff&=7;
	}
	if (x+w>(VIS_GC(vis)->clipbr.x)) {
		w=(VIS_GC(vis)->clipbr.x)-x;
	}
	if (w<=0) return 0;

	adr=((unsigned char *)(VIS_FB_LINEAR(vis)));
	adr+=(x+y*VIS_FB_WIDTH(vis))>>3;

	color=buff[0];
	if ((i=x&7)) {
		if ((j=w+i-8)<0) {
			mask=(0xff>>i)&(0xff<<-j); 
		} else {
			mask=0xff>>i;
		}

		foo = (*adr & ~mask);
		*adr = foo | ((color >> (i+=diff)) & mask);

		if (j<0)
			return EOK;
		else
			adr++;
	} else
		j=w;i+=diff;

	while ((j-=8)>=0) {
		color <<= (8-i);
		color |= *(++buff) >> i;
		*adr = color;
	}

    
	if (j&=7) {
		color <<= (8-i);
		color |= *(++buff) >> i;

		mask=~(0xff>>j);
		foo = (*adr & ~mask);
		*adr = foo | ((color >> i) & mask);
	}
  
	return 0;
}

int GGIgethline(ggi_visual *vis,int x,int y,int w,void *buffer)
{ 
	unsigned char *adr,*buff=(unsigned char *)buffer;
	int mask,i,j;
	unsigned char color;

	if (x<0||y<0||x+w>VIS_FB_WIDTH(vis)||y>=VIS_FB_HEIGHT(vis)) return -1;

	adr=((unsigned char *)(VIS_FB_LINEAR(vis)));
	adr+=(x+y*VIS_FB_WIDTH(vis))>>3;

	if ((i=x&7)) {
		if ((j=w+i-8)<0) {
			mask=(0xff>>i)&(0xff<<-j); 
		} else {
			mask=0xff>>i;
		}

		*buff = (*adr & mask) << (8-i);

		if (j<0)
			return EOK;
		else
			adr++;
	} else
		j=w;
	

	while ((j-=8)>=0) {
		color = *adr;
		*(buff++) |= color >> i;
		*(buff) = (color << (8-i));
	}

    
	if (j&=7) {
		mask=~(0xff>>j);
		*buff |= (*adr & mask) >> i;
	}
  
	return 0;
}

