/* 
   Graphics library for GGI. Vertical lines.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <string.h>

#include "ggi-dl.h"

/**********************************/
/* draw/get/put a vertical line */
/**********************************/

int GGIdrawvline(ggi_visual *vis,int x,int y,int height)
{
	unsigned char *adr;
	int i,sw,bm;

	/* Clipping */
	if (x< (VIS_GC(vis)->cliptl.x) ||
	    x>=(VIS_GC(vis)->clipbr.x)) return EOK;
	if (y< (VIS_GC(vis)->cliptl.y)) {
		int diff=(VIS_GC(vis)->cliptl.y)-y;
		y     +=diff;
		height-=diff;
	}
	if (y+height>(VIS_GC(vis)->clipbr.y)) {
		height=(VIS_GC(vis)->clipbr.y)-y;
	}
	if (height<=0) return 0;

	bm=(0x80>>(x&7));
	sw=(VIS_FB_WIDTH(vis))>>3;
	adr=((unsigned char *)(VIS_FB_LINEAR(vis)));
	adr+=(x>>3)+y*sw;

	if(VIS_GC_FGCOLOR(vis)&1) 
		for (i=height;i--;adr+=sw) *adr |= bm;
	else
		for (i=height;i--;adr+=sw) *adr &= ~bm;
  
	return 0;
}

int _GGIdrawvline(ggi_visual *vis,int x,int y,int height)
{
	unsigned char *adr;
	int i,sw,bm;

	bm=(0x80>>(x&7));
	sw=(VIS_FB_WIDTH(vis))>>3;
	adr=((unsigned char *)(VIS_FB_LINEAR(vis)));
	adr+=(x>>3)+y*sw;

	if(VIS_GC_FGCOLOR(vis)&1) 
		for (i=height;i--;adr+=sw) *adr |= bm;
	else
		for (i=height;i--;adr+=sw) *adr &= ~bm;
  
	return 0;
}

int GGIputvline(ggi_visual *vis,int x,int y,int height,void *buffer)
{ 
	unsigned char *adr,*buff=(unsigned char *)buffer;
	int mask=0x80,sw,i,bm;

	/* Clipping */
	if (x< (VIS_GC(vis)->cliptl.x) ||
	    x>=(VIS_GC(vis)->clipbr.x)) return EOK;
	if (y< (VIS_GC(vis)->cliptl.y)) {
		int diff=(VIS_GC(vis)->cliptl.y)-y;
		y     +=diff;
		height-=diff;
		buff +=diff>>3;
		mask>>=diff&7;
	}
	if (y+height>(VIS_GC(vis)->clipbr.y)) {
		height=(VIS_GC(vis)->clipbr.y)-y;
	}

	bm=(0x80>>(x&7));
	sw=(VIS_FB_WIDTH(vis))>>3;
	adr=((unsigned char *)(VIS_FB_LINEAR(vis)));
	adr+=(x>>3)+y*sw;

	for (i=0;i<height;i++,adr+=sw) {
		if (*buff & mask) 
			*adr |= bm;
		else
			*adr &= ~bm;
		mask >>= 1;
		if (mask==0) {
			mask=0x80;
			buff++;
		}
	}

  	return 0;
}

int GGIgetvline(ggi_visual *vis,int x,int y,int height,void *buffer)
{ 
	unsigned char *adr,*buff=(unsigned char *)buffer;
	int mask,sw,i,bm;

	if (x<0||y<0||x>=VIS_FB_WIDTH(vis)||y+height>VIS_FB_HEIGHT(vis)) return -1;

	sw=(VIS_FB_WIDTH(vis))>>3;
	adr=((unsigned char *)(VIS_FB_LINEAR(vis)));
	adr+=(x>>3)+y*sw;

	bm=(0x80>>(x&7));
	mask=0x80;
	for (i=0;i<height;i++,adr+=sw) {
		*buff |= (*adr & bm) ? mask : 0;
		mask>>=1;
		if (mask==0) {
			mask=0x80;
			buff++;
		}
	}

  	return 0;
}

