/* 
   Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ggi-dl.h"

int (*fallback_crossblit)(ggi_visual *src,int sx,int sy,int swidth,int sheight,
			  ggi_visual *dst,int dx,int dy,int dwidth,int dheight);

/* Cross blitting of a rectangular area. Speedup for 32 bit.
 */
int GGIcrossblit(ggi_visual *src,int sx,int sy,int sw,int sh,
		 ggi_visual *dst,int dx,int dy,int dw,int dh)
{
	char *srcp,*dstp;
	ggi_directbuffer_t buf=NULL;
        ggi_pixellinearbuffer *plb=NULL;
        	
	/* search for a directbuffer where we can blit to faster ... */

	while(1) {
		if (ggiDBGetBuffer(dst,&buf) || buf==NULL) 
			{ buf=NULL;break; }
		if (ggiDBGetLayout(buf) == blPixelLinearBuffer &&
		    (plb=ggiDBGetPLB(buf)) &&
		    plb->setup == sp32a32b8g8r8A8 ) break;
	}

	if (!buf || sx<0 || sy<0 || dx<0 || dy <0 ||
	            sx+sw>VIS_FB_WIDTH(src) || sy+sh>VIS_FB_HEIGHT(src) ||
	            dx+dw>VIS_FB_WIDTH(dst) || dy+dh>VIS_FB_HEIGHT(dst) )
		return fallback_crossblit(src,sx,sy,sw,sh,dst,dx,dy,dw,dh);

	/* O.K. - destination has a suitable buffer. Now let's fly ... */
	
	srcp=(char *)((uint32 *)VIS_FB_LINEAR(src)+sy*VIS_FB_WIDTH(src)+sx);
	dstp=(char *)plb->write+dy*plb->stride+dx*sizeof(uint32);

	if (sw==dw&&sh==dh) {	/* NOSTRETCH */

		dw<<=2;	/* set right size ... */
		while(sh--) {
			memcpy(dstp,srcp,dw);
			dstp+=plb->stride;
			srcp+=VIS_FB_WIDTH(src)*sizeof(uint32);
		}

	} else {		/* STETCH */

		ggi_uint pixel;
		int errx,erry,ddx,ssx,ty,tx,dtx,dty;
	
		ddx=dx;ssx=sx;erry=(dh-sh)>>1;
		ty =sy+sh;tx =sx+sw;
		dty=dy+dh;dtx=dx+dw;
	
		/* The basic stretching core. Please speed me up. Teunis ? */
		for(;sy<ty;sy++,srcp+=VIS_FB_WIDTH(src)*sizeof(uint32)) {
			erry+=dh;
			if (erry<=0) continue;

			while(erry>0) {
				dx=ddx;sx=ssx;errx=0;
				for(;sx<tx;sx++) {
					errx+=dw;
					if (errx<=0) continue;
					pixel=*((uint32 *)srcp+sx);

					while(errx>0 && dx<dtx) {
						*((uint32 *)dstp+dx++)=pixel;
						errx-=sw;
					}
				}
				dy++;erry-=sh;dstp+=plb->stride;
				if (dy>=dty) break;
			}
		}

	}		/* STETCH */
	return EOK;

}
