/* 
   Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ggi-dl.h"


/**********************/
/* draw/get/put a box */
/**********************/
int GGIdrawbox(ggi_visual *vis,int x,int y,int width,int length)
{
  /* Do correct clipping so we can use faster _ggiDrawHLine */
  
	if (y< (VIS_GC(vis)->cliptl.y)) {
		int diff=(VIS_GC(vis)->cliptl.y)-y;
		y     +=diff;
		length-=diff;
	}
	if (y+length>(VIS_GC(vis)->clipbr.y)) {
		length=(VIS_GC(vis)->clipbr.y)-y;
	}
	if (x< (VIS_GC(vis)->cliptl.x)) {
		int diff=(VIS_GC(vis)->cliptl.x)-x;
		x     +=diff;
		width-=diff;
	}
	if (x+width>(VIS_GC(vis)->clipbr.x)) {
		width=(VIS_GC(vis)->clipbr.x)-x;
	}

	if (length<=0 || width<=0 ) 
		return EOK;

	while(length--)
		_ggiDrawHLine(vis,x,y++,width);

	return EOK;
}

int GGIputbox(ggi_visual *vis,int x,int y,int width,int length,void *buffer)
{ 
	int rowadd = (VIS_FB_BPP(vis)*width+7)/8/sizeof(uint8);

	/* Pre-Clipping, so we don't call PutHline without need. */

	if (y< (VIS_GC(vis)->cliptl.y)) {
		int diff=(VIS_GC(vis)->cliptl.y)-y;
		y     +=diff;
		length-=diff;
		buffer = (uint8 *)(buffer)+rowadd*diff; 
	}
	if (y+length>=(VIS_GC(vis)->clipbr.y)) {
		length=(VIS_GC(vis)->clipbr.y)-y;
	}
	if (length<0||x>=VIS_GC(vis)->clipbr.x||x+width<=VIS_GC(vis)->cliptl.x) 
		return EOK;

	while(length--) {
		ggiPutHLine(vis,x,y++,width,buffer);
		buffer= (uint8 *)(buffer)+rowadd; 
	}

	return EOK;
}

int GGIgetbox(ggi_visual *vis,int x,int y,int width,int length,void *buffer)
{ 
	int rowadd = (VIS_FB_BPP(vis)*width+7)/8/sizeof(uint8);

	/* Unclipped - get */
	while (length--) {
		ggiGetHLine(vis,x,y++,width,buffer);
		buffer= (uint8 *)(buffer)+rowadd; 
	}
	return EOK;
}
