/*
 * This file contains the functions for a capture module that writes all
 * events it sees to a file. Handy for debugging and "demo-recodring".
 */

#include <stdio.h>
#include "evstack.h"

#define MAXSUBSTACK 16

typedef struct capture_state
{
	FILE *handle;
} capture_state;

static capture_state defaultstate = {
	NULL
};

static int capture_use;
static int capture_open(evstack *blank);
static int capture_close(evstack *slf);
static void capture_handler(evstack *slf,ggi_event *ev,int *left);
static int capture_uses(evstack *slf);

static evstack_operations capture_ops = {
	capture_open,
	capture_close,
	capture_handler,
	capture_uses,
	NULL,
	NULL
};

static int capture_open(evstack *blank)
{
	capture_state *dat;
	FILE *hand;
	char namebuffer[256];
	
	sprintf(namebuffer,"capture%d.ev",blank->id&EV_ID_MASK);
	if ((hand=fopen(namebuffer,"w"))==NULL) return -1;

	sprintf(blank->name,"Capture v1.0 (instance %.4x)",
		blank->id & EV_ID_MASK);

	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	dat=(capture_state *)ev_malloc(sizeof(capture_state));
	memcpy(dat,&defaultstate,sizeof(capture_state));
	dat->handle=hand;

	blank->private=dat;

	capture_use++;

	return 0;
}

static int capture_close(evstack *slf)
{
	struct capture_state *ss=slf->private;

	fclose(ss->handle);
	capture_use--;
	
	return 0;
}

/* Event handler code.
 */

static void capture_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	struct capture_state *ss=slf->private;

	for(;ev->size;ev=ev_evpage_next(ev))       /* Loop through all events in page */
	{
		fwrite(ev,ev->size,1,ss->handle);
		fflush(ss->handle);
	}
}

static int capture_uses(evstack *slf)
{
	return capture_use;
}

void init_capture(void)
{       
	ev_register_class("capture",&capture_ops);
	printk("Registered capture.\n");
}
