#include "evstack.h"

/* This module is used to debug event-handling. It will print out all events which
 * are currently on an event-page.
 */
#define CLASSNAME "debug"

static ev_id debug_id;
static int debug_use=0;

static int  debug_open(evstack *blank);
static int  debug_close(evstack *slf);
static void debug_handler(evstack *slf,ggi_event *ev,int *left);
static int  debug_uses(evstack *slf);

/* Operations for the module */
static evstack_operations debug_ops ={
	debug_open,
	debug_close,
	debug_handler,
	debug_uses,
	NULL,
	NULL
};

static int debug_open(evstack *blank)
{
	sprintf(blank->name,"Debug v1.0 (instance %.4x)",
		blank->id&EV_ID_MASK);

	/* Who is in charge of changig me ? */
	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	blank->private=NULL;

        MOD_INC_USE_COUNT;
	debug_use++;

	return 0;	/* Everything's OK! */
}

static int debug_close(evstack *slf)
{
        MOD_DEC_USE_COUNT;
	debug_use--;

	return 1;
}

/* A sample debug module
 */
static void debug_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	for(;ev->size;ev=ev_evpage_next(ev))       /* Loop through all events in page */
	{
		/* We are interested in all events (even consumed ones). Trying to dump them. */
		printk("%c ", ev_is_consumed(ev) ? 'c' : ' ');
		switch(ev->any.type&0x7f)
		{ case evKeyPress:
			printk("Key-Press : ");
			goto dumpkey;
		  case evKeyRelease:
			printk("Key-Release : ");
			goto dumpkey;
		  case evKeyRepeat:
			printk("Key-Repeat : ");
			goto dumpkey;
		  case evKeyState:
			printk("Key-State : ");

			dumpkey:
			{
				ggi_key_event *evdat;

			  	evdat=(ggi_key_event *)ev;

				printk(" Code=%4x Sym=%4x \n",evdat->code,evdat->sym);
			}
			break;
		  case evPtrRelative:
			printk("Ptr-Relative : ");
			goto dumpptrmove;
		  case evPtrAbsolute:
			printk("Ptr-Absolute : ");
			dumpptrmove:
			{
				ggi_pmove_event *evdat;

			  	evdat=(ggi_pmove_event *)ev;

				printk(" x=%-4d y=%-4d \n",evdat->x,evdat->y);
			}
			break;

		  case evPtrButtonPress:
			printk("Ptr-ButtPress : ");
			goto dumpptrbutt;
		  case evPtrButtonRelease:
			printk("Ptr-ButtRelease : ");
			dumpptrbutt:
			{
				ggi_pbutton_event *evdat;

			  	evdat=(ggi_pbutton_event *)ev;

				printk(" x=%4x y=%4x \n",evdat->state,evdat->button);
			}
			break;

		  case evPtrState:
			printk("Ptr-State : ");
			printk("\n");
			break;

		  case evValRelative:
			printk("Val-Relative : ");
			goto dumpvalmove;
		  case evValAbsolute:
			printk("Val-Absolute : ");
			dumpvalmove:
			{
				ggi_val_event *evdat;
				int cnt,cnt2;

			  	evdat=(ggi_val_event *)ev;

				for(cnt=cnt2=0;cnt<32;cnt++)
					if (evdat->changed & (1<<cnt)) {
						printk("Axis %-3d:%-8d",evdat->first+cnt,evdat->value[cnt2]);
						cnt2++;
					}
				printk("\n");
			}
			break;
		  case evValState:
			printk("Val-State : ");
			printk("\n");
			break;

		  case evCommand:
			printk("Command : ");
			{
				ggi_cmd_event *evdat=(ggi_cmd_event *)ev;
				int x;

				printk(" Command=%8x ( ",evdat->code);
				
				for(x=0;x<20;x++)
				  printk("%2x ",*(((unsigned char *)(evdat->data))+x));
				printk(")\n");
			}
			break;
		                                  
		  default: return;
		}
	}
}

static int debug_uses(evstack *slf)
{
	return debug_use;
}

#ifdef MODULE
#define init_debug init_module
#endif

int init_debug(void)
{ 
	debug_id=ev_register_class(CLASSNAME,&debug_ops);
	printk("Registered %s.\n",CLASSNAME);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	ev_unregister_class(debug_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif
