/*
 * This is the usermode part of the /dev/event interface.
 * It will read events from the kernel and distribute tem to usermode stacks.
 * Attention: This one is using threads, too.
 */

#ifdef __KERNEL__
#warning This is a usermode driver. MUST be compiled in user context.
#endif

#include <pthread.h>
#include <unistd.h>
#include <sys/fcntl.h>

#define CLASSNAME "devevuser"
#define MAXMINOR 32
#define READBUFSIZE 8192

#undef DEBUG
#ifdef DEBUG
#define DEBUG_LEVEL 2
#endif

#include "evstack.h"

typedef struct {
	int refcount;
	int handle;
} devevent_state;

static devevent_state defaultstate=
{	1,
	-1	/* file not yet open */
};

/* First the real driver. It is running in a seperate thread, to
 * allow for asynchronous execution.
 */

/* FIXME ! Make this a list at best */
static volatile evstack *thestack=NULL;

static void *devevent_readthread(void *parm)
{
	/* parm is not used. */
	evstack *slf;
	devevent_state *des;

	unsigned char buffer[256];

	while(1)
	{
		usleep(100000);
		slf=thestack;if (!slf) continue;
		des=slf->private;if (!des) continue;

		read(des->handle,buffer,1);
		read(des->handle,buffer+1,(*buffer)-1);
		printk("bla -\n");
		ev_send(NULL,NULL,(ggi_event *)buffer,slf->report_to);

	}
	return 0;	/* NOT REACHED */
}

/* Now for the evstack part.
 */

static ev_id devevent_id;
static int  devevent_use=0;

static int  devevent_open(evstack *blank);
static int  devevent_close(evstack *slf);
static void devevent_handler(evstack *slf,ggi_event *ev,int *left);
static int  devevent_uses(evstack *slf);

static evstack_operations devevent_ops={
	devevent_open,
	devevent_close,
	devevent_handler,
	devevent_uses,
	NULL,
	NULL
};

static int devevent_open(evstack *blank)
{
	devevent_state *des;
	int handle;

	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	if ((des=blank->private))
	{
		des->refcount++;
		DEBUG2("Stack cloned.",0);
	}
	else
	{	
		handle=open("/dev/event0",O_RDWR);
		if (handle==-1) printk("open failed.\n");
	        des=blank->private=ev_malloc(sizeof(devevent_state));
                memcpy(des,&defaultstate,sizeof(devevent_state));
                des->handle=handle;
                thestack=blank;
		DEBUG2("Stack created.",0);
	}
	sprintf(blank->name,"devevent-user v1.0 (instance %.4x)",
		blank->id&EV_ID_MASK);
	
	MOD_INC_USE_COUNT;
	devevent_use++;

	return 0;
}

static int devevent_close(evstack *slf)
{
	devevent_state *des=slf->private;
        
	if (--des->refcount==0)		/* No more instances left */
	{
		ev_free(des);
		DEBUG2("Stack destroyed (minor %d).",des->minor);
	}
	MOD_DEC_USE_COUNT;
	devevent_use--;

	return 0;
}

	
/* /dev/event module. Relays all commands to kernel space.
 */
static void devevent_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	devevent_state *des=slf->private;

	DEBUG2("Stack handler (minor %d).",des->minor);
	for(;ev->size;ev=ev_evpage_next(ev))       /* Loop through all events in page */
	{
		/* We are interested in sending unconsumed events to the kernel. */

		if ( ! ev_is_consumed(ev) ) 
		{
			write(des->handle,ev,ev->size);
		}
	}
}

static int devevent_uses(evstack *stack)
{
	return devevent_use;
}

#ifdef MODULE
#define init_devevent init_module
#endif

static pthread_t measure_thread;

int init_devevent(void)
{ 
	devevent_id=ev_register_class(CLASSNAME,&devevent_ops);
	printk("Registered %s.\n",CLASSNAME);

	pthread_create(&measure_thread,NULL,devevent_readthread,NULL);

	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	pthread_kill_other_threads_np();
	ev_unregister_class(devevent_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif

