#undef DEBUG
#ifdef DEBUG
#define DEBUG_LEVEL 2
#endif

#include "evstack.h"

typedef struct parser_state 
{	int refcount;
	int width,height;
	int curx,cury;
	ggi_color cur_fg,cur_bg;
} parser_state;

#define SCROLLERSTATE parser_state
#undef SCROLLERSTATE
#include "scroll.inc"

static parser_state default_state = 
{ 	1,
	80, 25, 
	 0,  0,
	{0xffff,0xffff,0xffff},	{0x0000,0x0000,0x0000}
};

/* A sample dumb terminal parser module
 */

#define CLASSNAME "dumbterm"

static ev_id dumb_id;
static int dumb_use = 0;

static int dumb_open(evstack *blank);
static int dumb_close(evstack *slf);
static void dumb_handler(evstack *slf,ggi_event *evpage,int *left);
static int dumb_uses(evstack *slf);
static int dumb_write(evstack *slf, char *str);
static int dumb_read(evstack *slf, int lineno, int maxlen, char *str);

static struct evstack_operations dumb_ops = {
	dumb_open,
	dumb_close,
	dumb_handler,
	dumb_uses,
	dumb_write,
	dumb_read
};

static int dumb_open(evstack *blank)
{
	parser_state *dat;

	sprintf(blank->name,"DumbTerm v1.0 (instance %.4x)",
		blank->id&EV_ID_MASK);
	
	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	if ((dat=blank->private))
	{
		dat->refcount++;
		DEBUG2("Stack cloned (rc %d).",des->refcount);
	}
	else
	{	
		blank->private=dat=ev_malloc(sizeof(parser_state));
		memcpy(dat,&default_state,sizeof(parser_state));
		DEBUG2("Stack created (rc %d).",des->refcount);
	}

        MOD_INC_USE_COUNT;
	dumb_use++;

	return 0;
}

static int dumb_close(evstack *slf)
{
	parser_state *des=slf->private;
        
	if (--des->refcount==0)		/* No more instances left */
	{
		ev_free(des);
		DEBUG2("Stack destroyed (minor %d).",des->minor);
	}

        MOD_DEC_USE_COUNT;
	dumb_use--;

	return 0;
}

/* write-char handler. Parts stolen form console.c
 */
static int dumb_do_write(evstack *slf,ggi_event *ev, int *left,
			 const char *buf, int count)
{
	char outbuf[256];
	DataScrollWriteChar dswc;
	struct parser_state *ps=slf->private;

	DEBUG1("count=%d",count);
	while(count--)
	{	
		switch(*buf)
	  	{ case 13: 
	  		scroll_gotoxy(ev,left,ps->curx=0,ps->cury,CUR_NORM,0);
	  		break;
		  case 10:
		  	ps->cury++;
			if ( ps->cury >= ps->height )
			{	scroll_scroll(ev,left,0,1,ps->width,ps->height-1,0,ps->height-ps->cury-1);
				ps->cury=ps->height-1;
			}
	  		scroll_gotoxy(ev,left,ps->curx,ps->cury,CUR_NORM,0);
	  		break;
		  case 8: 
			if (ps->curx) 
			{	ps->curx--;
				scroll_gotoxy(ev,left,ps->curx,ps->cury,CUR_NORM,0);
			}
	  		break;
#ifndef MODULE
	  	  case 3: /* FIXME DEMO ! */
			ps->cur_fg.r=random()&0xffff;
			ps->cur_fg.g=random()&0xffff;
			ps->cur_fg.b=random()&0xffff;
	  	  	break;
	  	  case 4: /* FIXME DEMO ! */
			ps->cur_bg.r=random()&0xffff;
			ps->cur_bg.g=random()&0xffff;
			ps->cur_bg.b=random()&0xffff;
	  	  	break;
#endif
	  	  default: 
			/* display the char */
			dswc.dx=1;dswc.dy=0;
			dswc.numchr=1;
			dswc.bg_col=ps->cur_bg;
			dswc.fg_col=ps->cur_fg;
			dswc.attrib=0;
			dswc.chr[0]=*buf;
			build_cmd_event(outbuf,CmdScrollWriteChar,&dswc,sizeof(dswc)-sizeof(dswc.chr)+sizeof(dswc.chr[0])*dswc.numchr);
                        DEBUG1("sending scroller command",0);
                        ev=ev_evpage_add(ev,left,(ggi_event *)&outbuf);

	  	  	if (++ps->curx>=ps->width) 
	  	  	{	ps->curx=0;ps->cury++; 
				if ( ps->cury >= ps->height )
				{	scroll_scroll(ev,left,0,1,ps->width,ps->height-1,0,ps->height-ps->cury-1);
					ps->cury=ps->height-1;
				}
				scroll_gotoxy(ev,left,ps->curx,ps->cury,CUR_NORM,0);
			}
                        break;
	  	}
	  buf++;
	}
	return 0;
}

/* Event handler code. Incoming part.
 */
static void send_key(ggi_event *ev,int *left,int symbol)
{
	DataTtyWriteChar evdata;
	char outchar[256];
	
	/* For testing purposes we simply echo it. */
	if ( ( KTYP(U(symbol))==KT_LATIN ||
	       KTYP(U(symbol))==KT_LETTER ) )
	{	
		evdata.size=1;
		evdata.data[0]=symbol;
		build_cmd_event(outchar,CmdTtyInWriteChar,&evdata,5);

		DEBUG1("sending echo command '%c'.\n",symbol);

               TRACE1({ int x;
                for(x=0;x<20;x++)
                     printk("%2x ",*(((unsigned char *)(outchar))+x));
                printk(")\n"); }
               );
                                                                                                  
		ev_evpage_add(ev,left,(ggi_event *)&outchar);
	}
}

/* Event handler code. Outgoing part.
 */
static void dumb_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	ggi_cmd_event *evdat;

	/* Loop through all events in page */
	for(;ev->size;ev=ev_evpage_next(ev))
	{
		/* We are interested in translating CmdTtyOutWriteChar to 
		   scroller-actions */

		switch(ev->any.type)
		{ 
		  case evCommand:
		  	evdat=(ggi_cmd_event *)ev;
		  	switch(evdat->code)
		  	{
		  	  case CmdTtyOutWriteChar:
		  	  	ev_consume(ev);
				DEBUG1("got event %x.\n",evdat->code);
				{ DataTtyWriteChar *dt=(DataTtyWriteChar *)evdat->data;
				  DEBUG1("%d chars :%*.*s.\n",dt->size
				  	  ,dt->size,dt->size,dt->data);
				  dumb_do_write(slf,ev,left,dt->data,dt->size);
				} break;
		  	  default:
		  	  	break;
			}
			break;
		  case evKeyPress:
		  case evKeyRepeat:
	  	  	ev_consume(ev);
			DEBUG1("got key event.\n",0);
			send_key(ev,left,ev->key.sym);
			break;
		  default: 
		  	break;
		}
	}
}

static int dumb_uses(evstack *slf)
{
	return dumb_use;
}

static int dumb_write(evstack *slf, char *str)
{
	return -EINVAL;
}

static int dumb_read(evstack *slf,int lineno,int maxlen,char *str)
{
	if (lineno==0) 
	{
		strcpy(str,"No info available."); return 0;
	}
	return -EINVAL;
}

#ifdef MODULE
#define init_dumb init_module
#endif

int init_dumb(void)
{ 
	dumb_id=ev_register_class(CLASSNAME,&dumb_ops);
	printk("Registered %s.\n",CLASSNAME);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	ev_unregister_class(dumb_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif
