/*
 * This is how it probably should be done, but the current kgi 
 * implementation will not allow multiple heads. This will hang 
 * the machine rocksolid. See #ifdef KGI_HAS_BEEN_FIXED.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <linux/kd.h>
#include <signal.h>
#include "evstack.h"
#include "parsearg.h"
#include <asm/io.h>
#include <unistd.h>
#include <pthread.h>

#ifdef __KERNEL__
#warning This is a usermode driver. It needs threads. Compile in user context.
#endif

#define CLASSNAME "joystick"

#undef DEBUG
#ifdef DEBUG
#define DEBUG_LEVEL 2
#endif

typedef struct {
	int refcount;		/* For cloning. Needed ? */
	int xold,yold;		/* Needed for smoothing */
	int smooth,wobble;	/* smooth-strength, max allowed wobble */
} joystick_state;

static joystick_state defaultstate=
{	1,
	0,0,
	0,5,
};

/* First the real driver. It is running in a seperate thread, to
 * allow for asynchronous execution.
 */

/* FIXME ! Make this a list at best */
static volatile evstack *thestack=NULL;

static int getaxis(int mask)
{
	int x=0;
	int mc=1000;

	while((inb(0x201)&mask) && mc--);
	mc=1000;outb(0x00,0x201);
	while((inb(0x201)&mask) && mc --) x++;
	return x;
}

static void *joy_measure(void *parm)
{
	/* parm is not used. */
	evstack *slf;
	joystick_state *jss;

	unsigned char buffer[256];
	ggi_val_event *valev=(ggi_val_event *)buffer;
	
	int x,y;

	/* We are userspace. ask for IO-permissions. */
	if (ioperm(0x201,1,1))
  		perror("ioperm");

	valev->size=sizeof(ggi_val_event);
	valev->type=evValAbsolute;
	valev->first=0;
	valev->changed=0x0003;	/* lowest 2 axis changed */
	
	while(1)
	{
		usleep(100000);
		slf=thestack;if (!slf) continue;
		jss=slf->private;if (!jss) continue;
		x=getaxis(1);
		y=getaxis(2);

		if ( abs(x-jss->xold) > jss->wobble ||
		     abs(y-jss->yold) > jss->wobble ) {
			valev->value[0]=x;
			valev->value[1]=y;
			ev_send(NULL,NULL,(ggi_event *)buffer,slf->report_to);
			jss->xold=x;jss->yold=y;
		}

	}
	return 0;	/* NOT REACHED */
}

/* Now for the evstack part.
 */

static ev_id joystick_id;
static int  joystick_use=0;

static int  joystick_open(evstack *blank);
static int  joystick_close(evstack *slf);
static void joystick_handler(evstack *slf,ggi_event *ev,int *left);
static int  joystick_uses(evstack *slf);

static evstack_operations joystick_ops={
	joystick_open,
	joystick_close,
	joystick_handler,
	joystick_uses,
	NULL,
	NULL
};

static int joystick_open(evstack *blank)
{
	joystick_state *des;

	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	if ((des=blank->private))
	{
		des->refcount++;
	}
	else
	{	
		if (thestack) return -1;	/* Only one for now. */
	        des=blank->private=ev_malloc(sizeof(joystick_state));
                memcpy(des,&defaultstate,sizeof(joystick_state));
                thestack=blank;
	}
	DEBUG2("Stack created.",0);
	sprintf(blank->name,"joystick v1.0 (inst. %.4x)",
		blank->id&EV_ID_MASK);
	
	MOD_INC_USE_COUNT;
	joystick_use++;

	return 0;
}



static int joystick_close(evstack *slf)
{
	joystick_state *des=slf->private;
        
	if (--des->refcount==0)		/* No more instances left */
	{
		DEBUG2("Stack destroyed.",0);
		thestack=NULL;
	}
	MOD_DEC_USE_COUNT;
	joystick_use--;

	return 0;
}

	
/* joystick module. Relays all commands to userspace.
 */
static void joystick_handler(evstack *slf,ggi_event *ev,int *left)
{ 
#if 0
	ggi_cmd_event *evdat;
	int size;
	unsigned char *head,*evptr;
	joystick_state *des=slf->private;

	DEBUG2("Stack handler (minor %d).",des->minor);
	for(;ev->size;ev=ev_evpage_next(ev))       /* Loop through all events in page */
	{
		/* We are not much interested in incoming events. */
		switch(ev->any.type)
		{ 
                  case evCommand:
		  default:
		  	break;
		}
	}
#endif
}

static int joystick_uses(evstack *stack)
{
	return joystick_use;
}

#ifdef MODULE
#define init_joystick init_module
#endif

static pthread_t measure_thread;

int init_joystick(void)
{ 
	joystick_id=ev_register_class(CLASSNAME,&joystick_ops);
	printk("Registered %s.\n",CLASSNAME);

	pthread_create(&measure_thread,NULL,joy_measure,NULL);

	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	pthread_kill_other_threads_np();
	ev_unregister_class(joystick_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif

