/* ----------------------------------------------------------------------------
**	selection "device". Reads pointer events, writes TtyIn
** ----------------------------------------------------------------------------
**
**	Copyright (C) 1996-1997, Andreas Beck [becka@uni-duesseldorf.de]
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** -----------------------------------------------------------------------------
*/

#define	MAINTAINER	"Andreas Beck, [becka@uni-duesseldorf.de]"

#undef	DEBUG		/* enable debugging code only on explicit request */
#ifdef DEBUG
#define DEBUG_LEVEL 2
#endif

#define TRACE TRACE2

#include "evstack.h"

enum { ST_IDLE, 	/* idle, cursor off */
       ST_CURSORON, 	/* mouse has been moved, cursor is on */
       ST_1CLICK, 	/* mouse has been clicked once */
       ST_1REL,		/* mouse released 1st time */
       ST_2CLICK, 	/* mouse has been clicked twice */
       ST_2REL,		/* mouse released 2nd time */
       ST_3CLICK, 	/* mouse has been clicked three times */
       ST_3REL,		/* mouse released 3rd time */
       ST_READY 
     };

typedef struct selection_state 
{	
	int pointerx,pointery;
	int startx,starty;
	int stopx,stopy;
	int width,height;
	int state;
} selection_state;

static selection_state default_state = 
{
	  0,  0,
	  0,  0,
	  0,  0,
	 80, 25,
	  0
};

#include "scroll.inc"

/* A selection cut/copy/paste module
 */
#define CLASSNAME "selection"

static ev_id selection_id;
static int selection_use = 0;

static int selection_open(evstack *blank);
static int selection_close(evstack *slf);
static void selection_handler(evstack *slf,ggi_event *evpage,int *left);
static int selection_uses(evstack *slf);
static int selection_write(evstack *slf, char *str);
static int selection_read(evstack *slf, int lineno, int maxlen, char *str);

static struct evstack_operations selection_ops = {
	selection_open,
	selection_close,
	selection_handler,
	selection_uses,
	selection_write,
	selection_read
};

static int selection_open(evstack *blank)
{
	sprintf(blank->name,"Selection v1.0 (instance %.4x)",
		blank->id&EV_ID_MASK);
	
	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	blank->private=ev_malloc(sizeof(selection_state));
	memcpy(blank->private,&default_state,sizeof(selection_state));

	selection_use++;

	return 0;
}

static int selection_close(evstack *slf)
{
	ev_free(slf->private);
	selection_use--;

	return 0;
}

/* Event handler code.
 */
static void selection_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	struct selection_state *ps=slf->private;

	/* Loop through all events in page */
	for(;ev->size;ev=ev_evpage_next(ev))
	{
		/* We are interested in translating CmdTtyOutWriteChar to 
		   scroller-actions */
		switch(ev->any.type)
		{ 
		  case evPtrButtonPress:
			ev_consume(ev);
			if (ev->pbutton.button & 1) {

				scroll_gotoxy(ev,left,ps->stopx=ps->pointerx,
					ps->stopy=ps->pointery, CUR_SELECTEND, 3);

			}
			if (ev->pbutton.button & 2) {
				DataTtyWriteChar evdata;
				char outchar[256];
	
				evdata.size=9;
				strcpy(evdata.data,"Selection");
				build_cmd_event(outchar,CmdTtyInWriteChar,&evdata,4+evdata.size);
				ev_evpage_add(ev,left,(ggi_event *)&outchar);
			}
			if (ev->pbutton.button & 4) {
				switch(ps->state)
				{ case ST_CURSORON: 
					ps->state=ST_1CLICK;
					scroll_gotoxy(ev,left,ps->startx=ps->pointerx,
						ps->starty=ps->pointery, CUR_SELECTSTART, 3);
					scroll_gotoxy(ev,left,ps->stopx=ps->pointerx,
						ps->stopy=ps->pointery, CUR_SELECTEND, 3);
					break;
				  case ST_1REL    : 
				  	ps->state=ST_2CLICK;
				  	break;
				  case ST_2REL    : 
				  	ps->state=ST_3CLICK;
					scroll_gotoxy(ev,left,ps->startx=0,
						ps->starty, CUR_SELECTSTART, 3);
					scroll_gotoxy(ev,left,ps->stopx=(ps->width<<3)-1,
						ps->stopy=ps->pointery, CUR_SELECTEND, 3);
				  	break;
				  case ST_3REL    : 
				  	ps->state=ST_1CLICK;
					scroll_gotoxy(ev,left,ps->startx=ps->pointerx,
						ps->starty=ps->pointery, CUR_SELECTSTART, 3);
				  	break;
				}
			}
			break;

		case evPtrButtonRelease:
			ev_consume(ev);

			switch(ps->state)
			{ case ST_1CLICK  : ps->state=ST_1REL;break;
			  case ST_2CLICK  : ps->state=ST_2REL;break;
			  case ST_3CLICK  : ps->state=ST_3REL;break;
			}
			break;

		case evPtrRelative:
		case evPtrAbsolute:
			if (ev->any.type == evPtrRelative) {
				ps->pointerx += ev->pmove.x;
				ps->pointery += ev->pmove.y;
			} else { /* evPtrAbsolute */
				ps->pointerx = ev->pmove.x;
				ps->pointery = ev->pmove.y;
			}
			ev_consume(ev);
			ps->pointerx = (ps->pointerx < 0 ) ? 0 : 
			                 (ps->pointerx >= ps->width<<3 ) ? (ps->width<<3) - 1 : 
			                    ps->pointerx;
			ps->pointery = (ps->pointery < 0 ) ? 0 : 
					 (ps->pointery >= ps->height<<3 ) ? (ps->height<<3) - 1 : 
					   ps->pointery;

			switch(ps->state)
			{ 
			  case ST_1REL:
			  case ST_2REL:
			  case ST_3REL: 
			  case ST_IDLE: 
				ps->state=ST_CURSORON;
			  case ST_CURSORON:
				break;
			  case ST_1CLICK:
			  case ST_2CLICK:
				scroll_gotoxy(ev,left,ps->stopx=ps->pointerx,
					ps->stopy=ps->pointery, CUR_SELECTEND, 3);
				break;
			  case ST_3CLICK:
				scroll_gotoxy(ev,left,ps->stopx=(ps->width<<3)-1,
					ps->stopy=ps->pointery, CUR_SELECTEND, 3);
				break;
			}
			scroll_gotoxy(ev,left,ps->pointerx,
				ps->pointery, CUR_MOUSE, 3);
			break;

		  case evKeyRepeat:
		  case evKeyPress:
		  case evKeyRelease:
		  	/* Any Key event will disable selection area */

			ps->state=ST_IDLE;
			if (ps->startx>=0)
			{
				ps->startx=ps->starty=-1;
				scroll_gotoxy(ev,left,-1,-1, CUR_SELECTSTART, 3);
				scroll_gotoxy(ev,left,-1,-1, CUR_SELECTEND  , 3);
				scroll_gotoxy(ev,left,-1,-1, CUR_MOUSE      , 3);
			}
		  	break;

		  default: 
		  	break;
		}
	}
}

static int selection_uses(evstack *slf)
{
	return selection_use;
}

static int selection_write(evstack *slf, char *str)
{
	return -EINVAL;
}

static int selection_read(evstack *slf,int lineno,int maxlen,char *str)
{
	if (lineno==0) 
	{
		strcpy(str,"No info available."); return 0;
	}
	return -EINVAL;
}

#ifdef MODULE
#define init_selection init_module
#endif

int init_selection(void)
{ 
	selection_id=ev_register_class(CLASSNAME,&selection_ops);
	printk("Registered %s.\n",CLASSNAME);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	ev_unregister_class(selection_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif
