#undef DEBUG
#ifdef DEBUG
#define DEBUG_LEVEL 2
#endif

#include "evstack.h"

typedef struct sound_state 
{	int busy;
} sound_state;

static sound_state default_state = 
{ 	0
};

/* A sample soundserver module to actually play mksound commands.
 */

#define CLASSNAME "sound"

static ev_id sound_id;
static int sound_use = 0;

static int sound_open(evstack *blank);
static int sound_close(evstack *slf);
static void sound_handler(evstack *slf,ggi_event *evpage,int *left);
static int sound_uses(evstack *slf);
static int sound_write(evstack *slf, char *str);
static int sound_read(evstack *slf, int lineno, int maxlen, char *str);

static struct evstack_operations sound_ops = {
	sound_open,
	sound_close,
	sound_handler,
	sound_uses,
	sound_write,
	sound_read
};

static int sound_open(evstack *blank)
{
	sprintf(blank->name,"SoundServ v1.0 (instance %.4x)",
		blank->id&EV_ID_MASK);
	
	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	if (!blank->private)
	{
		blank->private=ev_malloc(sizeof(sound_state));
		memcpy(blank->private,&default_state,sizeof(sound_state));
	}
	
	sound_use++;

	return 0;
}

static int sound_close(evstack *slf)
{
	ev_free(slf->private);
	sound_use--;

	return 0;
}

/* Event handler code.
 */
static void sound_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	ggi_cmd_event *evdat;

	/* Loop through all events in page */
	for(;ev->size;ev=ev_evpage_next(ev))
	{
		/* We are interested in translating CmdTtyOutWriteChar to 
		   scroller-actions */

		switch(ev->any.type)
		{ 
		  case evCommand:
		  	evdat=(ggi_cmd_event *)ev;
		  	switch(evdat->code)
		  	{
		  	  case CmdMakeSound:
		  	  	ev_consume(ev);
				DEBUG1("got event %x.\n",evdat->code);
				{ DataMakeSound *dt=(DataMakeSound *)evdat->data;
				  printk("FIXME: Implement Play-Sound %s,%d Hz,%d ms.\n",
				  	dt->type,dt->frequency,dt->duration);
				} break;
		  	  default:
		  	  	break;
			}
			break;
		  default: 
		  	break;
		}
	}
}

static int sound_uses(evstack *slf)
{
	return sound_use;
}

static int sound_write(evstack *slf, char *str)
{
	return -EINVAL;
}

static int sound_read(evstack *slf,int lineno,int maxlen,char *str)
{
	if (lineno==0) 
	{
		strcpy(str,"No info available."); return 0;
	}
	return -EINVAL;
}

#ifdef MODULE
#define init_sound init_module
#endif

int init_sound(void)
{ 
	sound_id=ev_register_class(CLASSNAME,&sound_ops);
	printk("Registered %s.\n",CLASSNAME);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	ev_unregister_class(sound_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif
