/*
 * This is a usermode keyboard handler which hooks into the traditional linux 
 * keyboard driver. It is assumed /dev/tty? is attached to stdin.
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <linux/kd.h>
#include <signal.h>
#include "evstack.h"
#include "parsearg.h"
#include <unistd.h>
#include <pthread.h>

#include <linux/kd.h>
#include <unistd.h>

#ifdef __KERNEL__
#warning This is a usermode driver. It needs threads. Compile in user context.
#endif

#define CLASSNAME "userkey"

#undef DEBUG
#ifdef DEBUG
#define DEBUG_LEVEL 2
#endif

typedef struct {
	int refcount;		/* For cloning. Needed ? */
} userkey_state;

static userkey_state defaultstate=
{	1
};

/* First the real driver. It is running in a seperate thread, to
 * allow for asynchronous execution.
 */

/* FIXME ! Make this a list at best */
static volatile evstack *thestack=NULL;

static void *key_measure(void *parm)
{
	/* parm is not used. */
	evstack *slf;
	userkey_state *jss;

	unsigned char buffer[256];
	ggi_key_event *keyev=(ggi_key_event *)buffer;
	
	unsigned char readkey;
	int hlp;

	/* First set up the keyboard in mediumraw-mode. */
	ioctl(0,KDSKBMODE,2);

	keyev->size=sizeof(ggi_key_event);
	keyev->type=evKeyPress;
	keyev->sym=0;
	
	while(1)
	{
		usleep(100000);
		slf=thestack;if (!slf) continue;
		jss=slf->private;if (!jss) continue;

		read(0,&readkey,1);
		if (readkey&0x80) {
			keyev->type=evKeyRelease;
			keyev->code=readkey&0x7f;
		} else {
			keyev->type=evKeyPress;
			keyev->code=readkey;
		}
		ev_send(NULL,NULL,(ggi_event *)buffer,slf->report_to);

	}
	return 0;	/* NOT REACHED */
}

/* Now for the evstack part.
 */

static ev_id userkey_id;
static int  userkey_use=0;

static int  userkey_open(evstack *blank);
static int  userkey_close(evstack *slf);
static void userkey_handler(evstack *slf,ggi_event *ev,int *left);
static int  userkey_uses(evstack *slf);

static evstack_operations userkey_ops={
	userkey_open,
	userkey_close,
	userkey_handler,
	userkey_uses,
	NULL,
	NULL
};

static int userkey_open(evstack *blank)
{
	userkey_state *des;

	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	if ((des=blank->private))
	{
		des->refcount++;
	}
	else
	{	
		if (thestack) return -1;	/* Only one for now. */
	        des=blank->private=ev_malloc(sizeof(userkey_state));
                memcpy(des,&defaultstate,sizeof(userkey_state));
                thestack=blank;
	}
	DEBUG2("Stack created.",0);
	sprintf(blank->name,"userkey v1.0 (inst. %.4x)",
		blank->id&EV_ID_MASK);
	
	MOD_INC_USE_COUNT;
	userkey_use++;

	return 0;
}



static int userkey_close(evstack *slf)
{
	userkey_state *des=slf->private;
        
	if (--des->refcount==0)		/* No more instances left */
	{
		DEBUG2("Stack destroyed.",0);
		thestack=NULL;
	}
	MOD_DEC_USE_COUNT;
	userkey_use--;

	return 0;
}

	
/* userkey module. Relays all commands to userspace.
 */
static void userkey_handler(evstack *slf,ggi_event *ev,int *left)
{ 
#if 0
	ggi_cmd_event *evdat;
	int size;
	unsigned char *head,*evptr;
	userkey_state *des=slf->private;

	DEBUG2("Stack handler (minor %d).",des->minor);
	for(;ev->size;ev=ev_evpage_next(ev))       /* Loop through all events in page */
	{
		/* We are not much interested in incoming events. */
		switch(ev->any.type)
		{ 
                  case evCommand:
		  default:
		  	break;
		}
	}
#endif
}

static int userkey_uses(evstack *stack)
{
	return userkey_use;
}

#ifdef MODULE
#define init_userkey init_module
#endif

static pthread_t measure_thread;

void myxit(void)
{
	ioctl(0,KDSKBMODE,1);
}

int init_userkey(void)
{ 
	userkey_id=ev_register_class(CLASSNAME,&userkey_ops);
	printk("Registered %s.\n",CLASSNAME);

	atexit(myxit);
	pthread_create(&measure_thread,NULL,key_measure,NULL);

	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	pthread_kill_other_threads_np();
	ev_unregister_class(userkey_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif

