#include <linux/autoconf.h>
#include <asm/io.h>
#include "evstack.h"

#define CLASSNAME "vgascroll"

typedef struct scroller_state
{	int curx,cury;
	int mousex,mousey;
	int selendx,selendy;
	int selstx,selsty;
	int width,height;
	short int *backbuffer;
	short int *hwbuffer;
	int in_bg;
} scroller_state;

static scroller_state defaultstate = { 
	0,0,
	-1,-1,
	-1,-1,
	-1,-1,
	80,25,
	NULL,
	(void *)0xb8000,
	0
};

/* A sample usermode scroller routine. This one uses GGI.
 */

static ev_id vgascroll_id;
static int vgascroll_use;
static int vgascroll_open(evstack *blank);
static int vgascroll_close(evstack *slf);
static void vgascroll_handler(evstack *slf,ggi_event *ev,int *left);
static int vgascroll_uses(evstack *slf);

static evstack_operations vgascroll_ops = {
	vgascroll_open,
	vgascroll_close,
	vgascroll_handler,
	vgascroll_uses,
	NULL,
	NULL
};

/* FIXME ! ALPHA needs this, but doesn't define it. */
static void my_memcpy_fromio(char *membuffer, char *hwbuffer,int numbytes)
{
	while(numbytes--) *membuffer++=readw(hwbuffer++);
}

static void toback(struct scroller_state *ss)
{
	my_memcpy_fromio(ss->backbuffer,ss->hwbuffer,2*ss->width*ss->height);
}

/* FIXME ! ALPHA needs this, but doesn't define it. */
static void my_memcpy_toio(char *hwbuffer,char *membuffer,int numbytes)
{
	while(numbytes--) writeb(*membuffer++,hwbuffer++);
}

static void fromback(struct scroller_state *ss)
{
	my_memcpy_toio(ss->hwbuffer,ss->backbuffer,2*ss->width*ss->height);
}

static void mem_CopyBox(struct scroller_state *ss,short int *ptr,int x1,int y1,int w,int h,int tx,int ty)
{
	int xx,yy;
	if (y1<ty)
	{
		for(yy=h-1;yy>=0;yy--)
			memmove(ptr+tx+(ty+yy)*ss->width,ptr+x1+(y1+yy)*ss->width,2*w);
	}
	else
	{
		for(yy=0;yy<h;yy++)
			memmove(ptr+tx+(ty+yy)*ss->width,ptr+x1+(y1+yy)*ss->width,2*w);
	}
}

static void mem_DrawBox(struct scroller_state *ss,short int *ptr,int x1,int y1,int w,int h)
{
	int yy;
	for(yy=0;yy<h;yy++)
		memset(ptr+x1+(y1+yy)*ss->width,0,w*2);
}

static int vgascroll_open(evstack *blank)
{
	scroller_state *dat;

	sprintf(blank->name,"VGA-Scroll v1.0 (instance %.4x)",
		blank->id & EV_ID_MASK);

	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	dat=(scroller_state *)ev_malloc(sizeof(scroller_state));
	memcpy(dat,&defaultstate,sizeof(scroller_state));
	dat->backbuffer=ev_malloc(dat->width*dat->height*sizeof(short int));
	memset(dat->backbuffer,0,dat->width*dat->height*sizeof(short int));

	blank->private=dat;

	vgascroll_use++;

	return 0;
}

static int vgascroll_close(evstack *slf)
{
	scroller_state *dat=slf->private;
	ev_free(dat->backbuffer);

	vgascroll_use--;
	
	return 0;
}


/* Event handler code.
 */
static void vgascroll_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	ggi_cmd_event *evdat;
	struct scroller_state *ss=slf->private;

	/* remove old HW-cursor */
	if (!ss->in_bg)
	{	writew(readw(ss->hwbuffer+ss->curx+ss->cury*ss->width)^0xFF00,
	                    (ss->hwbuffer+ss->curx+ss->cury*ss->width));
		if (ss->mousex>=0)
			writew(readw(ss->hwbuffer+ss->mousex+ss->mousey*ss->width)^0x8000,
	                    (ss->hwbuffer+ss->mousex+ss->mousey*ss->width));
		if (ss->selendx>=0&&ss->selstx>=0)
		{	short int *p1,*p2,*p3;
			p1=ss->hwbuffer+ss->selendx+ss->selendy*ss->width;
			p2=ss->hwbuffer+ss->selstx +ss->selsty *ss->width;
			for(p3=p1;p3<=p2;p3++) 
				writew(readw(p3)^0x2000,p3);
			for(p3=p2;p3<=p1;p3++)
				writew(readw(p3)^0x2000,p3);
		}
	}
	
	for(;ev->size;ev=ev_evpage_next(ev))       /* Loop through all events in page */
	{
		/* We are interested in translating CmdScrollWriteChar to 
		   scroller-actions */

		switch(ev->any.type)
		{ 
		  case evCommand:
		  	evdat=(ggi_cmd_event *)ev;
		  	switch(evdat->code)
		  	{
		  	  case CmdScrollWriteChar:
		  	  	ev_consume(ev);
				DEBUG1("got event %x.",evdat->code);
				{ DataScrollWriteChar *dt=(DataScrollWriteChar *)evdat->data;
				  int x;
				  for(x=0;x<dt->numchr;x++)
				  {
				  	short int color;
				  	if ( ss->curx < 0 ||
					     ss->curx >= ss->width ||
					     ss->cury < 0 ||
				       	     ss->cury >= ss->height ) break;

				  	DEBUG1("Drawing 0x%x at (%d,%d)",dt->chr[x],ss->curx,ss->cury);
					color=  ((dt->bg_col.r&0x8000)>>1)|
						((dt->bg_col.g&0x8000)>>2)|
						((dt->bg_col.b&0x8000)>>3)|
						((dt->fg_col.r&0x8000)>>5)|
						((dt->fg_col.g&0x8000)>>6)|
						((dt->fg_col.b&0x8000)>>7);
					if (ss->in_bg)
					  *(ss->backbuffer+ss->curx+ss->cury*ss->width) = dt->chr[x]|color;
					else
					  writew(dt->chr[x]|color,ss->hwbuffer+ss->curx+ss->cury*ss->width);
				  	ss->curx+=dt->dx;ss->cury+=dt->dy;
				  }
				} break;
		  	  case CmdScrollGotoXY:
		  	  	ev_consume(ev);
				DEBUG1("got event %x.",evdat->code);
				{ DataScrollGotoXY *dt=(DataScrollGotoXY *)evdat->data;
				  switch(dt->curnum)
				  { case CUR_NORM:
					  ss->curx=dt->x>>dt->fixedpointshift;
					  ss->cury=dt->y>>dt->fixedpointshift;
					  if ( ss->curx < 0 ) ss->curx=0;
					  if ( ss->curx >= ss->width ) ss->curx=ss->width-1;
					  if ( ss->cury < 0 ) ss->cury=0;
					  if ( ss->cury >= ss->height ) ss->cury=ss->height-1;
					  break;
				    case CUR_MOUSE:
					  ss->mousex=dt->x>>dt->fixedpointshift;
					  ss->mousey=dt->y>>dt->fixedpointshift;
					  if ( ss->mousex < 0 ||
					       ss->mousex >= ss->width ||
					       ss->mousey < 0 ||
					       ss->mousey >= ss->height )
						  ss->mousex=-1;
					  break;
				    case CUR_SELECTEND:
					  ss->selendx=dt->x>>dt->fixedpointshift;
					  ss->selendy=dt->y>>dt->fixedpointshift;
					  if ( ss->selendx < 0 ||
					       ss->selendx >= ss->width ||
					       ss->selendy < 0 ||
					       ss->selendy >= ss->height )
						  ss->selendx=-1;
					  break;
				    case CUR_SELECTSTART:
					  ss->selstx=dt->x>>dt->fixedpointshift;
					  ss->selsty=dt->y>>dt->fixedpointshift;
					  if ( ss->selstx < 0 ||
					       ss->selstx >= ss->width ||
					       ss->selsty < 0 ||
					       ss->selsty >= ss->height )
						  ss->selstx=-1;
					  break;
				  }
				} break;
		  	  case CmdScrollScroll:
		  	  	ev_consume(ev);
				DEBUG1("got event %x.",evdat->code);
				{ DataScrollScroll *dt=(DataScrollScroll *)evdat->data;

				  /* Check if box is on-screen. */
				  if ( dt->x < 0 ||
				       dt->x >= ss->width ||
				       dt->y < 0 ||
				       dt->y >= ss->height ) break;

				  /* adjust box. */
				  if ( dt->x + dt->width > ss->width )
				      dt->width=ss->width-dt->x;
				  if ( dt->y + dt->height > ss->height )
				      dt->height=ss->height-dt->x;

				  if (!ss->in_bg)
				  	toback(ss);
				  if ( dt->dx==0 && dt->dy==0 )
				  { 
					mem_DrawBox(ss,ss->backbuffer,dt->x,dt->y,dt->width,dt->height);
				  }
				  else
				  { 
#ifndef SPEED_TEST
				    mem_CopyBox(ss,ss->backbuffer,dt->x,dt->y,
						dt->width,dt->height,
						dt->x+dt->dx,dt->y+dt->dy);
				    if (dt->dy<0)
					mem_DrawBox(ss,ss->backbuffer,dt->x,dt->y+dt->height+dt->dy,dt->width,-dt->dy);
				    if (dt->dy>0)
					mem_DrawBox(ss,ss->backbuffer,dt->x,dt->y,dt->width,dt->dy);
				    if (dt->dx<0)
					mem_DrawBox(ss,ss->backbuffer,dt->x+dt->width+dt->dx,dt->y,-dt->dx,dt->height);
				    if (dt->dx>0)
					mem_DrawBox(ss,ss->backbuffer,dt->x,dt->y,dt->dx,dt->height);
#endif
				  }
				  if (!ss->in_bg)
				  	fromback(ss);

				  DEBUG1("Scroll (%d,%d,%d,%d -> %d,%d)",
				  	dt->x,dt->y,dt->width,dt->height,dt->dx,dt->dy);
				} break;

		  	  case CmdVtActivate:
		  	  	{ DataVtActivate *vta=(DataVtActivate *)evdat->data;
		  	  	  if (vta->onoff)
				  {	
				  	fromback(ss);ss->in_bg=0;
				  } else
				  {
				  	toback(ss);ss->in_bg=1;
#if 0
				  	memset_io(ss->hwbuffer,0,ss->width*ss->height*2);
#else
					{ int x;short int *dst;
					  dst=ss->hwbuffer;
					  x=ss->width*ss->height;
					  while(x--) writew(0,dst++);
					}
#endif
				  	/* Leave blank screen for others */
				  }
		  	  	} break;

		  	  default:
		  	  	break;
			}
		  default: 
		  	break;
		}
	}
	/* Write new HW-cursor */
	if (!ss->in_bg)
	{	writew(readw(ss->hwbuffer+ss->curx+ss->cury*ss->width)^0xFF00,
	                    (ss->hwbuffer+ss->curx+ss->cury*ss->width));
		if (ss->mousex>=0)
			writew(readw(ss->hwbuffer+ss->mousex+ss->mousey*ss->width)^0x8000,
	                    (ss->hwbuffer+ss->mousex+ss->mousey*ss->width));
		if (ss->selendx>=0&&ss->selstx>=0)
		{	short int *p1,*p2,*p3;
			p1=ss->hwbuffer+ss->selendx+ss->selendy*ss->width;
			p2=ss->hwbuffer+ss->selstx +ss->selsty *ss->width;
			for(p3=p1;p3<=p2;p3++) 
				writew(readw(p3)^0x2000,p3);
			for(p3=p2;p3<=p1;p3++)
				writew(readw(p3)^0x2000,p3);
		}
	}
}

static int vgascroll_uses(evstack *slf)
{
	return vgascroll_use;
}
                                                        
#ifdef MODULE
#define init_vgascroll init_module
#endif

int init_vgascroll(void)
{       
	/* Find out the VGA size here ! FIXME ! */

	vgascroll_id=ev_register_class(CLASSNAME,&vgascroll_ops);
	printk("Registered %s.\n",CLASSNAME);

	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	ev_unregister_class(vgascroll_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif
