/* ----------------------------------------------------------------------------
**	xterm console parser
** ----------------------------------------------------------------------------
**
**	Copyright (C) 1996-1997, Steffen Seeger [seeger@physik.tu-chemnitz.de]
**	EvStack conversion by Andreas Beck [becka@uni-duesseldorf.de]
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** -----------------------------------------------------------------------------
*/

#define	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"

#undef	STRICT_XTERM	/* #define to restrict to original XTERM	*/
#undef	SETEC_ASTRONOMY	/* #define to force display of concealed text	*/

/* ========================================================================== *\
|*  ---===>>> THERE IS NO USER-SERVICABLE CODE BELOW THIS LINE :-) <<<===---  *|
\* ========================================================================== */

#undef	DEBUG		/* enable debugging code only on explicit request */
#undef	DEBUG_XTERM
#ifdef DEBUG
#define DEBUG_LEVEL 2
#endif

#define TRACE TRACE2

#include "evstack.h"

#include <linux/ascii.h>
#include "xterm.h"

/*	forward declaration of the font translation tables. See the end of this
**	file for the initializer.
*/
static unicode translation[4][256];

typedef struct parser_state 
{	
	int refcount;
	int width,height;
	int curx,cury;
	int saved_curx,saved_cury;
	int pointerx,pointery;
	ggi_color cur_fg,cur_bg;
	unicode *translate;
	int ctmap[4],saved_ctmap[4];
	int charset,saved_charset;
	int attrib,saved_attrib;
	isochar utf_char;
	int utf_count;
	int mode,saved_mode;
	int flag;
	uint8 state;
	ggi_uint 	npar;
	uint32	par[MAX_NPAR];
	uint32	tab_stop[TAB_STOPS];	/* tabstops		*/
	int top,bottom;
	int bell_pitch, bell_duration;
	int numbkchar;
	char bkchar[40];
} parser_state;

#define SCROLLERSTATE parser_state
#include "scroll.inc"

static parser_state default_state = 
{ 	1,
	80, 25, 
	 0,  0,
	 0,  0,
	 0,  0,
	{0xffff,0xffff,0xffff},	{0x0000,0x0000,0x0000},
	translation[LAT1_MAP],
	{ LAT1_MAP,GRAF_MAP,IBMPC_MAP,USER_MAP },{ LAT1_MAP,GRAF_MAP,IBMPC_MAP,USER_MAP },
	LAT1_MAP,LAT1_MAP,
	0,0,
	0,0,
	0,0,
	0,
	0,
	0,
	{0,},
	{0,},
	80, 25,
	0,0,
	0,""
};

/* A sample xterm terminal parser module
 */

#define CLASSNAME "xterm"

static ev_id xterm_id;
static int xterm_use = 0;

static int xterm_open(evstack *blank);
static int xterm_close(evstack *slf);
static void xterm_handler(evstack *slf,ggi_event *evpage,int *left);
static int xterm_uses(evstack *slf);

static struct evstack_operations xterm_ops = {
	xterm_open,
	xterm_close,
	xterm_handler,
	xterm_uses,
	NULL,
	NULL
};

static int xterm_open(evstack *blank)
{
	parser_state *ps;

	sprintf(blank->name,"Xterm v1.0 (instance %.4x)",
		blank->id&EV_ID_MASK);
	
	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	if ((ps=blank->private))
	{
		ps->refcount++;
		DEBUG2("Stack cloned (minor %d).",ps->minor);
	}
	else
	{	
	        ps=blank->private=ev_malloc(sizeof(parser_state));
                memcpy(ps,&default_state,sizeof(parser_state));
		DEBUG2("Stack created (minor %d).",ps->minor);
	}

	MOD_INC_USE_COUNT;
	xterm_use++;

	return 0;
}


static int xterm_close(evstack *slf)
{
	parser_state *ps=slf->private;

	if (--ps->refcount==0)		/* No more instances left */
	{
		ev_free(ps);
		DEBUG2("Stack pstroyed (minor %d).",ps->minor);
	}
	MOD_DEC_USE_COUNT;
	xterm_use--;

	return 0;
}

static void xterm_do_set_mode(struct parser_state *ps, ggi_event *ev, int *left, int mode, int on)
{
#if 0 /* Shall we really emulate this ? */
	if (mode == reverse_video) {

		if (!TEST_FLAG(ps->mode,reverse_video) != !on) {

			#warning invert alternate screen too!

			fb_invert(FB, 0, FB_SIZE - 1);
			scroll_modified(cons, ORG, ORG + FSIZE);
			ps->attrib ^= AT_REVERSE;
		}
	}
#endif


#if 0	/* !!! implement alternate screen	*/
	if ((mode == alt_screen) && (!TEST_FLAG(ps->mode,alt_screen) != !on)) {

		int cnt = FB_SIZE_X*SIZE_Y;

		if (TEST_FLAG(ps->mode,alt_screen)) {

			if (ORG + cnt < FB_SIZE) {

				fb_memcpy_tofb(FB, ALT_SCREEN, ORG, cnt);

           		} else {

				fb_memcpy_tofb(FB, ALT_SCREEN, ORG,
					FB_SIZE-ORG);
				fb_memcpy_tofb(FB, ALT_SCREEN+FB_SIZE-ORG,
					0, ORG+cnt-FB_SIZE);
			}

		} else {

			if (ORG + cnt < FB_SIZE) {

				fb_memcpy_fromfb(FB, ALT_SCREEN, ORG, cnt);

           		} else {

				fb_memcpy_fromfb(FB, ALT_SCREEN, ORG,
					FB_SIZE - ORG);
				fb_memcpy_fromfb(FB, ALT_SCREEN + FB_SIZE-ORG,
					0, ORG+cnt-FB_SIZE);
			}
		}
	}
#endif

	/* !!!  if (mode == column132 && TEST_FLAG(ps->mode,allow_132mode))
	 * !!!    con_resize(cons, SIZE_Y, on ? 132 : 80);
	 */

	if (on) {

		SET_FLAG(ps->mode,mode);

	} else {
	
		CLEAR_FLAG(ps->mode,mode);
	}

	if (mode == origin) {

		scroll_gotoxy(ev,left, ps->curx=0, ps->cury=0, CUR_NORM, 0);
	}
}

static int xterm_par_to_mode(register int ques, register int par)
{
	if (ques) {

		switch (par) {

			case    1:	return cursor_key;
			case    3:	return column132;
			case    4:	return smooth_scroll;
			case    5:	return reverse_video;
			case    6:	return origin;
			case    7:	return auto_wrap;
			case    8:	return auto_repeat;
			case    9:	return x10_mouse;
			case   38:	return textronixmode;
			case   40:	return allow_132mode;
			case   41:	return curses_fix;
			case   44:	return margin_bell;
			case   45:	return reverse_wrap;
			case   46:	return logging;
			case   47:	return alt_screen;
			case 1000:	return report_mouse;
			case 1001:	return track_mouse;
		}

	} else {

		switch (par) {

			case    3:	return display_ctrl;
			case    4:	return insert;
			case   20:	return newline;
		}
	}

	return -1;
}

static void xterm_set_mode(struct parser_state *ps, ggi_event *ev, int *left, int on)
{
	int i, mode;

	for (i = 0; i <= ps->npar; i++) {

		if ((mode = xterm_par_to_mode(TEST_FLAG(ps->flag,question), ps->par[i])) >= 0) {

			xterm_do_set_mode(ps,ev,left, mode, on);
		}
	}
}

static void xterm_backup_mode(struct parser_state *ps, ggi_event *ev, int *left, int save)
{
	int i, mode;

	for (i = 0; i <= ps->npar; i++) {

		if ((mode = xterm_par_to_mode(TEST_FLAG(ps->flag,question), ps->par[i])) >= 0) {

			if (save) {

				TRANS_FLAG(ps->saved_mode,ps->mode,mode);

			} else {	

				if (TEST_FLAG(ps->mode,mode) != TEST_FLAG(ps->saved_mode,mode)) {

					xterm_do_set_mode(ps,ev,left, mode,
						TEST_FLAG(ps->saved_mode,mode));
				}
			}
		}
	}
}

static void xterm_save_cur(struct parser_state *ps)
{
	#warning save reverse_mode too!

	ps->saved_curx	= ps->curx;
	ps->saved_curx	= ps->cury;
	ps->saved_attrib= ps->attrib;
	ps->saved_charset	= ps->charset;

	memcpy(ps->saved_ctmap ,ps->ctmap ,sizeof(ps->ctmap));
}


static void xterm_restore_cur(struct parser_state *ps, ggi_event *ev, int *left)
{
	ps->curx	= ps->saved_curx;
	ps->cury	= ps->saved_cury;
	ps->attrib	= ps->saved_attrib;	
	ps->charset	= ps->saved_charset;

	memcpy(ps->ctmap ,ps->saved_ctmap ,sizeof(ps->ctmap));

	ps->translate = translation[ps->ctmap[ps->charset]];
	scroll_gotoxy(ev,left, ps->curx, ps->cury, CUR_NORM, 0);
	CLEAR_FLAG(ps->flag,need_wrap);
}


static void xterm_set_term(struct parser_state *ps)
{
	switch(ps->par[0]) {

	case 10:

		if (ps->npar >= 1) {

			ps->bell_pitch = ((ps->par[1]<20) || (ps->par[1]>32767))
					 ? 0 : ps->par[1];
		} else {
	
			ps->bell_pitch = 750;
		}
		break;

	case 11:

		if (ps->npar >= 1) {

			ps->bell_duration = (ps->par[1] < 2000) ?  ps->par[1] : 0;

		} else {

			ps->bell_duration = 125;
		}
		break;
	}
}

static void csi_m(struct parser_state *ps)
{
	int i;

	for (i = 0; i <= ps->npar; i++) {

		switch (ps->par[i]) {

		case 0:
			ps->cur_fg=DEFAULT_FG;
			ps->cur_bg=DEFAULT_BG;
			ps->attrib = 0;
			break;

		case 1:
#if 0
			ps->attrib &= ~AT_INTENSITY;
#endif
			ps->attrib |= AT_BOLD;
			break;

#ifndef STRICT_XTERM

		case 2:
#if 0
			ps->attrib &= ~AT_INTENSITY;
			ps->attrib |= AT_HALF;
#endif
			break;
#endif

		case 4:
			ps->attrib |= AT_UNDERLINE;
			break;

		case 5:
			ps->attrib |= AT_BLINK;
			break;

		case 7:
			if (TEST_FLAG(ps->mode,reverse_video)) {

#if 0
				ps->attrib &= ~AT_REVERSE;

			} else {

				ps->attrib |= AT_REVERSE;
#endif
			}
			break;

#ifndef STRICT_XTERM
	#ifndef SETEC_ASTRONOMY
		case 8:
#if 0	/*???*/
			ps->attrib &= ~AT_FGCOLOR;
			ps->attrib |= (ps->attrib << 8) & AT_FGCOLOR;
#endif
			break;
	#endif
		case 10:
			/*	ANSI X3.64-1979 (SCO-ish?)
			**	Select primary font, don't display
			**	control chars if defined, don't set
			**	bit 8 on output.
			*/
			ps->translate = translation[ps->ctmap[ps->charset]];
			CLEAR_FLAG(ps->mode,display_ctrl);
			CLEAR_FLAG(ps->mode,toggle_meta);
			break;

		case 11:
			/*	ANSI X3.64-1979 (SCO-ish?)
			**	Select first alternate font, let's
			**	chars < 32 be displayed as ROM chars.
			*/
			ps->translate = translation[ps->ctmap[IBMPC_MAP]];
			SET_FLAG(ps->mode,display_ctrl);
			CLEAR_FLAG(ps->mode,toggle_meta);
			break;

		case 12:
			/*
			**	ANSI X3.64-1979 (SCO-ish?)
			**	Select second alternate font, toggle
			**	high bit before displaying as ROM char.
			*/
			ps->translate = translation[ps->ctmap[IBMPC_MAP]];
			SET_FLAG(ps->mode,display_ctrl);
			SET_FLAG(ps->mode,toggle_meta);
			break;
#endif

		case 21:
		case 22:
#if 0
			ps->attrib &= ~AT_INTENSITY;
			ps->attrib |= AT_NORMAL;
#endif
			break;

		case 24:
			ps->attrib &= ~AT_UNDERLINE;
			break;

		case 25:
			ps->attrib &= ~AT_BLINK;
			break;

		case 27:
			if (TEST_FLAG(ps->mode,reverse_video)) {

#if 0
				ps->attrib |= AT_REVERSE;

			} else {

				ps->attrib &= ~AT_REVERSE;
#endif
			}
			break;

#ifndef STRICT_XTERM

		case 38:
			/*	ANSI X3.64-1979 (SCO-ish?)
			**	Enables underscore, white foreground
			**	with white underscore (Linux - use
			**	default foreground).
			*/
			ps->attrib |= AT_UNDERLINE;

			ps->cur_fg=colormap[COL_WHITE];
			break;

		case 39:
			/*
			**	ANSI X3.64-1979 (SCO-ish?)
			** ???	Disable underline option. Reset color
			** ???	to default?
			**	It did this before:
			*/
			ps->attrib &= ~AT_UNDERLINE;

			ps->cur_fg=DEFAULT_FG;
			ps->cur_bg=DEFAULT_BG;

			break;	

		case 49:

			ps->cur_bg=DEFAULT_BG;
			break;
#endif

		default:
			if ((ps->par[i] >= 30) && (ps->par[i] <= 37)) {

				ps->cur_fg = colormap[ps->par[i]-30]; 

			} else {

				if ((ps->par[i] >= 40) && (ps->par[i] <= 47)) {	

					ps->cur_bg = colormap[ps->par[i]-40]; 
				}
			}
			break;
		}
	}

}

void xterm_do_reset_state(struct parser_state *ps)
{
	ps->mode	= 0;
	SET_FLAG(ps->mode,show_cursor);
	SET_FLAG(ps->mode,auto_wrap);
	SET_FLAG(ps->mode,newline);
	SET_FLAG(ps->mode,auto_repeat); /* This was not in the original handler.*/
	/* But I suppose autorepeat is wanted by almost anyone ... */
	SET_FLAG(ps->mode,meta_esc);

	ps->saved_mode	= ps->mode;

	CLEAR_FLAG(ps->flag,question);
	CLEAR_FLAG(ps->flag,caps_lock);
	CLEAR_FLAG(ps->flag,caps_sticky);
	CLEAR_FLAG(ps->flag,num_lock);

	ps->state	= ESnormal;
	ps->utf_char	= 0;
	ps->utf_count	= 0;

	memset(ps->tab_stop,0x01,sizeof(ps->tab_stop));
	ps->tab_stop[0]	= 0x01010100;

#if 0
	ORIGIN		= 0;
	ORG		= 0;
#endif
	ps->top		= 0;
	ps->bottom	= ps->width;

	ps->bell_pitch	= 750 /* Hz */;
	ps->bell_duration = 125 /* msec */;

	ps->charset	= 0;
	ps->ctmap[0]	= LAT1_MAP;
	ps->ctmap[1]	= GRAF_MAP;
	ps->ctmap[2]	= IBMPC_MAP; 
	ps->ctmap[3]	= USER_MAP;

	ps->translate	= translation[ps->ctmap[ps->charset]];

	ps->npar = 0;

	ps->attrib = 0;
}

void xterm_do_reset(struct parser_state *ps, ggi_event *ev, int *left)
{
	xterm_do_reset_state(ps);
	/* FIXME - maybe send resize-request here. */

	scroll_gotoxy(ev,left, ps->curx=0, ps->cury=0, CUR_NORM, 0);	/* Set cursor */
	scroll_scroll(ev,left,0,0,ps->width,ps->height,0,0); /* Clear screen */
	scroll_flush (ev,left);
}

static void xterm_application_key(DataTtyWriteChar *evdat, uint8 key, int mode)
{
	evdat->data[evdat->size++]= ASCII_ESC;
	evdat->data[evdat->size++]= mode ? 'O' : '[';
	evdat->data[evdat->size++]= key;
}

static void xterm_function_key(DataTtyWriteChar *evdat, ggi_uint key)
{
	const char *s;
	if (key >=FUNCKEYNUM) return;else s= funckey[key];

	if (s) {

		while ( *s ) evdat->data[evdat->size++]= *s++;
	}
}

static const char *cur = "BDCA";
static const char *pad = "0123456789+-*/\015,.?";
static const char *app = "pqrstuvwxylSRQMnn?";

static void xterm_numpad_key(struct parser_state *ps,DataTtyWriteChar *evdat, uint8 val,
			     uint8 shift)
{

	if (TEST_FLAG(ps->mode,applic_key) && !(shift & (1 << KG_SHIFT))) {

		xterm_application_key(evdat, app[val], 1);
		return;
	}

	if (!TEST_FLAG(ps->flag,num_lock)) {

		switch(K(KT_PAD, val)) {

		case K_PCOMMA:
		case K_PDOT:
			xterm_function_key(evdat, KVAL(K_REMOVE));
			return;

		case K_P0:
			xterm_function_key(evdat, KVAL(K_INSERT));
			return;

		case K_P1:
			xterm_function_key(evdat, KVAL(K_SELECT));
			return;

		case K_P2:
			xterm_application_key(evdat, cur[KVAL(K_DOWN)], 0);
			return;

		case K_P3:
			xterm_function_key(evdat, KVAL(K_PGDN));
			return;

		case K_P4:
			xterm_application_key(evdat, cur[KVAL(K_LEFT)], 0);
			return;

		case K_P6:
			xterm_application_key(evdat, cur[KVAL(K_RIGHT)], 0);
			return;

		case K_P7:
			xterm_function_key(evdat, KVAL(K_FIND));
			return;

		case K_P8:
			xterm_application_key(evdat, cur[KVAL(K_UP)], 0);
			return;

		case K_P9:
			xterm_function_key(evdat, KVAL(K_PGUP));
			return;
		}

	}

	evdat->data[evdat->size++]= pad[val];

	if ((val == KVAL(K_PENTER)) && TEST_FLAG(ps->mode,newline)) {

		evdat->data[evdat->size++]= ASCII_LF;
	}
}

static void xterm_report(ggi_event *ev,int *left,struct parser_state *ps, ggi_uint what,int backrep)
{
	char buf[40], *s = buf;
	char cbutton = ' ';
	DataTtyWriteChar evdat;
	char outchar[256];
	evdat.size=0;

	#define VT100ID	"\033[?1;2c"	/* VT100 with advanced video option */
	#define VT102ID	"\033[?6c"

	#define MOUSEX	('!' + (char)(ps->pointerx>>3))
	#define MOUSEY	('!' + (char)(ps->pointery>>3))

	switch (what) {

		case cursor:
			sprintf(buf, "\033[%hd;%hdR",
				ps->cury + (TEST_FLAG(ps->mode,origin) ? ps->top+1 : 1), ps->curx+1);
			break;

		case id:
			s = VT102ID;
			break;
 			
		case status:
			s = "\033[0n";
			break;

		case mouseup:
			sprintf(buf, "\033[M#%c%c", MOUSEX, MOUSEY);
			break;

		case mousedown3:
			cbutton++;
		case mousedown2:
			cbutton++;
		case mousedown1:
			sprintf(buf, "\033[M%c%c%c", cbutton, MOUSEX, MOUSEY);
			break;
	}

	#undef MOUSEX
	#undef MOUSEY

	/* FIXME ! BACKREPORTING ! */
	if (backrep)
	{
		build_cmd_event(outchar,CmdStkReprocess,NULL,0);
		ev_evpage_add(ev,left,(ggi_event *)&outchar);
		while( *s && ps->numbkchar<sizeof(ps->bkchar) ) 
			ps->bkchar[ps->numbkchar++]=*s++;
	}
	else
	{
		while(*s) evdat.data[evdat.size++]=*s++;

		if (evdat.size)
		{
			build_cmd_event(outchar,CmdTtyInWriteChar,&evdat,4+evdat.size);
			ev_evpage_add(ev,left,(ggi_event *)&outchar);
		}
	}
}

void xterm_handle_key_event(struct parser_state *ps, ggi_event *ev, int *left)
{
	DataTtyWriteChar evdat;
	char outchar[256];
	unicode	sym = ev->key.sym;
	ggi_uint type = KTYP(sym);
	ggi_uint val  = KVAL(sym);

	evdat.size=0;

	if (type >= 0xF0) {

		type -= 0xF0;

		switch(type) {

			case KT_LETTER:

				if (!TEST_FLAG(ps->flag,caps_lock) != 
				    !TEST_FLAG(ps->flag,caps_sticky)) {

					if (('a' <= val) && (val <= 'z')) {

						val -= 'a' - 'A';

					} else {

						if (('A'<=val) && (val<='Z')) {

							val += 'a' - 'A';
						}
					}
				}
				/* fall through to KT_LATIN! */

			case KT_LATIN:
				evdat.data[evdat.size++]=val;
				break;

			case KT_META:
				if (TEST_FLAG(ps->mode,meta_esc)) {

					evdat.data[evdat.size++]=ASCII_ESC;
					evdat.data[evdat.size++]=val;

				} else {

					evdat.data[evdat.size++]=val|0x80;
				}
				break;

			case KT_FN:
				xterm_function_key(&evdat, val);
				break;

			case KT_SPEC:

				switch (K(KT_SPEC, val)) {

				case K_ENTER:
					evdat.data[evdat.size++]=ASCII_CR;
					if (TEST_FLAG(ps->mode,newline)) {

						evdat.data[evdat.size++]=ASCII_LF;
					}
					break;

#if 0	/* This shouldn't be here ! */
				case K_BREAK:
					tty_insert_flip_char(tty, 0, TTY_BREAK);
					break;
#endif					

				case K_CAPS:
					TOGGLE_FLAG(ps->flag,caps_lock);
/* !!!					kbd_toggle_led(&KBD_STATE, 
!!!						       KBD_LED_CAPSLOCK);
!!!					kbd_setleds(dev.kbd);
*/					return;

				case K_NUM:
					if (TEST_FLAG(ps->mode,applic_key)) {

						xterm_application_key(&evdat,
							'P', 1);

					} else {

						TOGGLE_FLAG(ps->flag,num_lock);
/* !!!						kbd_toggle_led(&KBD_STATE,
!!!							       KBD_LED_NUMLOCK);
!!!						kbd_setleds(dev.kbd);
*/						return;
					}
					break;
#if 0 /* This shouldn't be here ! */
				case K_HOLD:
					if (tty->stopped) {

						start_tty(tty);

					} else {

						stop_tty(tty);
					}
					return;

				case K_SCROLLFORW:
					scroll_forward(cons, 0);
					return;

				case K_SCROLLBACK:
					scroll_backward(cons, 0);
					return;

				case K_CAPSON:
					TOGGLE_FLAG(ps->flag,caps_sticky);
/* !!!					kbd_toggle_led(&KBD_STATE,
!!!						       KBD_LED_CAPSLOCK);
!!!					kbd_setleds(dev.kbd);
*/					return;

				case K_BARENUMLOCK:
					TOGGLE_FLAG(ps->flag,num_lock);
/* !!!					kbd_toggle_led(&KBD_STATE, 
!!!						       KBD_LED_NUMLOCK);
!!!					kbd_setleds(dev.kbd);
*/					return;
#endif

				default:
					return;
				}
				break;

			case KT_SHIFT:

				if (val == KVAL(K_NORMAL_SHIFT) &&
					TEST_FLAG(ps->flag,caps_lock)) {

					CLEAR_FLAG(ps->flag,caps_lock);
/* !!!					kbd_toggle_led(&KBD_STATE,
!!!						       KBD_LED_CAPSLOCK);
!!!					kbd_setleds(dev.kbd);
*/				}
				return;

			case KT_PAD:
				xterm_numpad_key(ps,&evdat, val, ev->key.normal);
				break;

			case KT_CUR:
				xterm_application_key(&evdat, cur[val],
					TEST_FLAG(ps->mode,cursor_key));
				break;

			case KT_DEAD:
				printk(KERN_ERR __FILE__
					": deadkeys not implemented - and they don't belong here.\n");
				return;
			}

		} else {

			if (TEST_FLAG(ps->mode,utf8)) {

				if (sym < 0x80) {

					evdat.data[evdat.size++]=sym;

				} else if (sym < 0x800) {

					evdat.data[evdat.size++]=0xC0 | (sym >> 6);
					evdat.data[evdat.size++]=0x80 |(sym & 0x3F);

				} else {

					evdat.data[evdat.size++]=0xE0 |(sym >> 12);
					evdat.data[evdat.size++]=0x80 |(sym >>  6);
					evdat.data[evdat.size++]=0x80 |(sym & 0x3F);
				}
				/*	UTF-8 is defined to up to 31 bits,
				**	but as the keyboard delivers unicode
				**	chars only we only need 16 bit here.
				*/
			}
		}

	if (evdat.size)
	{
		build_cmd_event(outchar,CmdTtyInWriteChar,&evdat,4+evdat.size);
		ev_evpage_add(ev,left,(ggi_event *)&outchar);
	}
}


/* Setting up chartable translation
 */
static void xterm_set_charset(struct parser_state *ps, ggi_uint nr, char c)
{
	switch(c) {

	case 'K':		/* ??? original console code did this?	*/
	case 'A':		/* United Kingdom/national		*/
		ps->ctmap[nr] = USER_MAP;
		break;

	case '1':		/* 'alternate ROM standard character'	*/
	case 'B':		/* US-ASCII (ISO-8859-1 = Latin1)	*/
		ps->ctmap[nr] = LAT1_MAP;
		break;

	case '2':		/* 'alternate ROM graphics'		*/
	case '0':		/* VT100 graphics			*/
		ps->ctmap[nr] = GRAF_MAP;
		break;


	case 'U':		/* ??? original console code did this?	*/
		ps->ctmap[nr] = IBMPC_MAP;
		break;
	}

	if (ps->charset == nr) {

		ps->translate = translation[ps->ctmap[nr]];
	}
}
/* write-char handler. Parts stolen form console.c
 */
static int xterm_do_write(evstack *slf,ggi_event *ev, int *left,
			 const char *buf, int count)
{
	char outbuf[256];
	isochar c;		/* 8-bit character to process	*/
	isochar	tc;		/* translated character		*/
	ggi_uint printable;
	struct parser_state *ps=slf->private;

	/* setup the write-char event */
	scroll_write_buffer.dx=1;
	scroll_write_buffer.dy=0;

	DEBUG1("count=%d",count);
	while(count--)
	{	
		c = *buf++ & 0xff;		/* make positive */

		/* ???	shouldn't toggle_meta mode *change* bit 0x80 ?? */
		tc = ps->translate[TEST_FLAG(ps->mode,toggle_meta) ? (c | 0x80) : c];

		printable = (c >= 32) || ((1 << c) & (TEST_FLAG(ps->mode,display_ctrl)
						? ~CTRL_ALWAYS : ~CTRL_ACTION));

		if (printable && (ps->state == ESnormal)) {
#if 0 /**/
			if (TEST_FLAG(ps->flag,need_wrap)) {

				scroll_modified(cons, old_pos, WPOS);
				if (TEST_FLAG(ps->mode,auto_wrap)) {

					scroll_lf(ps,ev,left);
				}
				scroll_cr(ps,ev,left);
				old_pos = WPOS;
			}

			if (TEST_FLAG(ps->mode,insert)) {

				TRACE(printk("<INS>"));
				scroll_insert_chars(cons, 1);
			}

			TRACE(printk("%.2x %.8x     ", c, SCROLL.mode));
			scroll_write(cons, tc);
#endif
			scroll_write_buffer.bg_col=ps->cur_bg;
			scroll_write_buffer.fg_col=ps->cur_fg;
			scroll_write_buffer.attrib=ps->attrib;
			scroll_write(ev,left,tc);

		  	if (++ps->curx>=ps->width) 
		  	{	ps->curx=0;ps->cury++; 
				if ( ps->cury >= ps->height )
				{	scroll_scroll(ev,left,0,1,ps->width,ps->height-1,0,ps->height-ps->cury-1);
					ps->cury=ps->height-1;
				}
				scroll_gotoxy(ev,left,ps->curx,ps->cury, CUR_NORM, 0);
			}

			continue;
		}

		/*	Control characters can be used in the *middle* of
		**	an escape sequence.
		*/
		switch (c) {

		case ASCII_BEL:

			TRACE(printk("<BEL>"));
			if (ps->bell_pitch && ps->bell_duration) {

				scroll_mksound(ev,left, ps->bell_pitch, ps->bell_duration, "termbell");
			}
			continue;
	
		case ASCII_BS:

			TRACE(printk("<BS>"));
			scroll_bs(ps,ev,left);
			continue;

		case ASCII_HT:

			TRACE(printk("<HT>"));

			while (ps->curx < ps->width - 1) {

				ps->curx++;
				if (ps->tab_stop[ps->curx >> 5] & 
				     (1 << (ps->curx & 31))) {

					break;
				}
			}
			continue;

		case ASCII_LF:
		case ASCII_VT:
		case ASCII_FF:

			TRACE(printk("<LF>"));
			scroll_lf(ps,ev,left);
	  		break;
	
			if (!TEST_FLAG(ps->mode,newline)) {

				continue;
			}

		case ASCII_CR:

			TRACE(printk("<CR>"));
	  		scroll_cr(ps,ev,left);
			continue;

		case ASCII_SO:

			TRACE(printk("<S0>"));
			ps->charset = 1;
			ps->translate = translation[1];
			SET_FLAG(ps->mode,display_ctrl);
			continue;

	 	case ASCII_SI:

			TRACE(printk("<SI>"));
			ps->charset = 0;
			ps->translate = translation[0];
			CLEAR_FLAG(ps->mode,display_ctrl);
			continue;

		case ASCII_CAN:
		case ASCII_SUB:

			TRACE(printk("<CAN>"));
			ps->state = ESnormal; /* !!! display checkerboard!	*/
			continue;

		case ASCII_ESC:

			TRACE(printk("<ESC"));
			ps->state = ESesc;
			continue;

		case 127:

			TRACE(printk("<DEL>"));
#if 0 /* Why was this commented out ? */
			/* !!! scroll_del(cons); */
#endif
			continue;

		case 128+27:

			TRACE(printk("<CSI"));	
			ps->state = ESsquare;
			continue;
		}

		switch(ps->state) {

		case ESesc:

			TRACE(printk("%c", (char) c));
			ps->state = ESnormal;

			/*	handle sequences that do not change cursor
			**	position first.
			*/
			switch(c) {

			case '[':	ps->state = ESsquare;
					continue;

			case '>':	/* DECKPNM - numeric keypad */
					TRACE(printk("= DECKPNM>\n"));
					CLEAR_FLAG(ps->mode,applic_key);
					continue;

			case '=':	/* DECKPAM - appl. keypad */
					TRACE(printk("= DECKPAM>\n"));
					SET_FLAG(ps->mode,applic_key);
				 	continue;

			case '(':	/* SCS G0 - select character set 0 */
					TRACE(printk("= SCS0>\n"));
					ps->state = ESset;
					ps->par[0] = 0;	/* charset to set */
					continue;

			case ')':	/* SCS G1 - select character set 1 */
					TRACE(printk("= SCS1>\n"));
					ps->state = ESset;
					ps->par[0] = 1;
					continue;

			case '*':	/* SCS G2 - select character set 2 */
					TRACE(printk("= SCS2>\n"));
					ps->state = ESset;
					ps->par[0] = 2;
					continue;

			case '+':	/* SCS G3 - select character set 3 */
					TRACE(printk("= SCS3>\n"));
					ps->state = ESset;
					ps->par[0] = 3;
					continue;

			case '#':	ps->state = EShash;
					continue;

			case 'H':	/* HTS - set horiz. tab stop */
					TRACE(printk("= HTS>\n"));
					ps->tab_stop[ps->curx >> 5] |= (1 << (ps->curx & 31));
					continue;

			case 'N':	TRACE(printk("= S2>\n"));
					ps->charset = 2;
					ps->translate = translation[2];
					CLEAR_FLAG(ps->mode,display_ctrl);
					continue;

			case 'O':	TRACE(printk("= S3>\n"));
					ps->charset = 3;
					ps->translate = translation[3];
					CLEAR_FLAG(ps->mode,display_ctrl);
					continue;

			case 'Z':	/* DECID - report ID */
					TRACE(printk("= DECID>\n"));
					xterm_report(ev,left,ps, id,1);
					continue;

			case 'n':  /*	!!! full locking shift select G2 set */
			case 'o':  /* 	!!! full locking shift select G3 set */
					continue;

			case '7':	/* DECSC - save cursor */
					TRACE(printk("= DECSC>\n"));
					xterm_save_cur(ps);
					continue;

			#ifndef STRICT_XTERM
			case '%':	ps->state = ESpercent;
					continue;
			#endif
			}

			/*	Now care of sequences that (potentially) alter
			**	the cursor position. Note that *all* cases have
			**	to be terminated with 'break'.
			*/

			switch (c) {

			case 'E':	/* NEL - next line */
					TRACE(printk("= NEL>\n"));
					scroll_cr(ps,ev,left);
					scroll_lf(ps,ev,left);
					break;

			case 'M':	/* RI - reverse index */
					TRACE(printk("= RI>\n"));
					scroll_reverse_lf(ps,ev,left);
					break;

			case 'D':	/* IND - index (down one line) */
					TRACE(printk("= IND>\n"));
					scroll_lf(ps,ev,left);
					break;

			case '8':	/* DECRC - restore cursor */
					TRACE(printk("= DECRC>\n"));
					xterm_restore_cur(ps,ev,left);
					break;

			case 'c':	/* RIS - reset to initial state */
					TRACE(printk("= RIS>\n"));
					xterm_do_reset(ps,ev,left);
					break;

			default:
					TRACE(printk("= unknown ESC>\n"));
			}
			continue;

		case ESsquare:

			for(ps->npar = 0; ps->npar < MAX_NPAR; ps->npar++) {

				ps->par[ps->npar] = 0;
			}

			ps->npar = 0;
			ps->state = ESgetpars;

			if (c == '[') {		/* Function key */

				TRACE(printk("%c", (char) c));
				ps->state=ESfunckey;
				continue;
			}

			if (c == '?') {

				TRACE(printk("%c", (char) c));
				SET_FLAG(ps->flag,question);
				continue;
			}
			/* fall through if ((c != '[') && (c != '?')) */

		case ESgetpars:

			if ((c == ';') && (ps->npar < MAX_NPAR-1)) {

				TRACE(printk("%c", (char) c));
				ps->npar++;
				continue;

			} else {

				if ((c >= '0') && (c <= '9')) {

					TRACE(printk("%c", (char) c));
					ps->par[ps->npar] *= 10;
					ps->par[ps->npar] += c-'0';
					continue;

				} else {

					ps->state = ESgotpars;
				}
			}
			/* fall through when parameters are complete! */

		case ESgotpars:

			ps->state = ESnormal;

			TRACE(printk("%c", (char) c));

			if (TEST_FLAG(ps->flag,question)) {

				/*	all these sequences may alter cursor
				**	position. Terminate cases with 'break'!
				*/
				switch(c) {

				case 'h':	/* SM - set mode */

					TRACE(printk("= SM>\n"));
					xterm_set_mode(ps,ev,left, 1);
					break;

				case 'l':	/* RM - reset mode */

					TRACE(printk("= RM>\n"));
					xterm_set_mode(ps,ev,left, 0);
					break;

				case 'r':	/* DECRM - restore mode */

					TRACE(printk("= DECRM>\n"));	
					xterm_backup_mode(ps,ev,left, 0);
					break;

				case 's':	/* DECSM - save mode */

					TRACE(printk("= DECSM>\n"));
					xterm_backup_mode(ps,ev,left, 1);
					break;

				default:

					TRACE(printk("= unknown ESC[? >\n"));
				}

				CLEAR_FLAG(ps->flag,question);
				continue;
			}

			/*	first care of sequences that do not alter
			**	cursor postion.
			*/
			switch(c) {

			case 'J':	/* ED - erase display	*/

				TRACE(printk("= ED>\n"));
				switch(ps->par[0])
				{
				   case 0:
					scroll_scroll(ev,left,ps->curx,ps->cury,ps->width-ps->curx,1,0,0);
					scroll_scroll(ev,left,0,ps->cury+1,ps->width,ps->height-ps->cury-1,0,0);
					break;
				   case 1:
					scroll_scroll(ev,left,0,0,ps->width,ps->cury,0,0);
					scroll_scroll(ev,left,0,ps->cury,ps->curx+1,1,0,0);
					break;
				   case 2:
					scroll_scroll(ev,left,0,0,ps->width,ps->height,0,0);
					break;
				}
				continue;

			case 'K':	/* EL - erase in line */

				TRACE(printk("= EL>\n"));
				switch(ps->par[0])
				{ 
				   case 0:
					scroll_scroll(ev,left,ps->curx,ps->cury,ps->width-ps->curx,1,0,0);
					break;
				   case 1:
					scroll_scroll(ev,left,0,ps->cury,ps->curx+1,1,0,0);
					break;
				    case 2:
					scroll_scroll(ev,left,0,ps->cury,ps->width,1,0,0);
					break;
				}
				continue;

			case 'P':

				TRACE(printk("= ???>\n"));
				if (ps->par[0] <= 0) {

					ps->par[0] = 1;
				}
				/* FIXME ! */
#if 0
				scroll_delete_chars(cons, ps->par[0]);
#endif
				continue;

			case 'c':	/* ??? DA - device attributes */

				TRACE(printk("= DA>\n"));
				if (!ps->par[0]) {

					xterm_report(ev,left,ps, id,1);
				}
				continue;

			case 'g':	/* TBC - tabulation clear */

				TRACE(printk("= TBC>\n"));
				if (!ps->par[0]) {

					ps->tab_stop[ps->curx >> 5] &= ~(1 << (ps->curx & 31));

				} else {

					if (ps->par[0] == 3) {

						memset(ps->tab_stop,0,sizeof(ps->tab_stop));
					}
				}
				continue;

			case 'm':	/* SGR - select graphic redition */

				TRACE(printk("= SGR>\n"));
				csi_m(ps);
				continue;

			case 'n':	/* DSR - device status report */

				TRACE(printk("= DSR>\n"));

				if (ps->par[0] == 5) {
					xterm_report(ev,left,ps, status,1);
				} else {
					if (ps->par[0] == 6) {
					     xterm_report(ev,left,ps, cursor,1);
					}
				}
				continue;

			case '@':	/* ICH - insert (blank) character(s) */

				TRACE(printk("= ICH>\n"));
				if (! ps->par[0]) {

					ps->par[0]++;
				}
#if 0
				scroll_insert_chars(cons, ps->par[0]);
#endif
				continue;
		#if 0
			case 'q':	/* ??? DECLL - load LED */
					/* ??? check this! */

				if (!ps->par[0]) {

					kbd_set_led_state(&KBD_STATE, 0);

				} else {

					if (ps->par[0] < 5) {

						kbd_set_led(&KBD_STATE, ps->par[0]-1);
						kbd_setleds(KBD);
					}
				}
				continue;
		#endif
		#ifndef STRICT_XTERM

			case 'X':

				/* FIXME - what does this do ? */
#if 0
				scroll_erase_chars(cons, ps->par[0]);
#endif
				continue;

			case ']':	/* setterm functions */

				xterm_set_term(ps);
				continue;
		#endif
			}

			/*	now handle those sequences that may alter
			**	cursor position. NOTE: all cases have to
			**	be terminated with 'break'.
			*/

			switch(c) {

			case 'A':	/* CUU - cursor up	*/

				TRACE(printk("= CUU>\n"));
				if (!ps->par[0]) {

					ps->par[0]++;
				}
				scroll_gotoxy(ev, left, ps->curx, ps->cury -= ps->par[0], CUR_NORM, 0);
				break;

			case 'B':	/* CUD - cursor down	*/
			case 'e':

				TRACE(printk("= CUD>\n"));
				if (!ps->par[0]) {

					ps->par[0]++;
				}
				scroll_gotoxy(ev, left, ps->curx, ps->cury += ps->par[0], CUR_NORM, 0);
				break;

			case 'C':	/* CUF - cursor forward	*/
			case 'a':

				TRACE(printk("= CUF>\n"));
				if (!ps->par[0]) {

					ps->par[0]++;
				}
				scroll_gotoxy(ev, left, ps->curx += ps->par[0], ps->cury, CUR_NORM, 0);
				break;

			case 'D':	/* CUB - cursor backward*/

				TRACE(printk("= CUB>\n"));
				if (!ps->par[0]) {

					ps->par[0]++;
				}
				scroll_gotoxy(ev, left, ps->curx -= ps->par[0], ps->cury, CUR_NORM, 0 );
				break;

			case 'H':	/* CUP - set cursor position */
			case 'f':	/* HVP - horiz. and vert. position */

				TRACE(printk(c=='H' ? "= CUP>\n" : "= HVP>\n"));
				if (ps->par[0]) {

					ps->par[0]--;
				}
				if (ps->par[1]) {

					ps->par[1]--;
				}
				if (TEST_FLAG(ps->mode,origin)) {

					ps->par[0] += ps->top;
				}
				scroll_gotoxy(ev, left, ps->curx=ps->par[1], ps->cury=ps->par[0], CUR_NORM, 0);
				break;

			case 'L':

				TRACE(printk("= ???>\n"));
				if (ps->par[0] <= 0) {

					ps->par[0] = 1;
				}
				scroll_down(ps, ev, left, ps->cury, ps->bottom, ps->par[0]);
				break;

			case 'M':

				TRACE(printk("= ???>\n"));
				if (ps->par[0] <= 0) {

					ps->par[0] = 1;
				}
				scroll_up(ps, ev, left, ps->cury, ps->bottom, ps->par[0]);
				break;

			case 'h':	/* SM - set mode */

				TRACE(printk("= SM>\n"));
				xterm_set_mode(ps,ev,left, 1);
				break;

			case 'l':	/* RM - reset mode */

				TRACE(printk("= RM>\n"));
				xterm_set_mode(ps,ev,left, 0);
				break;

			case 'r':	/* DECSTBM - set top & bottom margin */

				TRACE(printk("= DECSTBM>\n"));

				if (!ps->par[0]) {

					ps->par[0]++;
				}

				if (!ps->par[1]) {

					ps->par[1] = ps->height;
				}

				/* Minimum allowed region is 2 lines */
				if ((ps->par[0]<ps->par[1]) && (ps->par[1]<=ps->height)) {

					ps->bottom = ps->par[0]-1;
					ps->bottom = ps->par[1];
					scroll_gotoxy(ev, left, ps->curx=0, ps->cury=0, CUR_NORM, 0);
				}
				break;

			case 'x':	/* !!! DECREQTPARM - report parameter */

				TRACE(printk("= DECREQTPARM>\n"));
				break;

			#ifndef STRICT_XTERM
			case 's':

				TRACE(printk("= save cur>\n"));
				xterm_save_cur(ps);
				break;

			case 'u':
	
				TRACE(printk("= rest cur>\n"));
				xterm_restore_cur(ps,ev,left);
				break;

			case 'E':

				if (!ps->par[0]) {

					ps->par[0]++;
				}
				scroll_gotoxy(ev,left, ps->curx=0, ps->cury += ps->par[0], CUR_NORM, 0);
				break;

			case 'F':

				if (!ps->par[0]) {

					ps->par[0]++;
				}
				scroll_gotoxy(ev,left, ps->curx=0, ps->cury -= ps->par[0], CUR_NORM, 0);
				break;

			case 'G':
			case '`':

				if (ps->par[0]) {

					ps->par[0]--;
				}
				scroll_gotoxy(ev,left, ps->curx=ps->par[0], ps->cury, CUR_NORM, 0);
				break;

			case 'd':

				if (ps->par[0]) {

					ps->par[0]--;
				}
				if (TEST_FLAG(ps->mode,origin)) {

					ps->par[0] += ps->top;
				}
				scroll_gotoxy(ev,left, ps->curx, ps->cury=ps->par[0], CUR_NORM, 0);
				break;
			#endif

			default:

				TRACE(printk("= unknown ESC[>\n"));
			}
			continue;

		case ESpercent:

			ps->state = ESnormal;
			TRACE(printk("%c", (char) c));

			switch (c) {

			case '@':	/* defined in ISO 2022 */

				TRACE(printk("= UTF-8 off>\n"));
				CLEAR_FLAG(ps->mode,utf8);
				continue;

			case 'G':	/* prelim official escape code */
			case '8':	/* retained for compatibility */

				TRACE(printk("= UTF-8 on>\n"));
				SET_FLAG(ps->mode,utf8);
				continue;
			}
			TRACE(printk("= unkown ESC-percent>\n"));
			continue;

		case ESfunckey:

			ps->state = ESnormal;
			TRACE(printk("fn-key %c\n", (char) c));
			continue;

		case EShash:

			if (c == '8') {	/* DEC screen alignment test */

#if 0	/* We can't do this with the current interface ... */
				TRACE(printk("= DECSAT>\n"));
				ERASE = (ERASE & 0xff00) | 'E';
				/* FIXME ! */
				scroll_scroll(ev,left,0,0,ps->width,ps->height,0,0); /* Clear screen */
				ERASE = (ERASE & 0xff00) | ' ';
#endif

			} else {

				TRACE(printk("= unkown ESC-#>\n"));
			}
			ps->state = ESnormal;
			continue;

		case ESset:

			TRACE(printk("%c>\n", (char) c));
			xterm_set_charset(ps, ps->par[0], c);
			ps->state = ESnormal;
			continue;

		default:

			TRACE(printk("<%.2x, state %i>\n", c, ps->state));
			ps->state = ESnormal;
			continue;
		}
	}
	scroll_flush(ev,left);
	return 0;
}


/* Event handler code.
 */
static void xterm_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	ggi_cmd_event *evdat;
	struct parser_state *ps=slf->private;

	/* Flush pending answers. Maybe we should check, if this is the
	   round-trip induced by CmdStkReprocess cia ev->size==0 ??? */
	if (ps->numbkchar) 
	{
		DataTtyWriteChar evdat;
		char outchar[256];

		evdat.size=ps->numbkchar;
		memcpy(evdat.data,ps->bkchar,ps->numbkchar);
		ps->numbkchar=0;

		build_cmd_event(outchar,CmdTtyInWriteChar,&evdat,4+evdat.size);
		ev_evpage_add(ev,left,(ggi_event *)&outchar);
	}

	/* Loop through all events in page */
	for(;ev->size;ev=ev_evpage_next(ev))
	{
		/* We are interested in translating CmdTtyOutWriteChar to 
		   scroller-actions */

		switch(ev->any.type)
		{ 
		  case evCommand:
		  	evdat=(ggi_cmd_event *)ev;
		  	switch(evdat->code)
		  	{
		  	  case CmdTtyOutWriteChar:
		  	  	ev_consume(ev);
				DEBUG1("got event %x.\n",evdat->code);
				{ DataTtyWriteChar *dt=(DataTtyWriteChar *)evdat->data;
				  DEBUG1("%d chars :%*.*s.\n",dt->size
				  	  ,dt->size,dt->size,dt->data);
				  xterm_do_write(slf,ev,left,dt->data,dt->size);
				} break;
		  	  default:
		  	  	break;
			}
			break;

		  case evPtrButtonPress:

			if (TEST_FLAG(ps->mode,report_mouse) | 
			    TEST_FLAG(ps->mode,x10_mouse)) {

		  	  	ev_consume(ev);
				if (ev->pbutton.button & 1) {

					xterm_report(ev,left,ps,mousedown1,0);
				}
				if (ev->pbutton.button & 2) {

					xterm_report(ev,left,ps,mousedown2,0);
				}
				if (ev->pbutton.button & 4) {

					xterm_report(ev,left,ps,mousedown3,0);
				}
			}
			break;

		case evPtrButtonRelease:

			if (TEST_FLAG(ps->mode,report_mouse)) {

		  	  	ev_consume(ev);
				xterm_report(ev,left,ps, mouseup,0);
			}
			break;

		case evPtrRelative:
		case evPtrAbsolute:
			if (ev->any.type == evPtrRelative) {
				ps->pointerx += ev->pmove.x;
				ps->pointery += ev->pmove.y;
			} else { /* evPtrAbsolute */
				ps->pointerx = ev->pmove.x;
				ps->pointery = ev->pmove.y;
			}
			ps->pointerx = (ps->pointerx < 0 ) ? 0 : 
			                 (ps->pointerx >= ps->width<<3 ) ? (ps->width<<3) - 1 : 
			                    ps->pointerx;
			ps->pointery = (ps->pointery < 0 ) ? 0 : 
					 (ps->pointery >= ps->height<<3 ) ? (ps->height<<3) - 1 : 
					   ps->pointery;

			if (TEST_FLAG(ps->mode,report_mouse) | 
			    TEST_FLAG(ps->mode,x10_mouse)) {

		  	  	ev_consume(ev);
				scroll_gotoxy(ev,left,ps->pointerx,ps->pointery, CUR_MOUSE, 3);
			}
			break;

		  case evKeyRepeat:
		  	if (!TEST_FLAG(ps->mode,auto_repeat)) break;
			/* Fallthrough, if set. */
		  case evKeyPress:

	  	  	ev_consume(ev);
		  	xterm_handle_key_event(ps,ev,left);
		  	break;

		  default: 
		  	break;
		}
	}
}

static int xterm_uses(evstack *slf)
{
	return xterm_use;
}

#ifdef MODULE
#define init_xterm init_module
#endif

int init_xterm(void)
{ 
	xterm_do_reset_state(&default_state);
	xterm_id=ev_register_class(CLASSNAME,&xterm_ops);
	printk("Registered %s.\n",CLASSNAME);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	ev_unregister_class(xterm_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif


/* -----------------------------------------------------------------------------
**	font translation tables
** -----------------------------------------------------------------------------
*/

static unicode translation[][256] = {
  /* 8-bit Latin-1 mapped to Unicode -- trivial mapping */
  #define LAT1_MAP	0
  {
    0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007,
    0x0008, 0x0009, 0x000a, 0x000b, 0x000c, 0x000d, 0x000e, 0x000f,
    0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016, 0x0017,
    0x0018, 0x0019, 0x001a, 0x001b, 0x001c, 0x001d, 0x001e, 0x001f,
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, 0x007f,
    0x0080, 0x0081, 0x0082, 0x0083, 0x0084, 0x0085, 0x0086, 0x0087,
    0x0088, 0x0089, 0x008a, 0x008b, 0x008c, 0x008d, 0x008e, 0x008f,
    0x0090, 0x0091, 0x0092, 0x0093, 0x0094, 0x0095, 0x0096, 0x0097,
    0x0098, 0x0099, 0x009a, 0x009b, 0x009c, 0x009d, 0x009e, 0x009f,
    0x00a0, 0x00a1, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7,
    0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af,
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7,
    0x00b8, 0x00b9, 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf,
    0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x00c7,
    0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf,
    0x00d0, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x00d7,
    0x00d8, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x00dd, 0x00de, 0x00df,
    0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x00e7,
    0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef,
    0x00f0, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x00f7,
    0x00f8, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x00fd, 0x00fe, 0x00ff
  },
  /* VT100 graphics mapped to Unicode */
  #define GRAF_MAP	1
  {
    0x0000, 0x0001, 0x0002, 0x0003, 0x0004, 0x0005, 0x0006, 0x0007,
    0x0008, 0x0009, 0x000a, 0x000b, 0x000c, 0x000d, 0x000e, 0x000f,
    0x0010, 0x0011, 0x0012, 0x0013, 0x0014, 0x0015, 0x0016, 0x0017,
    0x0018, 0x0019, 0x001a, 0x001b, 0x001c, 0x001d, 0x001e, 0x001f,
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x2192, 0x2190, 0x2191, 0x2193, 0x002f,
    0x2588, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x00a0,
    0x25c6, 0x2592, 0x2409, 0x240c, 0x240d, 0x240a, 0x00b0, 0x00b1,
    0x2591, 0x240b, 0x2518, 0x2510, 0x250c, 0x2514, 0x253c, 0xf800,
    0xf801, 0x2500, 0xf803, 0xf804, 0x251c, 0x2524, 0x2534, 0x252c,
    0x2502, 0x2264, 0x2265, 0x03c0, 0x2260, 0x00a3, 0x00b7, 0x007f,
    0x0080, 0x0081, 0x0082, 0x0083, 0x0084, 0x0085, 0x0086, 0x0087,
    0x0088, 0x0089, 0x008a, 0x008b, 0x008c, 0x008d, 0x008e, 0x008f,
    0x0090, 0x0091, 0x0092, 0x0093, 0x0094, 0x0095, 0x0096, 0x0097,
    0x0098, 0x0099, 0x009a, 0x009b, 0x009c, 0x009d, 0x009e, 0x009f,
    0x00a0, 0x00a1, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7,
    0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af,
    0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x00b5, 0x00b6, 0x00b7,
    0x00b8, 0x00b9, 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf,
    0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x00c7,
    0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf,
    0x00d0, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x00d7,
    0x00d8, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x00dd, 0x00de, 0x00df,
    0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x00e7,
    0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef,
    0x00f0, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x00f7,
    0x00f8, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x00fd, 0x00fe, 0x00ff
  },
  /* IBM Codepage 437 mapped to Unicode */
  #define IBMPC_MAP	2
  {
    0x0000, 0x263a, 0x263b, 0x2665, 0x2666, 0x2663, 0x2660, 0x2022, 
    0x25d8, 0x25cb, 0x25d9, 0x2642, 0x2640, 0x266a, 0x266b, 0x263c,
    0x25b6, 0x25c0, 0x2195, 0x203c, 0x00b6, 0x00a7, 0x25ac, 0x21a8,
    0x2191, 0x2193, 0x2192, 0x2190, 0x221f, 0x2194, 0x25b2, 0x25bc,
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, 0x2302,
    0x00c7, 0x00fc, 0x00e9, 0x00e2, 0x00e4, 0x00e0, 0x00e5, 0x00e7,
    0x00ea, 0x00eb, 0x00e8, 0x00ef, 0x00ee, 0x00ec, 0x00c4, 0x00c5,
    0x00c9, 0x00e6, 0x00c6, 0x00f4, 0x00f6, 0x00f2, 0x00fb, 0x00f9,
    0x00ff, 0x00d6, 0x00dc, 0x00a2, 0x00a3, 0x00a5, 0x20a7, 0x0192,
    0x00e1, 0x00ed, 0x00f3, 0x00fa, 0x00f1, 0x00d1, 0x00aa, 0x00ba,
    0x00bf, 0x2310, 0x00ac, 0x00bd, 0x00bc, 0x00a1, 0x00ab, 0x00bb,
    0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x2561, 0x2562, 0x2556,
    0x2555, 0x2563, 0x2551, 0x2557, 0x255d, 0x255c, 0x255b, 0x2510,
    0x2514, 0x2534, 0x252c, 0x251c, 0x2500, 0x253c, 0x255e, 0x255f,
    0x255a, 0x2554, 0x2569, 0x2566, 0x2560, 0x2550, 0x256c, 0x2567,
    0x2568, 0x2564, 0x2565, 0x2559, 0x2558, 0x2552, 0x2553, 0x256b,
    0x256a, 0x2518, 0x250c, 0x2588, 0x2584, 0x258c, 0x2590, 0x2580,
    0x03b1, 0x00df, 0x0393, 0x03c0, 0x03a3, 0x03c3, 0x00b5, 0x03c4,
    0x03a6, 0x0398, 0x03a9, 0x03b4, 0x221e, 0x03c6, 0x03b5, 0x2229,
    0x2261, 0x00b1, 0x2265, 0x2264, 0x2320, 0x2321, 0x00f7, 0x2248,
    0x00b0, 0x2219, 0x00b7, 0x221a, 0x207f, 0x00b2, 0x25a0, 0x00a0
  }, 
  /* User mapping -- default to codes for direct font mapping */
  #define USER_MAP	3
  {
    0xf000, 0xf001, 0xf002, 0xf003, 0xf004, 0xf005, 0xf006, 0xf007,
    0xf008, 0xf009, 0xf00a, 0xf00b, 0xf00c, 0xf00d, 0xf00e, 0xf00f,
    0xf010, 0xf011, 0xf012, 0xf013, 0xf014, 0xf015, 0xf016, 0xf017,
    0xf018, 0xf019, 0xf01a, 0xf01b, 0xf01c, 0xf01d, 0xf01e, 0xf01f,
    0xf020, 0xf021, 0xf022, 0xf023, 0xf024, 0xf025, 0xf026, 0xf027,
    0xf028, 0xf029, 0xf02a, 0xf02b, 0xf02c, 0xf02d, 0xf02e, 0xf02f,
    0xf030, 0xf031, 0xf032, 0xf033, 0xf034, 0xf035, 0xf036, 0xf037,
    0xf038, 0xf039, 0xf03a, 0xf03b, 0xf03c, 0xf03d, 0xf03e, 0xf03f,
    0xf040, 0xf041, 0xf042, 0xf043, 0xf044, 0xf045, 0xf046, 0xf047,
    0xf048, 0xf049, 0xf04a, 0xf04b, 0xf04c, 0xf04d, 0xf04e, 0xf04f,
    0xf050, 0xf051, 0xf052, 0xf053, 0xf054, 0xf055, 0xf056, 0xf057,
    0xf058, 0xf059, 0xf05a, 0xf05b, 0xf05c, 0xf05d, 0xf05e, 0xf05f,
    0xf060, 0xf061, 0xf062, 0xf063, 0xf064, 0xf065, 0xf066, 0xf067,
    0xf068, 0xf069, 0xf06a, 0xf06b, 0xf06c, 0xf06d, 0xf06e, 0xf06f,
    0xf070, 0xf071, 0xf072, 0xf073, 0xf074, 0xf075, 0xf076, 0xf077,
    0xf078, 0xf079, 0xf07a, 0xf07b, 0xf07c, 0xf07d, 0xf07e, 0xf07f,
    0xf080, 0xf081, 0xf082, 0xf083, 0xf084, 0xf085, 0xf086, 0xf087,
    0xf088, 0xf089, 0xf08a, 0xf08b, 0xf08c, 0xf08d, 0xf08e, 0xf08f,
    0xf090, 0xf091, 0xf092, 0xf093, 0xf094, 0xf095, 0xf096, 0xf097,
    0xf098, 0xf099, 0xf09a, 0xf09b, 0xf09c, 0xf09d, 0xf09e, 0xf09f,
    0xf0a0, 0xf0a1, 0xf0a2, 0xf0a3, 0xf0a4, 0xf0a5, 0xf0a6, 0xf0a7,
    0xf0a8, 0xf0a9, 0xf0aa, 0xf0ab, 0xf0ac, 0xf0ad, 0xf0ae, 0xf0af,
    0xf0b0, 0xf0b1, 0xf0b2, 0xf0b3, 0xf0b4, 0xf0b5, 0xf0b6, 0xf0b7,
    0xf0b8, 0xf0b9, 0xf0ba, 0xf0bb, 0xf0bc, 0xf0bd, 0xf0be, 0xf0bf,
    0xf0c0, 0xf0c1, 0xf0c2, 0xf0c3, 0xf0c4, 0xf0c5, 0xf0c6, 0xf0c7,
    0xf0c8, 0xf0c9, 0xf0ca, 0xf0cb, 0xf0cc, 0xf0cd, 0xf0ce, 0xf0cf,
    0xf0d0, 0xf0d1, 0xf0d2, 0xf0d3, 0xf0d4, 0xf0d5, 0xf0d6, 0xf0d7,
    0xf0d8, 0xf0d9, 0xf0da, 0xf0db, 0xf0dc, 0xf0dd, 0xf0de, 0xf0df,
    0xf0e0, 0xf0e1, 0xf0e2, 0xf0e3, 0xf0e4, 0xf0e5, 0xf0e6, 0xf0e7,
    0xf0e8, 0xf0e9, 0xf0ea, 0xf0eb, 0xf0ec, 0xf0ed, 0xf0ee, 0xf0ef,
    0xf0f0, 0xf0f1, 0xf0f2, 0xf0f3, 0xf0f4, 0xf0f5, 0xf0f6, 0xf0f7,
    0xf0f8, 0xf0f9, 0xf0fa, 0xf0fb, 0xf0fc, 0xf0fd, 0xf0fe, 0xf0ff
  }
  #define LAST_MAP	4
};

/*	The standard kernel character-to-font mappings are not invertible
**	-- this is just a best effort.
*/
