/*
** checkmode
**
** test for all available modes and output a list of them.
**
** $Id: checkmode.c,v 1.4 1998/03/09 19:07:16 msundberg Exp $ 
** $Log: checkmode.c,v $
** Revision 1.4  1998/03/09 19:07:16  msundberg
** Demos now check returncode of ggiInit() and ggiOpen().
** Link flying_ggis and stars with -lm on Digital Unix.
**
** Revision 1.3  1998/02/26 01:06:09  msundberg
** Changed dots.c to ASYNC mode.
** Minor fix for checkmode.c
**
** Revision 1.2  1998/02/15 16:44:06  msundberg
** Fixed to use ASYNC mode.
**
** Revision 1.1  1998/02/10 18:17:59  becka
** checkmode added.
**
** Revision 1.1  1998/02/10 10:47:54  ha
** Initial revision
**
**
*/

#define NEED_GRAPHTYPE_SYMS
#include <ggi/ggi.h>
#include <ggi/libggi.h>
#include <stdio.h>
#include <unistd.h>  /* getopt */

/* #define OFFBYONE */ /* if the graphtype texts are wrong in ggi.h */

ggi_visual_t vis;

/* most graphics resolutions taken from XFree */
/* EGA/CGA, text resolutions from */
/* M. Uphff: Die Programmierung der EGA/VGA Grafikkarte */
struct resolution {
	int x;
	int y;
} defaultresolutions[]= { 
	{320,200},
	{320,240},
	{320,350}, /* EGA text */
	{360,400}, /* VGA text */
	{400,300},
	{480,300},
	{512,384},
	{640,200}, /* EGA */
	{640,350}, /* EGA */
	{640,400},
	{640,480},
	{720,350}, /* MDA text */
	{720,400}, /* VGA text */
	{800,600},
	{1024,768},
	{1152,864},
	{1200,1024},
	{1600,1200},
	{0,0}          /* End mark!! */
};

ggi_graphtype defaultgraphtypes[]={
	/* can't handle textmode yet */
	GT_1BIT,                /*  1 bpp graphics              */
        GT_4BIT,                /*  4 bpp graphics              */
        GT_8BIT,                /*  8 bpp graphics              */
        GT_15BIT,               /* 15 bpp graphics              */
        GT_16BIT,               /* 16 bpp graphics              */
        GT_24BIT,               /* 24 bpp graphics              */
        GT_32BIT,               /* 24 bpp word aligned          */
        GT_LAST
};

int setmod; /* true: do not only check, but set also */
int verbose;/* true: report all test results. False: small table */
int autof;   /* true: virtual size is GGI_AUTO, false: virtual = visible */

void checkoneresolution(int j, int x, int y, int xvir, int yvir )
{
	ggi_mode suggest;
	int res;

	if (verbose){
		printf ("\n %s:",
			3+GRAPHTYPE(defaultgraphtypes[j]));
	}
	/*printf("Checking %d %d %d %d ", x,y,xvir,yvir);*/

	res = ggiCheckGraphMode(vis,x,y,xvir,yvir,
				defaultgraphtypes[j],
				&suggest,
				NULL);
	if (res==0){
		if (setmod){
			if (ggiSetGraphMode(vis,x,y,xvir,yvir,
					    defaultgraphtypes[j])){
				/* failed */
				printf("S");
			} else {
				printf("+");
			}
		} else {
			printf("+");
		}
	} else {
		if (verbose){
			printf("- (%dx%d*%dx%d %s)",
			       suggest.visible.x,suggest.visible.y,
			       suggest.virt.x,suggest.virt.y,
			       3+GRAPHTYPE(suggest.graphtype));
		} else {
			if (suggest.graphtype > defaultgraphtypes[j]) {
				printf(">");
			} else if (suggest.graphtype < defaultgraphtypes[j] ){
				printf("<");
			} else if (((suggest.visible.x < x)&&
				   (suggest.visible.y <= y))||
				   ((suggest.visible.x <= x)&&
				    (suggest.visible.y < y))){
				printf("^");
			} else if (((suggest.visible.x > x)&&
				   (suggest.visible.y >= y))||
				   ((suggest.visible.x >= x)&&
				   (suggest.visible.y > y))){
				printf("v");
			} else if ((suggest.visible.x == x) &&
				   (suggest.visible.y == y) &&
				   ((suggest.virt.x != xvir) ||
				    (suggest.virt.y != yvir))){
				printf("*");
			} else {
				printf("o");
			}
		}
	}
	if (verbose){
		/*printf("\n");*/
	} else {
		printf("\t");
	}
}

void checkstandardresolutions(void)
{
	int i,j,x,y;

	if (!verbose){
		printf("\t\t");
		for (j=0;defaultgraphtypes[j]!=GT_LAST;j++){
			printf ("%s\t",3+GRAPHTYPE(defaultgraphtypes[j]));
		}
		printf("\n");
	}
	
	i=0;j=0;
	x=defaultresolutions[i].x;
	y=defaultresolutions[i].y;
	while (x!=0){
		printf("%s%dx%d%s",verbose?"Checking ":"",
		       x,y,
		       verbose?" :":"  \t");
		for (j=0;defaultgraphtypes[j]!=GT_LAST;j++){
			/*printf ("Depth %d [%s]:  ",
			  j,GRAPHTYPE(defaultgraphtypes[j]));*/
			if (autof){
				checkoneresolution(j,x,y,GGI_AUTO,GGI_AUTO);
			} else {
				checkoneresolution(j,x,y,x,y);
			}
		}
		i++;
		x=defaultresolutions[i].x;
		y=defaultresolutions[i].y;
		/*if (!verbose)*/{
			printf("\n");
		}

	}
}

void usage(char * s)
{
	printf("Usage: %s [options]\n"
	       "     available options:\n"
	       "     -s: set mode (not only check mode)\n"
	       "     -v: verbose\n"
	       "     -a: give a virtual size of GGI_AUTO (instead of virt=vis)\n"
	       "     planned options:\n"
	       , s);
	printf("The verbose output includes for each failing mode the\n"
	       "suggested mode, the short output gives a table containig\n"
	       " + for possible modes\n"
	       " > for 'higher' graphtypes (=bitdepth) suggested,\n"
	       " < for lower graphtypes\n"
	       " ^ for lower x AND y resolution\n"
	       " v for higher x AND y resolution\n"
	       " * for different virtual size (broken GGI_AUTO handling??)\n"
	       " o failing for other reasons.\n");
	printf(" S checking the mode succeeded, but setting it failed.\n"
	       "   (note that the original mode is set, not the suggested one.)\n");
	printf("$Id: checkmode.c,v 1.4 1998/03/09 19:07:16 msundberg Exp $\n");
	exit(0);
}

int main(int argc, char **argv)
{
	char * myself = argv[0];
	int op;

	int x,y;
	
	setmod=0;
	verbose=0;
	autof=0;

	while ( (op=getopt(argc,argv,"ahsv"))!=EOF ){
		switch(op){
		case 's': /*fprintf(stderr,"Setting modes also");*/ 
			setmod=1;
			break;
		case 'v': 
			verbose=1;
			break;
		case 'a':
			autof=1;
			break;
		case 'h': usage(myself); /* never return */ break;

		default: printf("Unknown option %c\n",op);
			usage(myself);
		}
	}

	
	/*usage(myself);*/

	if (ggiInit() != 0) {
		fprintf(stderr, "%s: unable to initialize libggi, exiting.\n",
			argv[0]);
		exit(1);
	}
	if ((vis=ggiOpen(NULL)) == NULL) {
		fprintf(stderr,
			"%s: unable to open default visual, exiting.\n",
			argv[0]);
		exit(1);
	}
	ggiSetInfoFlags(vis, GGIFLAG_ASYNC);
	
	checkstandardresolutions();

        ggiClose(vis);

        ggiExit();

	return (0);
}




