/* 
   Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#ifdef _AIX
int setitimer (int Which, const struct itimerval *value, struct itimerval *ovaue);
#endif

#include <signal.h>

#include "ggi-dl.h"
#include "Xvisual.h"

/* 
 * We are guaranteed to have been checked already...
 */

int GGIdlinit(ggi_visual *vis,const char *args)
{
	struct Xhooks *xinfo;

	DPRINT("X-lib starting.\n");
	VIS_PRIVATE(vis)=xinfo=malloc(sizeof(struct Xhooks));
	if (VIS_PRIVATE(vis)==NULL)
		return GGI_DL_ERROR;
	
	xinfo->display=NULL;
	xinfo->ximage =NULL;
	xinfo->cmap=0;
	xinfo->cmap_first=256;
	xinfo->cmap_last=0;
	xinfo->window=0;
	xinfo->gc=0;
#ifndef NO_SHM
	xinfo->have_shm=0;
	xinfo->shminfo.shmid=-1;
#endif
	xinfo->XLibLock=0;
	xinfo->viswidth=0;
	xinfo->visheight=0;
	xinfo->isasync=1;  /* Yes, this SHOULD be initialized to 1. */
	xinfo->relptr=0;
	xinfo->relptr_keymask = RELPTR_KEYINUSE;
	xinfo->colormap=NULL;
	
	/* We do not use the XLibLock here, as there is no concurrent thread yet. */
	
	DPRINT("X-lib wants display %s.\n",args);
	xinfo->display = XOpenDisplay(args);
	if (xinfo->display == NULL)
		return GGI_DL_ERROR;

	DPRINT("X-lib has display %s.\n",args);

#ifndef NO_SHM
	if (XShmQueryExtension(xinfo->display)) {
		
		DPRINT("Enabling use of XSHM extension.\n");
		xinfo->have_shm=1;
	}
#endif
            
	VIS_SELECT_FD(vis)=VIS_FD(vis)=ConnectionNumber(xinfo->display);

	xinfo->screen = DefaultScreen (xinfo->display);
	DPRINT("X-lib has screen.\n");
	DPRINT("X-lib fully up.\n");

	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 10) {
		/* Has mode management */
		vis->opdisplay->flush=GGIflush;
		vis->opdisplay->getmode=GGIgetmode;
		vis->opdisplay->setmode=GGIsetmode;
		vis->opdisplay->checkmode=GGIcheckmode;
	}

	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 15) {
		/* Has Event management */
		vis->opdisplay->eventpoll=GGIeventpoll;
		vis->opdisplay->eventread=GGIeventread;
	}

	return GGI_DL_OPDISPLAY;
}

int GGIdlcleanup(ggi_visual *vis)
{
	struct Xhooks *xinfo;

    ggiFlush(vis);
    
	/* prevents a soon-to-be unloaded function from being called */
	if (!(VIS_FLAGS(vis) & GGIFLAG_ASYNC))
		_mansync_stop(vis);

	/* FIXME !!! */
	if ((xinfo=VIS_PRIVATE(vis)) != NULL)
	{
		if (xinfo->cmap)
			XFreeColormap(xinfo->display,xinfo->cmap);
#ifndef NO_SHM
		if (xinfo->shminfo.shmid!=-1) {
			XShmDetach(xinfo->display, &(xinfo->shminfo));
			if (VIS_FB_LINEAR(vis)!=NULL) {
				shmdt(VIS_FB_LINEAR(vis));
			}
		}
#endif
		if (xinfo->ximage)
			XDestroyImage(xinfo->ximage);
		if (xinfo->gc)
			XFreeGC(xinfo->display,xinfo->gc);
        if (xinfo->window)
			XDestroyWindow(xinfo->display,xinfo->window);
		if (xinfo->display)
			XCloseDisplay(xinfo->display);
		free(VIS_PRIVATE(vis));
	}

	return EOK;
}

#include "ggidlinit.h"
