/* 
   Graphics library for GGI. Pixels.

   Copyright (C) 1998 Marcus Sundberg - marcus@ggi-project.org

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "ggi-dl.h"
#include "debug.h"
#include "Xvisual.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

/*******************************/
/* draw/get/put a single pixel */
/*******************************/

int GGIdrawpixel(ggi_visual *vis,int x,int y)
{
	XDrawPoint(XLIB_PRIV(vis)->display,
			   XLIB_PRIV(vis)->window,
			   XLIB_PRIV(vis)->gc,
			   x, y); 

	XLIB_DOSYNC(vis);
	return EOK;
}

int GGIputpixel(ggi_visual *vis,int x,int y,ggi_pixel col)
{ 
	XSetForeground(XLIB_PRIV(vis)->display,
				   XLIB_PRIV(vis)->tempgc,
				   col);
	XDrawPoint(XLIB_PRIV(vis)->display,
			   XLIB_PRIV(vis)->window,
			   XLIB_PRIV(vis)->tempgc,
			   x, y);

	XLIB_DOSYNC(vis);
	return EOK;
}

/* Hack to fall back when shm is not working. */
static int	geterror;

static int errorhandler (Display * disp, XErrorEvent * event)
{
	if (event->error_code == BadMatch)
		geterror=1;
	return 0;
}

int GGIgetpixel(ggi_visual *vis,int x,int y,ggi_uint *pixel)
{
	XImage *ximg;
	int	(*olderrorhandler) (Display *, XErrorEvent *);
	
	XLIB_DOSYNC(vis);

	if (XLIB_PRIV(vis)->XLibLock++!=0) 
		{XLIB_PRIV(vis)->XLibLock--;return -EAGAIN;}	/* FIXME use spinlocks here !*/
	geterror=0;
	olderrorhandler = XSetErrorHandler(errorhandler);
		/* This will cause a BadMatch error when the window is
		   iconified or on another virtual screen... */
	ximg = XGetImage(XLIB_PRIV(vis)->display, XLIB_PRIV(vis)->window, x, y,
			  1, 1, AllPlanes, ZPixmap);
	XSetErrorHandler(olderrorhandler);
	XLIB_PRIV(vis)->XLibLock--;	/* FIXME Spinlocks ... */
	*pixel=0;
	if (geterror)
		return -1;
	memcpy(pixel, ximg->data, VIS_FB_BPP(vis)/8);
	XDestroyImage(ximg);

	return EOK;
}
