/* 
   Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <signal.h>

#include <ggi/ggi_commands.h>
#include "ggi-dl.h"
#include "debug.h"

void _ignore_SIGBUS(int unused)
{ 
	signal(SIGBUS,_ignore_SIGBUS);
	sleep(1);	/* Ignore the SIGBUSes */
}

/* 
 * We are guaranteed to have been checked already...
 */

int GGIdlinit(ggi_visual *vis,const char *args)
{
	VIS_FD(vis)=VIS_SELECT_FD(vis)=open(args,O_RDWR);
	if (VIS_FD(vis)<0)
		return GGI_DL_ERROR;

	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 10) {
		/* Has mode management */
		vis->opdisplay->getmode=GGIgetmode;
		vis->opdisplay->setmode=GGIsetmode;
		vis->opdisplay->checkmode=GGIcheckmode;
	}

	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 15) {
		/* Has Event management */
		vis->opdisplay->eventpoll=GGIeventpoll;
		vis->opdisplay->eventread=GGIeventread;
	}

	vis->op2d->setsplitline=GGIsetsplitline;
	vis->op2d->setorigin=GGIsetorigin;

	/* temporary hack to do away with the SIGBUS ... */
	signal(SIGBUS,_ignore_SIGBUS);

	return GGI_DL_OPDISPLAY|GGI_DL_OP2D;
}

int GGIdlcleanup(ggi_visual *vis)
{
	if (VIS_FD(vis) > -1)
		close(VIS_FD(vis));

	return EOK;
}
		
#include "ggidlinit.h"
