/* 
   Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ggi-dl.h"
#include "debug.h"

/**********************/
/* draw/get/put a box */
/**********************/
int GGIdrawbox(ggi_visual *vis,int x,int y,int width,int length)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiDrawBox(*vislist++,x,y,width,length);

	return err;
}

int GGIputbox(ggi_visual *vis,int x,int y,int width,int length,void *buffer)
{ 
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiPutBox(*vislist++,x,y,width,length,buffer);

	return err;
}

int GGIgetbox(ggi_visual *vis,int x,int y,int width,int length,void *buffer)
{ 
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	err=ggiGetBox(*vislist,x,y,width,length,buffer);

	return err;
}

int GGIfillscreen(ggi_visual *vis)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiFillscreen(*vislist++);

	return err;
}

int GGIputc(ggi_visual *vis,int x,int y,char c)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiPutc(*vislist++,x,y,c);

	return err;
}


int GGIputs(ggi_visual *vis,int x,int y,const char *str)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiPuts(*vislist++,x,y,str);

	return err;
}

int GGIdrawhline(ggi_visual *vis,int x,int y,int w)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiDrawHLine(*vislist++,x,y,w);

	return err;
}

int GGIdrawline(ggi_visual *vis,int x1,int y1,int x2,int y2)
{ 
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiDrawLine(*vislist++,x1,y1,x2,y2);

	return err;
}

int GGIdrawvline(ggi_visual *vis,int x,int y,int height)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiDrawVLine(*vislist++,x,y,height);

	return err;
}

int GGIsetgammamap(ggi_visual *vis,int start,int len,ggi_color *colormap)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiSetGammaMap(*vislist++,start,len,colormap);

	return err;
}

int GGIgetgammamap(ggi_visual *vis,int start,int len,ggi_color *colormap)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	err=ggiGetGammaMap(*vislist,start,len,colormap);

	return err;
}


int GGIgetgamma(ggi_visual *vis,ggi_float *r,ggi_float *g,ggi_float *b)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	err=ggiGetGamma(*vislist,r,g,b);

	return err;
}

int GGIsetgamma(ggi_visual *vis,ggi_float r,ggi_float g,ggi_float b)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiSetGamma(*vislist++,r,g,b);

	return err;
} 

ggi_pixel GGImapcolor(ggi_visual *vis,ggi_color *col)
{
	ggi_uint err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	err=ggiMapColor(*vislist,col);

	return err;
}

int GGIunmappixel(ggi_visual *vis,ggi_uint pixel,ggi_color *col)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	err=ggiUnmapPixel(*vislist,pixel,col);

	return err;
}

int GGIsetpalvec(ggi_visual *vis,int start,int len,ggi_color *colormap)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiSetPaletteVec(*vislist++,start,len,colormap);

	return err;
}

int GGIgetpalvec(ggi_visual *vis,int start,int len,ggi_color *colormap)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	err=ggiGetPaletteVec(*vislist,start,len,colormap);

	return err;
}
  
int GGIsetorigin(ggi_visual *vis,int x,int y)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiSetOrigin(*vislist++,x,y);

	return err;
}

int GGIsetsplitline(ggi_visual *vis,int y)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiSetSplitline(*vislist++,y);

	return err;
}

ggi_event_mask GGIeventpoll(ggi_visual_t vis,ggi_event_mask mask,
			    struct timeval *t)
{
	ggi_event_mask err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	err=ggiEventPoll(*vislist,mask,t);

	return err;
}

int GGIeventread(ggi_visual_t vis,ggi_event *ev,ggi_event_mask mask)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	err=ggiEventRead(*vislist,ev,mask);

	return err;
}


int GGIdrawpixel(ggi_visual *vis,int x,int y)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiDrawPixel(*vislist++,x,y);

	return err;
}

int GGIputpixel(ggi_visual *vis,int x,int y,ggi_uint col)
{ 
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiPutPixel(*vislist++,x,y,col);

	return err;
}

int GGIgetpixel(ggi_visual *vis,int x,int y,ggi_uint *col)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	err=ggiGetPixel(*vislist,x,y,col);

	return err;
}

int GGIputhline(ggi_visual *vis,int x,int y,int w,void *buffer)
{ 
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiPutHLine(*vislist++,x,y,w,buffer);

	return err;
}

int GGIgethline(ggi_visual *vis,int x,int y,int w,void *buffer)
{ 
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	err=ggiGetHLine(*vislist,x,y,w,buffer);

	return err;
}

int GGIputvline(ggi_visual *vis,int x,int y,int height,void *buffer)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiPutVLine(*vislist++,x,y,height,buffer);

	return err;
}

int GGIgetvline(ggi_visual *vis,int x,int y,int height,void *buffer)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	err=ggiGetVLine(*vislist,x,y,height,buffer);

	return err;
}

int GGIcopybox(ggi_visual *vis,int x,int y,int w,int h,int nx,int ny)
{
	int err=EOK;
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) err=ggiCopyBox(*vislist++,x,y,w,h,nx,ny);

	return err;
}
