/* 
   Graphics library for GGI.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>

#include "ggi-dl.h"
#include "debug.h"

/* 
 * We are guaranteed to have been checked already...
 */

int GGIdlinit(ggi_visual *vis,const char *args)
{
	char hlpbuffer[1024],*curr;
	ggi_visual_t *vislist;
	int x,visanz;

	VIS_FD(vis)=-1;

	if (! args) {
	
		fprintf(stderr,"display-multi needs the real targets as arguments.\n");
		return -1;
	}

	strncpy(hlpbuffer,args,sizeof(hlpbuffer));
	hlpbuffer[sizeof(hlpbuffer)-1]='\0';
	
	for(visanz=1,curr=hlpbuffer;*curr;curr++) 
		if (*curr==';') { visanz++; *curr='\0'; }

	VIS_PRIVATE(vis)=vislist=malloc((visanz+1)*sizeof(ggi_visual_t));
	if (! vislist) {
	
		fprintf(stderr,"display-multi: Out of memory allocating vislist.\n");
		return -1;
	}
	

	for(x=0,curr=hlpbuffer;x<visanz;x++,curr+=strlen(curr)+1) {

		DPRINT("here %s %d %d",curr,visanz,x);
		if (! (vislist[x]=ggiOpen(curr,NULL)) ) {

			fprintf(stderr,"display-multi: Opening of target %s failed.\n",curr);
			free(vislist);
			return -1;
		}
		/* Make a shared GC for the children ... shudder ... */
		free(VIS_GC(vislist[x])); /* FIXME - this will be freed multiple times ! */
		VIS_GC(vislist[x])=VIS_GC(vis);
	}
	vislist[x]=NULL;	/* Terminate it */
	
	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 10) {
		/* Has mode management */
		vis->opdisplay->getmode=GGIgetmode;
		vis->opdisplay->setmode=GGIsetmode;
		vis->opdisplay->checkmode=GGIcheckmode;
                vis->opdisplay->flush=GGIflush;                                 
	}
	if (GGI_VERSION_FUNCS(vis->opdisplay->version) >= 15) {
		/* Has event management */
		vis->opdisplay->eventpoll=GGIeventpoll;
		vis->opdisplay->eventread=GGIeventread;
	}

	vis->op2d->fillscreen=GGIfillscreen;

	vis->op2d->setsplitline=GGIsetsplitline;
	vis->op2d->setorigin=GGIsetorigin;

	vis->op2d->putc=GGIputc;
	vis->op2d->puts=GGIputs;

	vis->op2d->drawpixel=GGIdrawpixel;
	vis->op2d->putpixel=GGIputpixel;
	vis->op2d->getpixel=GGIgetpixel;

	vis->op2d->drawhline=GGIdrawhline;
	vis->op2d->puthline=GGIputhline;
	vis->op2d->gethline=GGIgethline;

	vis->op2d->drawvline=GGIdrawvline;
	vis->op2d->putvline=GGIputvline;
	vis->op2d->getvline=GGIgetvline;

	vis->op2d->drawbox=GGIdrawbox;
	vis->op2d->putbox=GGIputbox;
	vis->op2d->getbox=GGIgetbox;

	vis->op2d->drawline=GGIdrawline;
	vis->op2d->copybox=GGIcopybox;

	vis->opcolor->getgamma=GGIgetgamma;                                  
	vis->opcolor->setgamma=GGIsetgamma;                                  
	vis->opcolor->getgammamap=GGIgetgammamap;                            
	vis->opcolor->setgammamap=GGIsetgammamap;                            

	vis->opcolor->mapcolor=GGImapcolor;                                  
	vis->opcolor->unmappixel=GGIunmappixel;                              
	vis->opcolor->setpalvec=GGIsetpalvec;                                
	vis->opcolor->getpalvec=GGIgetpalvec;                                

	return GGI_DL_OPDISPLAY;
}

int GGIdlcleanup(ggi_visual *vis)
{
	ggi_visual_t *vislist=VIS_PRIVATE(vis);
	while(*vislist) ggiClose(*vislist++);

	return EOK;
}

#include "ggidlinit.h"
