/* 
   Graphics library for GGI.

   Copyright (C) 1998 Andreas Beck - andreas.beck@ggi-project.org
                      Rodolphe Ortalo

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* FIXME: Should we adjust parent clipping for "get" calls ? */

#include "ggi-dl.h"
#include "debug.h"
#include "SubVisual.h"

#define IS_SAVING_GC	ggi_pixel _foreground, _background; \
			int _l,_t,_r,_b;

#define SETUP_AND_SAVE_GC \
		do {	_foreground = VIS_GC(hook->parent)->fg_color; \
			_background = VIS_GC(hook->parent)->bg_color; \
			_l = VIS_GC(hook->parent)->cliptl.x; \
			_t = VIS_GC(hook->parent)->cliptl.y; \
			_r = VIS_GC(hook->parent)->clipbr.x; \
			_b = VIS_GC(hook->parent)->clipbr.y; \
			VIS_GC(hook->parent)->fg_color = VIS_GC(vis)->fg_color; \
			VIS_GC(hook->parent)->bg_color = VIS_GC(vis)->bg_color; \
			VIS_GC(hook->parent)->cliptl.x=VIS_GC(vis)->cliptl.x+hook->position.x; \
		   	VIS_GC(hook->parent)->cliptl.y=VIS_GC(vis)->cliptl.y+hook->position.y; \
		   	VIS_GC(hook->parent)->clipbr.x=VIS_GC(vis)->clipbr.x+hook->position.x; \
		   	VIS_GC(hook->parent)->clipbr.y=VIS_GC(vis)->clipbr.y+hook->position.y; \
			/* check limits... (in case leaving visual) */ \
			if (VIS_GC(hook->parent)->clipbr.x > hook->botright.x) \
				VIS_GC(hook->parent)->clipbr.x = hook->botright.x; \
			if (VIS_GC(hook->parent)->clipbr.y > hook->botright.y) \
				VIS_GC(hook->parent)->clipbr.y = hook->botright.y; \
			/* ensure update in KGI */ \
			VIS_GC(hook->parent)->version++; \
		} while (0);

#define RESTORE_GC \
		do {	VIS_GC(hook->parent)->fg_color = _foreground; \
			VIS_GC(hook->parent)->bg_color = _background; \
			VIS_GC(hook->parent)->cliptl.x= _l; \
		   	VIS_GC(hook->parent)->cliptl.y= _t; \
		   	VIS_GC(hook->parent)->clipbr.x= _r; \
		   	VIS_GC(hook->parent)->clipbr.y= _b; \
			/* ensure update in KGI */ \
			VIS_GC(hook->parent)->version++; \
		} while (0);


/*
**	Functions undergoing a translation
**	with GC saving/restoring
*/

/* NB: Here the src is not directly translated... */
int GGIcrossblit(ggi_visual *src, int sx, int sy, int sw, int sh,
		 ggi_visual *vis, int dx, int dy, int dw, int dh)
{
	SubHook *hook=VIS_PRIVATE(vis); // Destination is subvisual
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiCrossBlit(src, sx, sy, sw, sh, hook->parent,
		      dx+hook->position.x, dy+hook->position.y, dw, dh);
	RESTORE_GC;
	return err;
}

int GGIdrawbox(ggi_visual *vis,int x,int y,int width,int length)
{
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiDrawBox(hook->parent,x+hook->position.x,y+hook->position.y,width,length);
	RESTORE_GC;
	return err;
}

int GGIputbox(ggi_visual *vis,int x,int y,int width,int length,void *buffer)
{ 
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiPutBox(hook->parent,x+hook->position.x,y+hook->position.y,width,length,buffer);
	RESTORE_GC;
	return err;
}

int GGIgetbox(ggi_visual *vis,int x,int y,int width,int length,void *buffer)
{ 
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiGetBox(hook->parent,x+hook->position.x,y+hook->position.y,width,length,buffer);
	RESTORE_GC;
	return err;
}

int GGIfillscreen(ggi_visual *vis)
{
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiDrawBox(hook->parent, hook->position.x,hook->position.y,
					hook->botright.x-hook->position.x,
					hook->botright.y-hook->position.y);
	RESTORE_GC;
	return err;
}

int GGIputc(ggi_visual *vis,int x,int y,char c)
{
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiPutc(hook->parent,x+hook->position.x,y+hook->position.y,c);
	RESTORE_GC;
	return err;
}


int GGIputs(ggi_visual *vis,int x,int y,const char *str)
{
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiPuts(hook->parent,x+hook->position.x,y+hook->position.y,str);
	RESTORE_GC;
	return err;
}

int GGIdrawhline(ggi_visual *vis,int x,int y,int w)
{
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiDrawHLine(hook->parent,x+hook->position.x,y+hook->position.y,w);
	RESTORE_GC;
	return err;
}

int GGIdrawline(ggi_visual *vis,int x1,int y1,int x2,int y2)
{ 
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiDrawLine(hook->parent,x1+hook->position.x,y1+hook->position.y,
			x2+hook->position.x,y2+hook->position.y);
	RESTORE_GC;
	return err;
}

int GGIdrawvline(ggi_visual *vis,int x,int y,int height)
{
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiDrawVLine(hook->parent,x+hook->position.x,y+hook->position.y,height);
	RESTORE_GC;
	return err;
}

int GGIdrawpixel(ggi_visual *vis,int x,int y)
{
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiDrawPixel(hook->parent,x+hook->position.x,y+hook->position.y);
	RESTORE_GC;
	return err;
}

int GGIputpixel(ggi_visual *vis,int x,int y,ggi_uint col)
{ 
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiPutPixel(hook->parent,x+hook->position.x,y+hook->position.y,col);
	RESTORE_GC;
	return err;
}

int GGIgetpixel(ggi_visual *vis,int x,int y,ggi_uint *col)
{
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiGetPixel(hook->parent,x+hook->position.x,y+hook->position.y,col);
	RESTORE_GC;
	return err;
}

int GGIputhline(ggi_visual *vis,int x,int y,int w,void *buffer)
{ 
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiPutHLine(hook->parent,x+hook->position.x,y+hook->position.y,w,buffer);
	RESTORE_GC;
	return err;
}

int GGIgethline(ggi_visual *vis,int x,int y,int w,void *buffer)
{ 
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiGetHLine(hook->parent,x+hook->position.x,y+hook->position.y,w,buffer);
	RESTORE_GC;
	return err;
}

int GGIputvline(ggi_visual *vis,int x,int y,int height,void *buffer)
{
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiPutVLine(hook->parent,x+hook->position.x,y+hook->position.y,height,buffer);
	RESTORE_GC;
	return err;
}

int GGIgetvline(ggi_visual *vis,int x,int y,int height,void *buffer)
{
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiGetVLine(hook->parent,x+hook->position.x,y+hook->position.y,height,buffer);
	RESTORE_GC;
	return err;
}

int GGIcopybox(ggi_visual *vis,int x,int y,int w,int h,int nx,int ny)
{
	SubHook *hook=VIS_PRIVATE(vis);
	int err;
	IS_SAVING_GC;

	SETUP_AND_SAVE_GC;
	err = ggiCopyBox(hook->parent,x+hook->position.x,y+hook->position.y,w,h,nx,ny);
	RESTORE_GC;
	return err;
}

/*
**	Simple "call parent" functions
*/

int GGIsetgammamap(ggi_visual *vis,int start,int len,ggi_color *colormap)
{
	SubHook *hook=VIS_PRIVATE(vis);

	/* Should we allow that ? */
	return ggiSetGammaMap(hook->parent,start,len,colormap);
}

int GGIgetgammamap(ggi_visual *vis,int start,int len,ggi_color *colormap)
{
	SubHook *hook=VIS_PRIVATE(vis);

	return ggiGetGammaMap(hook->parent,start,len,colormap);
}


int GGIgetgamma(ggi_visual *vis,ggi_float *r,ggi_float *g,ggi_float *b)
{
	SubHook *hook=VIS_PRIVATE(vis);

	return ggiGetGamma(hook->parent,r,g,b);
}

int GGIsetgamma(ggi_visual *vis,ggi_float r,ggi_float g,ggi_float b)
{
	SubHook *hook=VIS_PRIVATE(vis);

	/* Should we allow that ? */
	return ggiSetGamma(hook->parent,r,g,b);
} 

ggi_pixel GGImapcolor(ggi_visual *vis,ggi_color *col)
{
	SubHook *hook=VIS_PRIVATE(vis);

	return ggiMapColor(hook->parent,col);
}

int GGIunmappixel(ggi_visual *vis,ggi_uint pixel,ggi_color *col)
{
	SubHook *hook=VIS_PRIVATE(vis);

	return ggiUnmapPixel(hook->parent,pixel,col);
}

int GGIsetpalvec(ggi_visual *vis,int start,int len,ggi_color *colormap)
{
	SubHook *hook=VIS_PRIVATE(vis);

	/* Should we allow that ? */
	return ggiSetPaletteVec(hook->parent,start,len,colormap);
}

int GGIgetpalvec(ggi_visual *vis,int start,int len,ggi_color *colormap)
{
	SubHook *hook=VIS_PRIVATE(vis);

	return ggiGetPaletteVec(hook->parent,start,len,colormap);
}

ggi_event_mask GGIeventpoll(ggi_visual_t vis,ggi_event_mask mask,
			    struct timeval *t)
{
	SubHook *hook=VIS_PRIVATE(vis);

	return ggiEventPoll(hook->parent,mask,t);
}

int GGIeventread(ggi_visual_t vis,ggi_event *ev,ggi_event_mask mask)
{
	SubHook *hook=VIS_PRIVATE(vis);

	return ggiEventRead(hook->parent,ev,mask);
}

/*
**	Functions that are deactivated !
*/ 

int GGIsetorigin(ggi_visual *vis,int x,int y)
{
	return -1;
}

int GGIsetsplitline(ggi_visual *vis,int y)
{
	return -1;
}


