/* 
   Graphics library for GGI. Horizontal lines.

   Copyright (C) 1998 Marcus Sundberg - marcus@ggi-project.org
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ggi-dl.h"
#include "vgavisual.h"

/**********************************/
/* draw/get/put a horizontal line */
/**********************************/

int GGIdrawhline(ggi_visual *vis,int x,int y,int w)
{
	/* Clipping */
	if (y<(VIS_GC(vis)->cliptl.y) || y>=(VIS_GC(vis)->clipbr.y)) return 0;
	if (x< (VIS_GC(vis)->cliptl.x)) {
		int diff=(VIS_GC(vis)->cliptl.x)-x;
		x+=diff;
		w-=diff;
	}
	if (x+w>(VIS_GC(vis)->clipbr.x)) {
		w=(VIS_GC(vis)->clipbr.x)-x;
	}
	if (w>0) {
		vga_setcolor(VIS_GC_FGCOLOR(vis));
		vga_drawline(x,y,x+w-1,y);
	}
	return 0;
}

int _GGIdrawhline(ggi_visual *vis,int x,int y,int w)
{
	vga_setcolor(VIS_GC_FGCOLOR(vis));
	vga_drawline(x,y,x+w-1,y);
	return 0;
}

int GGIputhline(ggi_visual *vis,int x,int y,int w,void *buffer)
{ 
	/* Clipping */
	if (y<(VIS_GC(vis)->cliptl.y) || y>=(VIS_GC(vis)->clipbr.y)) return 0;
	if (x< (VIS_GC(vis)->cliptl.x)) {
		int diff=(VIS_GC(vis)->cliptl.x)-x;
		x+=diff;
		w-=diff;
		buffer=((char *)buffer)+diff;
	}
	if (x+w>(VIS_GC(vis)->clipbr.x)) {
		w=(VIS_GC(vis)->clipbr.x)-x;
	}
	if (w>0) {
		if (SVGA_PRIV(vis)->ismodex && x%4 != 0) {
			for (;x%4 != 0;x++) {
				ggiPutPixel(vis, x, y, *((char*)buffer));
				buffer = ((char*)buffer)+1;
			}
			for (;w%4 != 0;w--)
				ggiPutPixel(vis, x, y, *((char*)buffer+w));	
		}
        vga_drawscansegment(buffer, x, y, w*VIS_FB_BPP(vis)/8);
	}
	return EOK;
}

int GGIgethline(ggi_visual *vis,int x,int y,int w,void *buffer)
{ 
	if (x<0||y<0||x+w>VIS_FB_WIDTH(vis)||y>=VIS_FB_HEIGHT(vis)) return -1;

    vga_getscansegment(buffer, x, y, w*VIS_FB_BPP(vis)/8);

	return 0;
}
