/* 
 * LibGGI SVGAlib target README:
 * 
 * Programs need SUID root. You really should use KGI :-)
 * 
 * Uses raw keyboard mode! That means "a crash might render the console
 * unusable. (but shouldn't)." Of course, SVGAlib is inherently dangerous,
 * so this is nothing new.
 * 
 * When using the SVGAlib target, LibGGI programs will always segfault at
 * the end. This happens because SVGAlib likes to attach itself to the end
 * of programs to do some 'cleanup'. The problem is, SVGAlib would be
 * unloaded along with the target by ggiClose(), and hence it won't be there
 * at the end of the program (cannot access memory). However, this behavior
 * normally shouldn't cause any problems.
 *
 * This behaviour is avoided by doing a signal(SIGSEGV,_exit) in
 * ggiClose(). This is only done if LIBGGI_DEBUG is != 0.
 * To completely disable this, define ALWAYS_SEGFAULT_ON_EXIT below.
 * 
 * Likewise, SVGAlib will also install its own signal handlers, so beware.
 * 
 * See source for miscellaneous stuff. 
 * The SVGAlib target now supports banked and modeX modes.
 * 
 * ------------------------------------------------------------------------
 *      Copyright (C) 1998 Marcus Sundberg <marcus@ggi-project.org>
 *      Copyright (C) 1998 Steve Cheng <elmert@ipoline.com>
 * ------------------------------------------------------------------------
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU Library General Public License as 
 *	published by the Free Software Foundation; either version 2, or 
 *	(at your option) any later version.
 * 
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Library General Public License for more details.
 *
 *	You should have received a copy of the GNU Library General Public 
 *	License	along with this program; see the file COPYING.LGPL .  
 *	If not, write to the Free Software Foundation, 675 Mass Ave, 
 *	Cambridge, MA 02139, USA.
 *
 * ------------------------------------------------------------------------
 */ 

#include <vga.h>
#include <vgakeyboard.h>
#include <vgamouse.h>

/*
 * ------------------------------------------------------------------------
 * Things you might want to change.
 * ------------------------------------------------------------------------
 */ 

/* Splitline support.
   This is experimental and not guaranteed to work on all
   cards.
   
   It has been successfully tested on CL5430 and Matrox
   Millenium (I and II) in all modes supported by SVGAlib.

   However this does not mean it will work on your card.
   Defining this might cause anything to happen to your
   computer.
   
   Now that you've been warned you're welcome to test it
   and send me bugreports and fixes. Just be sure to sync
   all disks before testing this. */
/*
#define WANT_DANGEROUS_SPLITLINE
*/

/* The sample rate of your mouse.
   Most people need not change this. */

#define SAMPLE_RATE MOUSE_DEFAULTSAMPLERATE

/* Never prevent apps from segfaulting on exit */
/*
#define ALWAYS_SEGFAULT_ON_EXIT
*/

/*
 * ------------------------------------------------------------------------
 * Do not change anything below this line.
 * ------------------------------------------------------------------------
 */ 

typedef struct {
	short          w, h;
	ggi_graphtype  gt;
	short          bpp;
	char           ismodex;
} svga_ggimode;

struct svga_priv {
	unsigned short CRTI,CRTD;
	int            nocols;
	ggi_color     *colormap;
	svga_ggimode  *availmodes;
	char           islinear;
	char           ismodex;
	char           isbanked;
};

#define SVGA_PRIV(vis) ((struct svga_priv*)VIS_PRIVATE(vis))

void _GGIkeyboardhandler(int scancode, int press);
void _GGImousehandler(int button, int dx, int dy, int dz, int drx, int dry, int drz);
