/* 
   Graphics library for GGI. Mode management.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
   Copyright (C) 1998 MenTaLboY - mentalboy@geocities.com

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <time.h>
#include <ctype.h>

#include "TIvisual.h"

#include "ggi-dl.h"
#include "../common/evqueue.inc"

static int _ggiGetAndQueueEvents(ggi_visual_t vis)
{
	ggi_event ev;
	int key, status;
	struct TIhooks *tiinfo;

	status = 0;

	tiinfo = (struct TIhooks *)VIS_PRIVATE(vis);

	lock_ncurses();
	set_term(tiinfo->scr);
	key = wgetch(stdscr);
	switch (key) {
		case ERR: {
			status = -1;
		} break;
#if ( NCURSES_MOUSE_VERSION == 1 )		
		case KEY_MOUSE: {
			ggi_mode *mode;
			MEVENT mev;

			mode = VIS_MODE(vis);

			getmouse(&mev);
			ev.any.focus = 0;
			ev.any.device = mev.id;
			ev.any.time = time(NULL);
			ev.pmove.effect = 
				( ( mev.bstate | BUTTON_SHIFT ) ? mkShift : 0 );
			ev.pmove.effect |=
				( ( mev.bstate | BUTTON_CTRL ) ? mkCtrl : 0 );
			ev.pmove.effect |=
				( ( mev.bstate | BUTTON_ALT ) ? mkAlt : 0 );

			switch (mev.bstate) {
				case BUTTON1_PRESSED:
				case BUTTON2_PRESSED:
				case BUTTON3_PRESSED:
				case BUTTON4_PRESSED: {
					ev.any.type = evPtrButtonPress;
					ev.any.size = sizeof(ggi_pbutton_event);
					switch (mev.bstate) {
						case BUTTON1_PRESSED: ev.pbutton.button = 1;
						case BUTTON2_PRESSED: ev.pbutton.button = 4;
						case BUTTON3_PRESSED: ev.pbutton.button = 2;
						case BUTTON4_PRESSED: ev.pbutton.button = 8;
					}
					ev.pbutton.state = 1;
				} break;
				case BUTTON1_RELEASED:
				case BUTTON2_RELEASED:
				case BUTTON3_RELEASED:
				case BUTTON4_RELEASED: {
					ev.any.type = evPtrButtonRelease;
					ev.any.size = sizeof(ggi_pbutton_event);
					switch (mev.bstate) {
						case BUTTON1_RELEASED: ev.pbutton.button = 1;
						case BUTTON2_RELEASED: ev.pbutton.button = 4;
						case BUTTON3_RELEASED: ev.pbutton.button = 2;
						case BUTTON4_RELEASED: ev.pbutton.button = 8;
					}
					ev.pbutton.state = 0;
				} break;
				default: {
					ev.any.type = evPtrAbsolute;
					ev.any.size = sizeof(ggi_pmove_event);
					ev.pmove.x = mev.x * mode->dpp.x;
					ev.pmove.y = mev.y * mode->dpp.y;
					ev.pmove.wheel = mev.z;
				}
			}
			ev.key.normal = ev.key.effect;
			_ggiEvQueueAdd(vis, &ev);
			status = 1;
		} break;
#endif /* NCURSES_MOUSE_VERSION */
		case '\033': {
			timeout(1);
			key = wgetch(stdscr);
			timeout(0);
			if ( key != ERR ) {
				ev.any.size = sizeof(ggi_key_event);
				ev.any.focus = 0;
				ev.any.device = 0;
				ev.any.time = time(NULL);
				ev.any.type = evKeyPress;
				ev.key.effect = mkAlt;
				ev.key.sym = translate_key(key, &ev.key.effect);
				ev.key.normal = ev.key.effect;
				ev.key.sticky = ev.key.locked = 0;
				ev.key.code = key;
				_ggiEvQueueAdd(vis, &ev);
				ev.any.size = sizeof(ggi_key_event);
				ev.any.focus = 0;
				ev.any.device = 0;
				ev.any.time = time(NULL);
				ev.any.type = evKeyRelease;
				ev.key.effect = mkAlt;
				ev.key.sym = translate_key(key, &ev.key.effect);
				ev.key.normal = ev.key.effect;
				ev.key.sticky = ev.key.locked = 0;
				ev.key.code = key;
				_ggiEvQueueAdd(vis, &ev);
				status = 2;
				break;
			}
		} 
		default: {
			ev.any.size = sizeof(ggi_key_event);
			ev.any.focus = 0;
			ev.any.device = 0;
			ev.any.time = time(NULL);
			ev.any.type = evKeyPress;
			ev.key.effect = 0;
			ev.key.sym = translate_key(key, &ev.key.effect);
			ev.key.normal = ev.key.effect;
			ev.key.sticky = ev.key.locked = 0;
			ev.key.code = key;
			_ggiEvQueueAdd(vis, &ev);
			ev.any.size = sizeof(ggi_key_event);
			ev.any.focus = 0;
			ev.any.device = 0;
			ev.any.time = time(NULL);
			ev.any.type = evKeyRelease;
			ev.key.effect = 0;
			ev.key.sym = translate_key(key, &ev.key.effect);
			ev.key.normal = ev.key.effect;
			ev.key.sticky = ev.key.locked = 0;
			ev.key.code = key;
			_ggiEvQueueAdd(vis, &ev);
			status = 2;
		}
	}
	unlock_ncurses();
	
	return status;
}

/************************** GGI Functions ***************************/
/* Event Handling */
ggi_event_mask GGIeventpoll(ggi_visual_t vis, ggi_event_mask mask,
			    struct timeval *t)
{
	fd_set fds;
	int err, fd;
	ggi_event_mask evmask;

if (t==NULL) {
	DPRINT("GGIeventpoll(%p,0x%.8x,NULL)\n",vis,mask);
} else {
	DPRINT("GGIeventpoll(%p,0x%.8x,{%d,%d})\n",vis,mask,t->tv_sec,t->tv_usec);
}

	evmask = _ggiEvQueueSeen(vis, mask);
	if ( evmask != 0 )  
		return evmask;

	fd = fileno(( (struct TIhooks *)VIS_PRIVATE(vis) )->f_in);

	do {
		if ( _ggiGetAndQueueEvents(vis) <= 0 ) {
			FD_ZERO(&fds);
			FD_SET(fd, &fds);

			err = select(fd+1, &fds, NULL, NULL, t);
			if ( err < 0 ) 
				return 0;
			if (FD_ISSET(fd, &fds)) {
				if ( _ggiGetAndQueueEvents(vis) > 0 ) {
					evmask = _ggiEvQueueSeen(vis, mask);
				}
			}
		} else {
			evmask = _ggiEvQueueSeen(vis, mask);
		}
	} while ( evmask == 0 && t == NULL );

	return evmask;
}

int GGIeventread(ggi_visual_t vis, ggi_event *ev, ggi_event_mask mask)
{
	/* Block if we don't have anything queued... */
	GGIeventpoll(vis, mask, NULL);
	return _ggiEvQueueRelease(vis, ev, mask);
}
