/* 
   Graphics library for GGI.

   Copyright (C) 1998 Marcus Sundberg - marcus@ggi-project.org
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ggi-dl.h"
#include "vgaglvis.h"


/**********************/
/* draw/get/put a box */
/**********************/
int GGIdrawbox(ggi_visual *vis,int x,int y,int width,int height)
{
	if (y< (VIS_GC(vis)->cliptl.y)) {
		int diff=(VIS_GC(vis)->cliptl.y)-y;
		y     +=diff;
		height-=diff;
	}
	if (y+height>(VIS_GC(vis)->clipbr.y)) {
		height=(VIS_GC(vis)->clipbr.y)-y;
	}
	if (x< (VIS_GC(vis)->cliptl.x)) {
		int diff=(VIS_GC(vis)->cliptl.x)-x;
		x     +=diff;
		width-=diff;
	}
	if (x+width>(VIS_GC(vis)->clipbr.x)) {
		width=(VIS_GC(vis)->clipbr.x)-x;
	}

	if (height<=0 || width<=0 ) 
		return EOK;

	gl_fillbox(x,y,width,height,VIS_GC_FGCOLOR(vis));
	return EOK;
}

int GGIputbox(ggi_visual *vis,int x,int y,int width,int height,void *buffer)
{ 
	int rowadd = (VIS_FB_BPP(vis)*width+7)/8/sizeof(uint8);
	ggi_coord  tl=VIS_GC(vis)->cliptl, br=VIS_GC(vis)->clipbr;

	/* Pre-Clipping, so we don't call PutHline without need. */
	if (y< tl.y) {
		int diff=tl.y-y;
		y     +=diff;
		height-=diff;
		buffer = (uint8 *)(buffer)+rowadd*diff; 
	}
	if (y+height>=br.y) {
		height=br.y-y;
	}
	if (x< tl.x) {
		int diff=tl.x-x;
		x     +=diff;
		width -=diff;
		buffer = (uint8 *)(buffer)+diff;
		}
	if (x+width>=br.x) {
		width=br.x-x;
	}
	if (height>0&&width>0)
		gl_putbox(x,y,width,height,buffer);
	return EOK;
}

int GGIgetbox(ggi_visual *vis,int x,int y,int width,int height,void *buffer)
{ 
	/* Unclipped - get */
	gl_getbox(x,y,width,height,buffer);
	return EOK;
}
