/* 
   Graphics library for GGI.

   Copyright (C) 1998 Marcus Sundberg - marcus@ggi-project.org
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ggi-dl.h"
#include "vgaglvis.h"

/********************************/
/* draw a vertical line */
/********************************/

int GGIdrawvline(ggi_visual *vis,int x,int y,int height)
{
	/* Clipping */
	if (x< (VIS_GC(vis)->cliptl.x) ||
	    x>=(VIS_GC(vis)->clipbr.x)) return EOK;
	if (y< (VIS_GC(vis)->cliptl.y)) {
		int diff=(VIS_GC(vis)->cliptl.y)-y;
		y     +=diff;
		height-=diff;
	}
	if (y+height>(VIS_GC(vis)->clipbr.y)) {
		height=(VIS_GC(vis)->clipbr.y)-y;
	}
	if (height<1)
		return EOK;
	
	gl_line(x,y,x,y+height-1,VIS_GC_FGCOLOR(vis));
	return EOK;
}

int _GGIdrawvline(ggi_visual *vis,int x,int y,int height)
{
	gl_line(x,y,x,y+height-1,VIS_GC_FGCOLOR(vis));	
	return EOK;
}

int GGIputvline(ggi_visual *vis,int x,int y,int height,void *buffer)
{
        /* Clipping */
	if (x< (VIS_GC(vis)->cliptl.x) ||
		x>=(VIS_GC(vis)->clipbr.x)) return EOK;
	if (y< (VIS_GC(vis)->cliptl.y)) {
		int diff=(VIS_GC(vis)->cliptl.y)-y;
		y     +=diff;
		height-=diff;
		buffer =((char *)buffer)+diff;
	}
	if (y+height>(VIS_GC(vis)->clipbr.y)) {
		height=(VIS_GC(vis)->clipbr.y)-y;
	}
	
	if (height>0) {
        gl_putbox(x, y, 1, height, buffer);
	}
	return EOK;
}

int GGIgetvline(ggi_visual *vis,int x,int y,int height,void *buffer)
{
	if (x<0||y<0||x>=VIS_FB_WIDTH(vis)||y+height>VIS_FB_HEIGHT(vis))
		return -1;

	gl_getbox(x, y, 1, height, buffer);
	return EOK;
}
