/* 
   Graphics library for GGI. 
   Palette driver for XF86.

   Copyright (C) 1997 Steve Cheng <elmert@ipoline.com>
   Copyright (C) 1997 Andreas Beck - becka@uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ggi-dl.h"
#include "Xvisual.h"
#include <ggi/ggi_commands.h>

ggi_uint GGImapcolor(ggi_visual *vis,ggi_color *col)
{
	int closest=0;
	int i;
	unsigned short closedist=0xffff,dist,d;
/* Don't use XAllocColor for this. */
/*	if (VIS_MODE(vis)->graphtype != GT_8BIT) {
		XColor xcol;
		
		xcol.red=col->r;
		xcol.green=col->g;
		xcol.blue=col->b;

		XAllocColor(XLIB_PRIV(vis)->display, XLIB_PRIV(vis)->cmap, &xcol);
		
		return xcol.pixel;
	} else */{
/* Do closest-match-mapping. I use a || x ||_1 norm here. This is faster
 * and the measure for "closest" istn't well defined for RGB anyway.
 */
#undef ABS
#define ABS(a,b) ( (a>b) ? a-b : b-a )
		for (i=0;i<XLIB_PRIV(vis)->nocols;i++) {
			dist=ABS(col->r,XLIB_PRIV(vis)->colormap[i].r);
			if (dist >= closedist) continue;
			d   =ABS(col->g,XLIB_PRIV(vis)->colormap[i].g);
			if (d>dist) dist=d;
			if (dist >= closedist) continue;
			d   =ABS(col->b,XLIB_PRIV(vis)->colormap[i].b);
#undef ABS
			if (d>dist) dist=d;
			if (dist >= closedist) continue;
			closest=i;
			if (dist==0) break;
			closedist=dist;
		}
		DPRINT("colmap: chose %x,%x,%x (%d) for %x,%x,%x.\n",
			   XLIB_PRIV(vis)->colormap[closest].r,
			   XLIB_PRIV(vis)->colormap[closest].g,
			   XLIB_PRIV(vis)->colormap[closest].b,closest,
			   col->r,col->g,col->b
			);
	}
	return closest;
}

/* Map pixel to color
 */
int GGIunmappixel(ggi_visual *vis,ggi_pixel pixel,ggi_color *col)
{
	XColor xcol;

	if (col==NULL)
		return -1;
	
	xcol.pixel = pixel;
	XQueryColor(XLIB_PRIV(vis)->display, XLIB_PRIV(vis)->cmap, &xcol);
	
	col->r = xcol.red;
	col->g = xcol.green;
	col->b = xcol.blue;

	return EOK;	
}

int GGIsetpalvec(ggi_visual *vis, int start, int len,ggi_color *colormap) {
	int i;
	XColor xcol;
	struct Xhooks *xhook=VIS_PRIVATE(vis);

	if (colormap==NULL || start+len >XLIB_PRIV(vis)->nocols)
		return -1;
                        
	memcpy(XLIB_PRIV(vis)->colormap+start,colormap,len*sizeof(ggi_color));

	if (xhook->XLibLock++!=0)
		{xhook->XLibLock--;return -EAGAIN;}     /* FIXME use spinlocks here !*/

	for(i = start; i<start+len; i++) {
		xcol.red  =XLIB_PRIV(vis)->colormap[i].r;
		xcol.green=XLIB_PRIV(vis)->colormap[i].g;
		xcol.blue =XLIB_PRIV(vis)->colormap[i].b;
		xcol.pixel=i;
		xcol.flags= DoRed | DoGreen | DoBlue ;
		XStoreColor(xhook->display, xhook->cmap,&xcol);
	}

	XF86DGAInstallColormap(xhook->display, xhook->screen, xhook->cmap);

	xhook->XLibLock--;      /* FIXME Spinlocks ... */

	return 0;
}

int GGIgetpalvec(ggi_visual *vis,int start,int len,ggi_color *colormap)
{
	if (colormap==NULL || start+len >XLIB_PRIV(vis)->nocols)
		return -1;

	memcpy(colormap,XLIB_PRIV(vis)->colormap+start,len*sizeof(ggi_color));

	return 0;
}
