/* 
 *  Graphics library for GGI. Mode management.
 *
 *  Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
 *
 *  If you do any modifications, I would like you to send diffs to me
 *  to allow for collecting a more and more complete set of drivers and
 *  to improve existing ones.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "internal.h"

/************************** GGI Functions ***************************/

/* Event Handling */
ggi_event_mask ggiEventPoll(ggi_visual_t vis,ggi_event_mask mask,
			    struct timeval *t)
{ return vis->opdisplay->eventpoll(vis,mask,t); }

int ggiEventRead(ggi_visual_t vis,ggi_event *ev,ggi_event_mask mask)
{ return vis->opdisplay->eventread(vis,ev,mask); }

void ggiSetEventMask(ggi_visual_t vis, ggi_event_mask evm)
{ vis->evmask = evm; }

ggi_event_mask ggiGetEventMask(ggi_visual_t vis)
{ return vis->evmask; }


#ifdef __sgi

int ggiEventPut(ggi_event_buffer *buf, ggi_event *ev)
{
	register ggi_uint start = buf->write;
	register ggi_uint end = start + ev->size;

	ASSERT(ev->any.type != evNothing);

	if (end <= (buf->read + buf->size)) {

		buf->write = end;
		
		DEBUG2("copy event to %i-%i", start, end);

		if ((end <= buf->size) || (buf->size <= start)) {

			DEBUG2("single burst");
			memcpy((uint8 *)buf->buf + (start & (buf->size-1)), ev,
				ev->size);
		} else {

			register ggi_uint cnt = buf->size - start;

			DEBUG2("dual burst");
			memcpy((uint8 *)buf->buf + start, ev, cnt);
			memcpy(buf->buf, ((uint8 *)ev) + cnt, ev->size - cnt);
		}
		return EOK;

	} else {

		NOTICE2("event buffer overflow");
		buf->overflow = 1;
		return -ENOMEM;
	}
}

void ggiEventSkip(ggi_event_buffer *buf)
{
	ggi_uint size = *((uint8 *)((uint8 *)buf->buf + buf->read));

	if (buf->read < buf->write) {

		DEBUG2("skip event from %i size %i", buf->read, size);
		buf->read += size;
		if (buf->read >= buf->size) {

			buf->read -= buf->size;
			buf->write -= buf->size;
		}
		DEBUG2("buf->read %i, buf->write %i", buf->read, buf->write);
	}
}

int ggiEventCopy(ggi_event_buffer *buf, ggi_event *ev)
{
	register ggi_uint start = buf->read;
	ggi_uint size = *((uint8 *)((uint8 *)buf->buf + start)); 
	register ggi_uint end = start + size;

	if (start < buf->write) {

		DEBUG2("copy event from %i-%i", start, end);
		
		if ((end <= buf->size) || (buf->size <= start)) {

			NOTICE2("single burst");
			memcpy(ev, (uint8 *)buf->buf + (start & (buf->size-1)), size);

		} else {

			register ggi_uint cnt = buf->size - start;

			NOTICE2("dual burst");
			memcpy(ev, (uint8 *)buf->buf + start, cnt);
			memcpy(((uint8 *)ev) + cnt, buf->buf, size - cnt);
		}
		return EOK;

	} else {

		NOTICE2("no event available");
		return -EAGAIN;	/* no event yet */
	}
}

#endif /* #ifdef __sgi */
