/* 
   Graphics library for GGI. Mode management.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define NEED_GRAPHTYPE_SYMS
#include <ctype.h>

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <sys/ioctl.h>

#include "internal.h"
#include <ggi/ggi_commands.h>

ggi_mode ggiDefaultMode={
	1,                      /* 1 frame [???] */
	{GGI_AUTO,GGI_AUTO},    /* Default size */
	{GGI_AUTO,GGI_AUTO},    /* Virtual */
	{300,220},              /* mm don't care */
	GGI_AUTO,               /* Mode */
	{GGI_AUTO,GGI_AUTO},    /* Font size */
	/*0*/
};
                                                                                                                                                  
static void _ggiCheck4Defaults(ggi_mode *tm)
{
#define DOCHECK(what)	if (tm->what == GGI_AUTO) tm->what=ggiDefaultMode.what

	DOCHECK(visible.x);
	DOCHECK(visible.y);
	DOCHECK(virt.x);
	DOCHECK(virt.y);
	DOCHECK(dpp.x);
	DOCHECK(dpp.y);
	DOCHECK(graphtype);
}

/********************************/
/* set any mode (text/graphics) */
/********************************/
int ggiSetMode(ggi_visual *vis,ggi_mode *tm)
{ 
	int retval;
	_ggiLock(vis);

	DPRINT("setmode trying (vis %dx%d virt %dx%d)\n",
		tm->visible.x,tm->visible.y,tm->virt.x,tm->virt.y);
	_ggiCheck4Defaults(tm);
               
	retval=vis->opdisplay->setmode(vis,tm);
	if (retval<0) {
		fprintf(stderr,"libggi: Failed to set requested mode\n");
		goto ender;
	}

	DPRINT("setorigin\n");
	ggiSetOrigin(vis,0,0);
	DPRINT("setsplit\n");
	ggiSetSplitline(vis,tm->visible.y*tm->dpp.y);
	DPRINT("set GC\n");
	ggiSetGCForeground(vis,0);
	ggiSetGCBackground(vis,0);
	ggiSetGCClipping(vis,0,0,tm->virt.x,tm->virt.y);
	DPRINT("fillscreen\n");
	ggiFillscreen(vis);
	DPRINT("setmode success (vis %dx%d virt %dx%d)\n",
		tm->visible.x,tm->visible.y,tm->virt.x,tm->virt.y);

ender:

	DPRINT("setmode - done\n");
	_ggiUnlock(vis);
	return retval;
}

/**********************************/
/* check any mode (text/graphics) */
/**********************************/
int ggiCheckMode(ggi_visual *vis,ggi_mode *tm)
{
	_ggiCheck4Defaults(tm);
	return vis->opdisplay->checkmode(vis,tm);
}

/************************/
/* get the current mode */
/************************/
int ggiGetMode(ggi_visual *vis,ggi_mode *tm)
{
	return vis->opdisplay->getmode(vis,tm);
}

/******************/
/* set a textmode */
/******************/
int ggiSetTextMode(ggi_visual *vis,int cols,int rows,
				   int vcols,int vrows,
				   int fontsizex,int fontsizey,
				   ggi_graphtype mode)
{
  ggi_mode tim={	1,		/* 1 frame [???] */
			{80,25},	/* Default size */
        	        {80,25},
        	        {300,220},
                        GT_TEXT16,
                	{9,14},
                	/*0*/           };
                        
  if (fontsizex<4||fontsizex>32) return -EINVAL;	/* Maybe some card can do it ... */
  if (fontsizey<4||fontsizey>32) return -EINVAL;

  tim.virt.x=vcols;
  tim.virt.y=vrows;
  tim.visible.x=cols;
  tim.visible.y=rows;
  tim.dpp.x=fontsizex;
  tim.dpp.y=fontsizey;
  tim.graphtype=mode;

  return ggiSetMode(vis,&tim);
}

/*************************/
/* check a text mode */
/*************************/
int ggiCheckTextMode(ggi_visual *vis,int cols,int rows,
				     int vcols,int vrows,
				     int fontsizex,int fontsizey,
				     ggi_graphtype mode,
				     ggi_mode *md,...)
{
  int rc;
  ggi_mode tim={	1,		/* 1 frame [???] */
			{80,25},	/* Default size */
        	        {80,25},
        	        {300,220},
                        GT_TEXT16,
                	{9,14},
                	/*0*/           };
                        
  if (fontsizex<4||fontsizex>32) return -EINVAL;	/* Maybe some card can do it ... */
  if (fontsizey<4||fontsizey>32) return -EINVAL;

  tim.virt.x=vcols;
  tim.virt.y=vcols;
  tim.visible.x=cols;
  tim.visible.y=rows;
  tim.dpp.x=fontsizex;
  tim.dpp.y=fontsizey;

  rc=ggiCheckMode(vis,&tim);
  if (md) *md=tim;	/* give back the mode if asked for. */
  return rc;
}

/***********************/
/* set a graphics mode */
/***********************/
int ggiSetGraphMode(ggi_visual *vis,int xsize,int ysize,
		    int xvirtual,int yvirtual,ggi_graphtype type)
{
  ggi_mode tim={ 1,
		 { 800, 600},
	         {1024,1024},
	         { 300, 220},
	         GT_8BIT,
	         {1,1},
	         /*0*/      };
  tim.visible.x=xsize;
  tim.visible.y=ysize;
  tim.virt.x=xvirtual;
  tim.virt.y=yvirtual;                       
  tim.dpp.x = tim.dpp.y = 1;
  tim.graphtype=type;

  return ggiSetMode(vis,&tim);
}

/*************************/
/* check a graphics mode */
/*************************/
int ggiCheckGraphMode(ggi_visual *visual,int xsize,int ysize,
		      int xvirtual,int yvirtual,ggi_graphtype type,
		      ggi_mode *md,...)
{
  int rc;
  ggi_mode tim={ 1,
		 { 800, 600},
	         {1024,1024},
	         { 300, 220},
	         GT_8BIT,
	         {1,1},
	         /*0*/      };
  tim.visible.x=xsize;
  tim.visible.y=ysize;
  tim.virt.x=xvirtual;
  tim.virt.y=yvirtual;                       
  tim.dpp.x = tim.dpp.y = 1;
  tim.graphtype=type;

  rc=ggiCheckMode(visual,&tim);
  if (md) *md=tim;	/* give back the mode if asked for. */
  return rc;
}

/*
 * Setup the vis->info.fb stuff, from vis->mode
 *
 * Provides initial guesses for:
 *     VIS_FB_WIDTH
 *     VIS_FB_HEIGHT
 *     VIS_FB_BPP	bits/pixel 
 *     VIS_FB_BPC	bits for color/pixel
 *     VIS_FB_BPA	bits for alpha/pixel 
 *
 * NOTE: These can be overriden by each fb driver.
 */
int _ggiSetupMode(ggi_visual *vis)
{
	int size;

	VIS_FB_WIDTH(vis)=VIS_MODE(vis)->virt.x;
	VIS_FB_HEIGHT(vis)=VIS_MODE(vis)->virt.y;

	size=VIS_FB_WIDTH(vis)*VIS_FB_HEIGHT(vis);

	VIS_FB_BPP(vis)=0;
	VIS_FB_BPC(vis)=0;
	VIS_FB_BPA(vis)=0;

	switch (VIS_MODE(vis)->graphtype) {
		case GT_TEXT16:
		case GT_TEXT32:
			size *= (VIS_MODE(vis)->graphtype==GT_TEXT16) ? 2 : 4;
			VIS_FB_BPP(vis) = (VIS_MODE(vis)->graphtype == GT_TEXT16) ? 16 : 32;
			VIS_FB_BPC(vis) = VIS_FB_BPP(vis);
			break;
		case GT_1BIT:
			size /= 8;
			VIS_FB_BPP(vis)=1;
			VIS_FB_BPC(vis)=1;
			break;
		case GT_4BIT:
			size /= 2;
			VIS_FB_BPP(vis)=4;
			VIS_FB_BPC(vis)=4;
			break;
		case GT_8BIT:
			VIS_FB_BPP(vis)=8;
			VIS_FB_BPC(vis)=8;
			break;
		case GT_15BIT:
			size *= 2;
			VIS_FB_BPP(vis)=16;
			VIS_FB_BPC(vis)=15;
			VIS_FB_BPA(vis)=1;
			break;
		case GT_16BIT:
			size *= 2;
			VIS_FB_BPP(vis)=16;
			VIS_FB_BPC(vis)=16;
			break;
		case GT_24BIT:
			size *= 3;
			VIS_FB_BPP(vis)=24;
			VIS_FB_BPC(vis)=24;
			break;
		case GT_32BIT:
			size *= 4;
			VIS_FB_BPP(vis)=32;
			VIS_FB_BPC(vis)=32;
			break;
		default:
			fprintf(stderr,"Unknown bit size used!\n");
			return -1;
	}

	/* round up to a multible of 64K */

	return (size + 0xffff) & ~0xffff;
}

/*******************/
/* print mode      */
/*******************/

#define _MODESTRING "%dx%d#%dx%d D%dx%d F %d [%s]",m->visible.x,m->visible.y,m->virt.x,m->virt.y,m->dpp.x,m->dpp.y,m->frames,GRAPHTYPE(m->graphtype)


int ggiPrintMode(ggi_mode * m)
{
	return printf(_MODESTRING);
}

int ggiSPrintMode(char * s, ggi_mode * m)
{
	return sprintf(s,_MODESTRING);
}

int ggiFPrintMode(FILE * s, ggi_mode * m)
{
	return fprintf(s,_MODESTRING);
}



/*
** parse a (commandline) mode string to a mode
** (c) Hartmut Niemann 1998 
**
** definition:
**
** X x Y # XVIRT x YVIRT D dpp.x x dpp.y F frames (T) [bpp] + XOFF + YOFF
** X x Y # XVIRT x YVIRT D dpp.x x dpp.y F frames [graphtype] + XOFF + YOFF
**
** Everything (!) can be omitted, all ommitted values default to GGI_AUTO,
** with the exception of a textmode with unspecified size which defaults
** to TEXT16.
** Whitespace is ignored.
**
** Please give full specs until the drivers understand GGI_AUTO:
**  320x200#320x200D1x1[8]
**
** this is is a full (AFAIK) superset of what the X geometry parser supports. 
**
** examples include:
** 640x480           just a visible size
** 640x480#640x960   same size, but double-height virtual screen
** #1024x768         only virtual size defined
**
** 80x40T            (default-bitsized) text mode with 80x40 characters
** #x100T            text mode with 100 virtual lines
** #x100[T]          same
** 640x400[8]        640x400 at 8 bits per pixel
** 640x400[8bpp]  640x400[8BIT]  640x400[GT_8BIT] all the same 
** [T32] [32T] [GT_TEXT32] [TEXT32]   32bit text mode
**
** currently unsupported by GGI, but prepared in the parser:
** 640x400+2+2       position window at upper left corner 
**                   like X does it.
**
** The only way of specifying GGI_AUTO is omitting the parameter;
** especially -1 (which is the internal representation) is not supported.
** (But you can do  x#x[]+-6 to specify only y position :-)
**
** Returncode:
**  0 on success, i.e. the string was correct.
**    ignored characters, like GT_ and a position information
**    (even a double T in [T16T]) do not make it fail
**    a missing ] from the bitdepth field is ignored, no failure
** -1 if there is text that can not be parsed. 
**    This text is printed to stderr.
**    All paramters parsed so far are written into m!
** So m contains all paramters that have been successfully parsed.
** For most applications there will be no need for testing parsemode
** for failure.
*/


int ggiParseMode(const char * s, ggi_mode * m)
{
	/* FIXME ! no range checking yet. */
	int bitdepth=0;
	int textmode=0;
	int negative=0;  /* negative flag for positions */

	int xposition=0;
	int yposition=0;

	*m=ggiDefaultMode;

	/* skipping leading text before the first number can't be easily done*/
	/* think   -geometryx50 : is 50 visible.x or visible.y??*/
	/*
	** while ( (*s != '\000') && (!isdigit(*s)) ){
	**   s++;
	** }
	*/
#define SKIPSPACE   while ( (*s!='\000') && (isspace(*s)) ) s++;
 
/* scan the integer from the string pointer s */
#define SCANINT(x)  SKIPSPACE;\
		    if (isdigit(*s)){\
		       x=*s-'0';\
	               s++;\
		       while (isdigit(*s)){\
			  x = x*10+ (*s) -'0';\
			  s++;\
		       }\
		    }\
		    SKIPSPACE

	/* first number is visible-x: */
	SCANINT(m->visible.x);
	if (*s== 'x'){ /* now for the y */
		s++;
		SCANINT(m->visible.y);
	}
	if (*s == '#'){ /* virtual starts here */
		s++;
		SCANINT(m->virt.x);
		if (*s== 'x'){ /* now for the y */
			s++;
			SCANINT(m->virt.y);
		}
	}
	if (*s == 'D'){ /* dpp starts here */
		s++;
		SCANINT(m->dpp.x);
		if (*s== 'x'){ /* now for the y */
			s++;
			SCANINT(m->dpp.y);
		}
	}
	if (*s == 'F'){ /* frames starts here */
		s++;
		SCANINT(m->frames);
	}
	if (*s=='T'){
		s++;
		textmode=1;
	} 
	if (*s == '['){ /* resolution starts here */
		s++;
		if ((*s=='G')&&(*(s+1)=='T')&&(*(s+2)=='_')){
			s +=3; /* skip a GT_ */
		}
		if ((*s=='T')&&(*(s+1)=='E')&&(*(s+2)=='X')&&(*(s+3)=='T')){
			s +=4; /* TEXT */
			if (textmode) {
				fprintf(stderr,
				     "warning: double Textmode identifier\n");
			}
			textmode=1;
		} else if (*s=='T'){
			s++;
			if (textmode) {
				fprintf(stderr,
				     "warning: double Textmode identifier\n");
			}
			textmode=1;
		}
		SCANINT(bitdepth);
		if ((*s=='B')&&(*(s+1)=='I')&&(*(s+2)=='T')){
			s +=3; /* skip a BIT */
		} else if ((*s=='b')&&(*(s+1)=='p')&&(*(s+2)=='p')){
			s +=3; /* skip bpp */
		} 
		if (*s=='T'){
			s++;
			if (textmode) {
				fprintf(stderr,
				     "warning: double Textmode identifier\n");
			}
			textmode=1;
		}
		if (*s == ']'){
			s++;
		} else {
			fprintf(stderr,"warning: ] missing.");
		}
	}
	if ((*s=='-')||(*s=='+')){ /* x position starts */
		if (*s=='-'){
			negative=1;
		}
		s++;
		SCANINT(xposition);
		if (negative){
			negative=0;
			xposition = - xposition;
		}
		fprintf(stderr,"X position %d ignored.\n",xposition);
	}
       	if ((*s=='-')||(*s=='+')){ /* y position starts */
		if (*s=='-'){
			negative=1;
		}
		s++;
		SCANINT(yposition);
		if (negative){
			negative=0;
			yposition = - yposition;
		}
		fprintf(stderr,"Y position %d ignored.\n",yposition);
	}
	if (*s !='\000'){
		fprintf(stderr,"trailing text %s ignored.\n"
			"parsed mode is ",s);
		ggiFPrintMode(stderr,m);
		fprintf(stderr,"\n");
	}
	
#undef SCANINT
	/*printf("mode: %d %s ",bitdepth,textmode?"T":"");*/
	/* parse bitdepth and textmode to graphtype */
	if (textmode){
		switch (bitdepth){
		case 0:
		case 16:
			m->graphtype=GT_TEXT16; /* default type */
			break;
		case 32:
			m->graphtype=GT_TEXT32;
			break;
		default:
			fprintf(stderr,"warning: unsupported bpp, assuming TEXT16\n");
			m->graphtype=GT_TEXT16;
			break;
		} 
	} else {
		switch (bitdepth){
		case 1:
			m->graphtype=GT_1BIT;
			break;
		case 4:
			m->graphtype=GT_4BIT;
			break;
		case 8:
			m->graphtype=GT_8BIT;
			break;
		case 15:
			m->graphtype=GT_15BIT;
			break;
		case 16:
			m->graphtype=GT_16BIT;
			break;
		case 24:
			m->graphtype=GT_24BIT;
			break;
		case 32:
			m->graphtype=GT_32BIT;
			break;
		case 0:
			m->graphtype=GGI_AUTO;
			break;
		default:
			fprintf(stderr,"warning: unsupported bpp, assuming GGI_AUTO\n");
			m->graphtype=GGI_AUTO;
			break;
		} 
	}
	return (*s=='\000')?0:-1;
}
