/* 
   Graphics library for GGI. Mode management.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <sys/ioctl.h>

#include "internal.h"
#include <ggi/ggi_commands.h>

extern int _ggiExtensionSize;	/* init.c*/

static int _default_error(void)
{
        DPRINT("_default_error() called\n");
	return -1;
}

static int _default_zero(void)
{
	return -1;
}

static void *_default_null(void)	/* currently unused. Might come in handy again ... */
{
	return NULL;
}

/***************************************
 * Creation and destruction of visuals *
 ***************************************/
typedef int   (*__simp_int)(void);
typedef void *(*__simp_ptr)(void);

static void _ggiZapOpDisplay(ggi_visual *vis)
{
	__simp_int *funcarr;	
	int i;

	_ggiCloseDL(vis,GGI_DL_OPDISPLAY);

	if (vis->opdisplay==NULL)
		vis->opdisplay=(struct ggi_visual_opdisplay *)malloc(sizeof(struct ggi_visual_opdisplay));
	vis->opdisplay->version=GGI_VERSION_VISUAL_OPDISPLAY;
	vis->opdisplay->dlhandle=NULL;

	/** Display **/
	funcarr=(__simp_int *)(&vis->opdisplay->null);
	for (i=0;i<GGI_VERSION_FUNCS(vis->opdisplay->version);i++)
		funcarr[i]=_default_error;
}

static void _ggiZapOpColor(ggi_visual *vis)
{
	__simp_int *funcarr;	
	int i;

	_ggiCloseDL(vis,GGI_DL_OPCOLOR);

	if (vis->opcolor==NULL)
		vis->opcolor=(struct ggi_visual_opcolor *)malloc(sizeof(struct ggi_visual_opcolor));
	vis->opcolor->version=GGI_VERSION_VISUAL_OPCOLOR;
	vis->opcolor->dlhandle=NULL;

	/** Color **/
	funcarr=(__simp_int *)(&vis->opcolor->null);
	for (i=0;i<GGI_VERSION_FUNCS(vis->opcolor->version);i++)
		funcarr[i]=_default_error;
	vis->opcolor->mapcolor=(ggi_pixel (*)(ggi_visual *, ggi_color *)) _default_zero;
}

static void _ggiZapOp2D(ggi_visual *vis)
{
	__simp_int *funcarr;	
	int i;

	_ggiCloseDL(vis,GGI_DL_OP2D);

	if (vis->op2d==NULL)
		vis->op2d=(struct ggi_visual_op2d *)malloc(sizeof(struct ggi_visual_op2d));
	vis->op2d->version=GGI_VERSION_VISUAL_OP2D;
	vis->op2d->dlhandle=NULL;
	
	/** 2D **/
	funcarr=(__simp_int *)(&vis->op2d->null);
	for (i=0;i<GGI_VERSION_FUNCS(vis->op2d->version);i++)
		funcarr[i]=_default_error;

	/* Exceptions...
	 */
}

static void _ggiZapOpGC(ggi_visual *vis)
{
	__simp_int *funcarr;	
	int i;

	_ggiCloseDL(vis,GGI_DL_OPGC);

	if (vis->opgc==NULL)
		vis->opgc=(struct ggi_visual_opgc *)malloc(sizeof(struct ggi_visual_opgc));
	vis->opgc->version=GGI_VERSION_VISUAL_OPGC;
	vis->opgc->dlhandle=NULL;
	
	/** GC **/
	funcarr=(__simp_int *)(&vis->opgc->null);
	for (i=0;i<GGI_VERSION_FUNCS(vis->opgc->version);i++)
		funcarr[i]=NULL;
}

void _ggiZapMode(ggi_visual *vis,ggi_uint type)
{
	if (type & GGI_DL_OPDISPLAY)
		_ggiZapOpDisplay(vis);

	if (type & GGI_DL_OPCOLOR)
		_ggiZapOpColor(vis);

	if (type & GGI_DL_OP2D)
		_ggiZapOp2D(vis);

	if (type & GGI_DL_OPGC)
		_ggiZapOpGC(vis);
}

ggi_visual *_ggiNewVisual(void)
{
	ggi_visual *vis;

	vis=(ggi_visual *)calloc(1,sizeof(ggi_visual)+_ggiExtensionSize);

	_ggiLockInit(vis);
	vis->version=GGI_VERSION_VISUAL;
	VIS_FD(vis)=VIS_SELECT_FD(vis)=-1;
	VIS_DB(vis)=NULL;
	VIS_MODE(vis)=(ggi_mode *)calloc(1,sizeof(ggi_mode));
	VIS_GC(vis)=(ggi_gc *)calloc(1,sizeof(ggi_gc));
 	VIS_GC(vis)->version=0;
	VIS_EVMASK(vis)=emAll;
	vis->size=_ggiExtensionSize;

	_ggiZapMode(vis,~0);

	return vis;
}

void _ggiDestroyVisual(ggi_visual *vis)
{
	_ggiCloseDL(vis,~0);
	free(vis->opdisplay);
	free(vis->opcolor);
	free(vis->op2d);
	free(vis->opgc);
	free(VIS_MODE(vis));
	free(VIS_GC(vis));
	free(vis);
}

int _ggiGetVisualFD(ggi_visual *vis)
{
	return vis->fd;
}

ggi_graphtype ggiBPP2GT(int bpp)
{
	switch (bpp) {
	case 1:  return GT_1BIT;
	case 4:  return GT_4BIT;
	case 8:  return GT_8BIT;
	case 15: return GT_15BIT;
	case 16: return GT_16BIT;
	case 24: return GT_24BIT;
	case 32: return GT_32BIT;
	default: return GT_LAST;
	}
}

int ggiGT2BPP(ggi_graphtype graphtype)
{
	switch (graphtype) {
	case GT_1BIT:  return 1;
	case GT_4BIT:  return 4;
	case GT_8BIT:  return 8;
	case GT_15BIT: return 15;
	case GT_16BIT: return 16;
	case GT_24BIT: return 24;
	case GT_32BIT: return 32;
	default:       return -1;
	}
}
