/* 
 *  setup
 *
 *  Copyright (C) 1998 Thomas Tanner. See CREDITS for details.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "lib.h"

/*----------------------------------------------------------------------*/

static void update(ggi_visual_t vis, int change)
{
	ggi2d_funcs *func = GGI2D_FUNC(vis);
	int	dashed = GGI2D_GC_DASHSIZE(vis);
	int	append = GGI2D_GC_APPEND(vis);
	int	textured = GGI2D_GC_TEXTURED(vis);
	int	alpha = GGI2D_GC_ALPHA(vis);
	int	op = GGI2D_GC_OP(vis);

	DPRINT("updating 2d linear-8\n");
	if (alpha == GGI_MAX_ALPHA) {
		func->putPixel = GGINAME(putPixel)[op];
		func->drawPixel = GGINAME(drawPixel)[op];
		func->drawPixelA = GGINAME(drawPixelA)[op];
		func->fillPixelA = GGINAME(fillPixelA);
		if (op == GGI_NOOP) {
			func->scanline = GGINAME(drawHLine_NOOP);
			if (!dashed) func->hLine = GGINAME(drawHLine_NOOP);
		} else if (op == GGI_INVERT) {
			func->scanline = GGINAME(drawHLine_INVERT);
			if (!dashed) func->hLine = GGINAME(drawHLine_INVERT);
		} else {
			func->scanline = GGINAME(scanline);
			if (!dashed) func->hLine = GGINAME(hLine);
		}
	} else {
		func->putPixel = GGINAME(putPixel_a)[op];
		func->drawPixel = GGINAME(drawPixel_a)[op];
		func->drawPixelA = GGINAME(drawPixelA_a)[op];
		func->fillPixelA = GGINAME(fillPixelA_a);
		if (op == GGI_NOOP) {
			func->scanline = GGINAME(drawHLine_NOOP);
			if (!dashed) func->hLine = GGINAME(drawHLine_NOOP);
		} else if (op == GGI_INVERT) {
			func->scanline = GGINAME(drawHLine_INVERT);
			if (!dashed) func->hLine = GGINAME(drawHLine_INVERT);
		} else {
			func->scanline = GGINAME(scanline_a);
			if (!dashed) func->hLine = GGINAME(hLine_a);
		}
	}
	if (!textured && op == GGI_SET && alpha == GGI_MAX_ALPHA) {
		func->fillRect = GGINAME(fillRect_SET);
	}
}

int GGILIB_INIT(ggi_visual_t visual, char *version)
{
	DPRINT("initializing 2d linear-8\n");
	GGI2DregisterUpdateFunc(visual, update, 
		GGI2D_UPD_LINEMODE + 
		GGI2D_UPD_ANTIALIAS + 
		GGI2D_UPD_APPENDMODE + 
		GGI2D_UPD_TEXTUREMODE + 
		GGI2D_UPD_ALPHAMODE + 
		GGI2D_UPD_OPERATOR);
	return EOK;
}

int GGILIB_CLEANUP(ggi_visual_t visual)
{
	return EOK;
}
		
/*----------------------------------------------------------------------*/
