/* 
 *  setup
 *
 *  Copyright (C) 1998 Thomas Tanner. See CREDITS for details.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "lib.h"
#include "util.h"

/*----------------------------------------------------------------------*/

static void	update(ggi_visual_t vis, int change)
{
 	int	dash, antialias, mode;
	ggi2d_funcs *func = GGI2D_FUNC(vis);

	DPRINT("updating 2d stubs\n");
	dash = GGI2D_GC_DASHSIZE(vis) ? 1 : 0;
	antialias = GGI2D_GC_ANTIALIAS(vis) ? 1 : 0;
  	mode = antialias * 2 + dash;
	
	func->hLine = GGINAME(hLine)[dash];
	func->vLine = GGINAME(vLine)[dash];
	func->drawRect = GGINAME(drawRect)[dash];
	func->drawPoly = GGINAME(drawPoly)[dash];
	
	func->fillCircle = GGINAME(fillCircle)[antialias];
	func->fillEllipse = GGINAME(fillEllipse)[antialias];
	func->fillArc = GGINAME(fillArc)[antialias];
	func->bezier = GGINAME(bezier)[antialias];
	func->trapezoid = GGINAME(trapezoid)[antialias];
	func->triangle = GGINAME(triangle)[antialias];
	func->drawPoly = GGINAME(drawPoly)[antialias];
	func->fillPoly = GGINAME(fillPoly)[antialias];
	func->fillPolys = GGINAME(fillPolys)[antialias];
	
	func->line = GGINAME(line)[mode];
	func->linef = GGINAME(linef)[mode];
	func->drawArc = GGINAME(drawArc)[mode];
	func->drawCircle = GGINAME(drawCircle)[mode];
	func->drawEllipse = GGINAME(drawEllipse)[mode];
	func->drawArc = GGINAME(drawArc)[mode];
	func->bezier = GGINAME(bezier)[mode];

	func->drawPixels = GGINAME(drawPixels);
	func->scanlines = GGINAME(scanlines);
	func->fillRect = GGINAME(fillRect);
	func->drawLines = GGINAME(drawLines);
	func->blit = GGINAME(blit);
	func->blitOp = GGINAME(blitOp);
	func->stretchBlit = GGINAME(stretchBlit);
	func->stretchBlitOp = GGINAME(stretchBlitOp); 
}

int GGILIB_INIT(ggi_visual_t visual,const char *version)
{
	DPRINT("initializing 2d stubs\n");
	GGI2DregisterUpdateFunc(visual, update, 
		GGI2D_UPD_ARCMODE + 
		GGI2D_UPD_POLYMODE + 
		GGI2D_UPD_LINEMODE + 
		GGI2D_UPD_ANTIALIAS + 
		GGI2D_UPD_APPENDMODE + 
		GGI2D_UPD_TEXTUREMODE + 
		GGI2D_UPD_ALPHAMODE + 
		GGI2D_UPD_OPERATOR);
	return EOK;
}

int GGILIB_CLEANUP(ggi_visual_t visual)
{
	return EOK;
}
		
/*----------------------------------------------------------------------*/
