/* 
 *  GGI Utility library
 *
 *  Copyright (C) 1998 Uwe Maurer - uwe_maurer@t-online.de
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <internal.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>


static void remove_param(int i,int *argc,char **argv[])
{
	int j;
	char **str;
	if (i>=*argc)
	{
		DPRINT("remove_param: Error!\n");
		return;
	}
	str=&(*argv)[i];
	for (j=i;j<*argc-1;j++)
	{
		*str++=(*argv)[j+1];
	}		
	(*argc)--;
}

static char *get_params(char *buf,const char *param,int *argc,
			char **argv[],char *def)

{
	int i;
	char *str=NULL;
	*buf=0;

	if (def)
	{
		str=strstr(def,param);
		if (str)
		{
			str+=strlen(param);
			while (isspace(*str)) str++;
			i=0;
			while (*str && !isspace(*str)) buf[i++]=*str++;
			buf[i]=0;
			str=buf;
		}
	}
	for (i=1;i<*argc;i++)
	{
		if (strncmp((*argv)[i],param,strlen(param))==0)
		{
			remove_param(i,argc,argv);
			if (i<*argc) 
			{
				str=buf;
				strcpy(buf,(*argv)[i]);
				remove_param(i,argc,argv);
				i--;
			}
			else
			{
				DPRINT("Not enough parameters! %s\n",param);
			}
		}
	}
	return str;
}

ggi_visual_t gguCmdLineVisual(int *argc,char **argv[],char *def)
{
	const char *display="--ggidisplay";
	const char *mode="--ggimode";
	ggi_mode m=ggiDefaultMode;
	ggi_visual_t vis;
	char buf[1000];
	char *str;
	int rc=0;

	DPRINT("ggiCmdLineVisual:\n");
	
	str=get_params(buf,display,argc,argv,def);
	vis= (str) ? ggiOpen(str,NULL) : ggiOpen(NULL);
	if (vis==NULL) return NULL;

	str=get_params(buf,mode,argc,argv,def);
	if (str) 
	{
		gguStringToMode(str,&m);
		rc=ggiSetMode(vis,&m);
	}

	if (!str || rc)
	{
		rc=ggiSetMode(vis,&ggiDefaultMode);
		if (rc)
		{
			ggiPanic("Can't set default mode(LIBGGI_DEFMODE)!\n");
		}
	}

	return vis;
}

int gguModeToString(ggi_mode *mode,char *str)
{
	char buf[100];
	sprintf(buf,"%ix%i",mode->visible.x,mode->visible.y);
	strcpy(str,buf);
	if (mode->visible.x!=mode->virt.x || mode->visible.y!=mode->virt.y)
	{
		sprintf(buf,"[%ix%i]",mode->virt.x,mode->virt.y);
		strcat(str,buf);
	}	
	sprintf(buf,"x%i",gguBPP(mode->graphtype));
	strcat(str,buf);
	return 0;	
}

int gguStringToMode(char *str,ggi_mode *mode)
{
	char *pos;
	ggi_mode m=*mode;

	m.visible.x=strtol(str,&pos,10);
	if (pos==str || *pos!='x') return -1;
	str=pos+1;
	m.visible.y=strtol(str,&pos,10);
	if (pos==str) return -1;
	str=pos+1;
	if (*pos=='[')
	{
		m.virt.x=strtol(str,&pos,10);
		if (pos==str || *pos!='x') return -1;
		str=pos+1;
		m.virt.y=strtol(str,&pos,10);
		if (pos==str || *pos!=']') return -1;
		str=pos+2;
		pos++;
	}
	else m.virt=m.visible;

	if (*pos=='x')
	{
		m.graphtype=gguGraphtype(strtol(str,&pos,10));
		if (pos==str || *pos!=0) return -1;
	}
	else return -1;
			
	*mode=m;
	return 0;
}

ggi_graphtype gguGraphtype(int bpp)
{
	switch(bpp)
	{
		case 8:	return GT_8BIT;
		case 15: return GT_15BIT;
		case 16: return GT_16BIT;
		case 24: return GT_24BIT;
		case 32: return GT_32BIT;
		case 1:	return GT_1BIT;
		case 4:	return GT_4BIT;
	}

	return GT_LAST;
}

int gguBPP(ggi_graphtype gt)
{
	switch(gt)
	{
		case GT_1BIT: return 1;
		case GT_4BIT: return 4;
		case GT_8BIT: return 8;
		case GT_15BIT: return 15;
		case GT_16BIT: return 16;
		case GT_24BIT: return 24;
		case GT_32BIT: return 32;
		default: break;
	}
	return -1;
}

ggi_color gguColor(float r,float g,float b)
{
	ggi_color col;

	if (r>1) r=1; else if (r<0) r=0;
	if (g>1) g=1; else if (g<0) g=0;
	if (b>1) b=1; else if (b<0) b=0;

	col.r=(int)(r*(float)((1<<GGI_COLOR_PRECISION)-1));
	col.g=(int)(g*(float)((1<<GGI_COLOR_PRECISION)-1));
	col.b=(int)(b*(float)((1<<GGI_COLOR_PRECISION)-1));
	return col;
}

ggi_pixellinearbuffer * gguGetPLB(ggi_visual_t vis)
{
	
	while (ggiDBGetBuffer(vis,&GGU(vis)->db)==0 && GGU(vis)->db) 
	{
		if (ggiDBGetLayout(GGU(vis)->db)==blPixelLinearBuffer)
			return ggiDBGetPLB(GGU(vis)->db);
	}
	return NULL;
	
}


