/* 
 *  GGI Utility library
 *
 *  Copyright (C) 1998 Uwe Maurer - uwe_maurer@t-online.de
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <internal.h>
#include <string.h>

#define SHIFT	10
#define MAX ((1<<GGI_COLOR_PRECISION)-1)

ggi_color gguWhite={MAX,MAX,MAX};
ggi_color gguLightGray={MAX*3/4,MAX*3/4,MAX*3/4};
ggi_color gguGray={MAX/2,MAX/2,MAX/2};
ggi_color gguDarkGray={MAX/4,MAX/4,MAX/4};
ggi_color gguBlack={0,0,0};
ggi_color gguRed={MAX,0,0};
ggi_color gguPink={MAX,MAX*7/10,MAX*7/10};
ggi_color gguOrange={MAX,MAX*8/10,0};
ggi_color gguYellow={MAX,MAX,0};
ggi_color gguGreen={0,MAX,0};
ggi_color gguMagenta={MAX,0,MAX};
ggi_color gguCyan={0,MAX,MAX};
ggi_color gguBlue={0,0,MAX};


static int _gguSetPalette(ggi_visual_t vis,int start,int len,ggi_color *col)
{
	usleep(1000000/70); 		// WaitRetrace  
	return	ggiSetPaletteVec(vis,start,len,col);
}

int gguPalFadeOut(ggi_visual_t vis)
{
	ggi_color black[256];
	memset(black,0,sizeof(black));
	return gguPalFadeTo(vis,black);
}

int gguPalFadeTo(ggi_visual_t vis,ggi_color *dest)
{
	
	int pal[256][3];
	int delta[256][3];

	ggi_color col[256];

	int i,j,start=GGU(vis)->palstart,length=GGU(vis)->pallen;
	int steps=GGU(vis)->steps;

	if (ggiGetPaletteVec(vis,start,length,&col[start])) return -1;

	for (i=start;i<start+length;i++) 
	{
		pal[i][0]=(int)col[i].r<<SHIFT;
		pal[i][1]=(int)col[i].g<<SHIFT;
		pal[i][2]=(int)col[i].b<<SHIFT;
		delta[i][0]=((int)(dest[i].r-col[i].r)<<SHIFT)/steps;
		delta[i][1]=((int)(dest[i].g-col[i].g)<<SHIFT)/steps;
		delta[i][2]=((int)(dest[i].b-col[i].b)<<SHIFT)/steps;
	}

	for (i=0;i<steps;i++)
	{
		for (j=start;j<start+length;j++)
		{
			col[j].r=pal[j][0]>>SHIFT;
			col[j].g=pal[j][1]>>SHIFT;
			col[j].b=pal[j][2]>>SHIFT;
			pal[j][0]+=delta[j][0];
			pal[j][1]+=delta[j][1];
			pal[j][2]+=delta[j][2];
		}
		_gguSetPalette(vis,start,length,col+start);
	}

	_gguSetPalette(vis,start,length,dest+start);
	return 0;
}


int gguPalBlend(ggi_color *pal,ggi_color src,ggi_color dest,
		int start,int len)
{
	int delta[3];
	int col[3];
	int i;
	if (len<2) return -1;
	col[0]=(int)src.r << SHIFT;
	col[1]=(int)src.g << SHIFT;
	col[2]=(int)src.b << SHIFT;
	delta[0]=((int)(dest.r-src.r)<<SHIFT)/(len-1);
	delta[1]=((int)(dest.g-src.g)<<SHIFT)/(len-1);
	delta[2]=((int)(dest.b-src.b)<<SHIFT)/(len-1);

	for (i=start;i<start+len;i++)
	{
		pal[i].r=col[0]>>SHIFT;
		pal[i].g=col[1]>>SHIFT;
		pal[i].b=col[2]>>SHIFT;
		col[0]+=delta[0];
		col[1]+=delta[1];
		col[2]+=delta[2];
	}
	return 0;
}

int gguPalMove(ggi_visual_t vis,int step)
{
	int start=GGU(vis)->palstart,length=GGU(vis)->pallen;
	int end=start+length;
	ggi_color col[256];
	ggi_color old[256];
	
	if (step==0) return 0;
	if (ggiGetPaletteVec(vis,start,length,&col[start])) return -1;

	if (step>0)
	{	
		if (step>=length) return -1;
		memcpy(old,&col[start],step*sizeof(ggi_color));
		memmove(&col[start],&col[start+step],
			(length-step)*sizeof(ggi_color));
		memcpy(&col[end-step],old,step*sizeof(ggi_color));
	}
	else
	{
		step=-step;
		if (step>=length) return -1;
		memcpy(old,&col[end-step],step*sizeof(ggi_color));
		memmove(&col[start+step],&col[start],
			(length-step)*sizeof(ggi_color));
		memcpy(&col[start],old,step*sizeof(ggi_color));
	}

	_gguSetPalette(vis,start,length,col+start);
	return 0;
}

int gguPalRange(ggi_visual_t vis,int start,int len)
{
	if (gguBPP(GGU_GT(vis))>8) return -1;
	GGU(vis)->palstart=start;
	GGU(vis)->pallen=len;
	return 0;
}

int gguPalFadeSteps(ggi_visual_t vis,int steps)
{
	if (gguBPP(GGU_GT(vis))>8) return -1;
	if (steps<1) steps=1;
	GGU(vis)->steps=steps;
	return 0;
}

