/*
 *  Input library for GGI.
 *
 *  Copyright (C) 1997 GGI Project
 *
 *  If you do any modifications, I would like you to send diffs to me
 *  to allow for collecting a more and more complete set of drivers and
 *  to improve existing ones.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "gii_structs.h"

/**************************************************************************/

/* Allocating and destroying descriptors */

gii_descriptor *gii_alloc_descriptor(char *name,enum gii_desctype dt);
void            gii_free_descriptor (gii_descriptor *desc);

/* Reading/writing them to config-files */

int gii_write_descriptor(FILE *file,gii_descriptor *desc);
int gii_read_descriptor (FILE *file,gii_descriptor **desc);

/* Checking if controls have been triggered */

uint32 gii_check (ggi_event *event,gii_descriptor *desc);


/**************************************************************************/

/* Working with features. */

int gii_get_feature_name(gii_descriptor *desc,int featurenumber,
                         char *buffer,int maxlen);

/* These are for "rolling your own". Please use gii_desctype>=PRIVATE for this. 
 * You may use the delete-operation on unnecessary features of the standard
 * descriptor types to speed up gii_check a little.
 */

int gii_delete_feature(gii_descriptor *desc,int featurenumber);
int gii_add_feature(gii_descriptor *desc,int featurenumber,
                    enum gii_featuremask type,char *name);

/**************************************************************************/

/* Working with controls. */

int gii_get_control_name(gii_descriptor *desc,int featurenumber,int controlnumber,
                         char *buffer,int maxlen);
int gii_delete_control(gii_descriptor *desc,int featurenumber,int controlnumber);

int gii_autoconfig_control(gii_descriptor *desc,int featurenumber,int controlnumber);


/**************************************************************************/

/* DLL interface for external control-recognizers */

int gii_register_recognizer(control_ops *cops);
