/* ---------------------------------------------------------------------------
**	/dev/graphic special device file driver
** ---------------------------------------------------------------------------
**
**	Copyright (C) 1995-1997
**		by Andreas Beck and Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**	
** ---------------------------------------------------------------------------
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Andreas_Beck
#undef	DEBUG
#define	DEBUG_LEVEL	1

#include <linux/version.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <asm/io.h>

#define	NEED_GRAPHIC_DEFS
#include <ggi/module.h>
#include <ggi/ggi_commands.h>

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)

	#include <asm/uaccess.h>
	#include <linux/poll.h>
        #include <linux/wrapper.h>
#else

	#define copy_from_user(x,ptr,len) memcpy_fromfs(x,ptr,len)
	#define copy_to_user(x,ptr,len) memcpy_tofs(x,ptr,len) 
#endif

#undef IS_CRITICAL
#undef ENTER_CRITICAL
#undef LEAVE_CRITICAL
#define IS_CRITICAL
#define ENTER_CRITICAL
#define LEAVE_CRITICAL

#define TO_PGPROT(x) (protection_map[(x)&0xF])

/*	mmap services
**
** NOTE	The fast_remap_page_range() and fast_unmap_page_range() should
** NOTE	move to the linux memory mapper code. Don't know if they are i386
** NOTE	specific, so we just make sure.
**
** NOTE (jmcc) The general functions zap_page_range and remap_page_range are
** NOTE used under Linux 2.1.31 and higher.
*/
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,31) 

static inline int fast_remap_page_range(struct mm_struct *mm,
	unsigned long addr, unsigned long end, unsigned long dest,
	pgprot_t prot)
{ return remap_page_range(addr,dest,end-addr,prot); }

static inline void fast_unmap_page_range(struct mm_struct *mm,
	unsigned long addr, unsigned long end)
{ zap_page_range(mm,addr,end-addr); }

#else

#ifdef __i386__
/*	The following code has been rewritten from scratch using the Linux
**	kernel (asm/pgtable.h in particular) as a model. This is just to
**	handle memory mapped I/O regions where no pages need to be allocated.
**	when a fault occurs but several mappings need to be changed.
*/

#include <asm/pgtable.h>
	
static inline void fast_remap_page_range(struct mm_struct *mm,
	unsigned long addr, unsigned long end, unsigned long dest,
	pgprot_t prot)
{
	pte_t pte_entry = pte_mkold(mk_pte(dest, prot));
	pte_t *ptr;
	unsigned long next_pmd, beg = addr;;

	next_pmd = addr & PMD_MASK;
  
	while(addr < end) {

		ptr = pte_offset(pmd_offset(pgd_offset(mm,addr),addr),addr);
		next_pmd += PMD_SIZE;

		if (next_pmd > end) {

			next_pmd = end;
		}

		while(addr < next_pmd) {

			*ptr++ = pte_entry;
			pte_val(pte_entry) += PAGE_SIZE;
			addr += PAGE_SIZE;
		}
	}
	flush_tlb_range(mm, beg, end);
}

/*	unmap a range of pages in the mapping table <mm> that are mapped
**	from logical address <addr> to logical address <end>.
*/
static inline void fast_unmap_page_range(struct mm_struct *mm,
	unsigned long addr, unsigned long end)
{
	pte_t *ptr;
	unsigned long next_pmd, beg = addr;
  
	next_pmd = addr & PMD_MASK;
  
	while(addr < end) {

		ptr = pte_offset(pmd_offset(pgd_offset(mm,addr),addr),addr);
		next_pmd += PMD_SIZE;

		if (next_pmd > end) {

			next_pmd = end;
		}

		while(addr < next_pmd) {

			pte_clear(ptr++);
			addr += PAGE_SIZE;
		}
	}
	flush_tlb_range(mm, beg, end);
}

#else	/* #ifdef __i386__ */

	#error Check mapping code for non-i386 architecture!
#endif
#endif /* Linux version >= 2.1.31 */

#define	VM(x)			(vma->vm_ ## x)

static ggi_uint mmapcnt[MAX_NR_CONSOLES];	/* number of mmap() references	*/

/*	unmap a list of mappings that share the same mmio_region.
*/
static inline void graph_unmap_mappings(struct kgi_mmio_mapping *map)
{
	TRACE(struct kgi_mmio_mapping *omap=map);
	IS_CRITICAL;

	ENTER_CRITICAL;

		while (map) {

			TRACE(printk(KERN_DEBUG"%s:%i: unmapping %p %lx-%lx\n",\
				__FILE__,__LINE__, map->mm, map->mstart,
				map->mend));
			fast_unmap_page_range(map->mm, map->mstart, map->mend);
			map->mstart = map->mend = 0;
			map = map->next;
		}

	LEAVE_CRITICAL;

	TRACE(printk(KERN_DEBUG"%s:%i: all pages for %p unmapped\n",\
		__FILE__,__LINE__, omap));
}

/*	graph_save() is called from a bottom half and should do the neccessary
**	unmapping. State is kept in *graph, so we only care about mapping.
*/
static void graph_save(struct kgi_device *dev)
{
	#define	graph	((struct kgi_graphic *) dev)
	ggi_uint i;

	ASSERT(DEV.type == KGI_GRAPHIC_DEV);
	TRACE(printk("%s:%i: unmap device %i\n", __FILE__,__LINE__,DEV.id));

	for (i = 0; i < MAX_NR_MMIO_REGIONS; i++) {

		graph_unmap_mappings(MMIO_MAP[i]);
	}

	#undef graph
}

/*	graph_restore() is called from a bottom half and should do the
**	neccessary mapping stuff. The device mode, CLUT, splitline, in short
**	all state in *dev, is already set, so we only care about mapping.
*/
static void graph_restore(struct kgi_device *dev)
{
	#define	graph	((struct kgi_graphic *) dev)

	ASSERT(DEV.type == KGI_GRAPHIC_DEV);
	TRACE(printk("%s:%i: restore device %i\n",__FILE__,__LINE__, DEV.id));

	#undef graph
}

static unsigned long graph_mmio_nopage(struct vm_area_struct *vma,
	unsigned long addr, int no_share)
{
	struct kgi_mmio_mapping *map = VM(private);
	struct kgi_graphic *graph = (void *) kgidev[map->devid];
	IS_CRITICAL;

	TRACE(printk("%s:%i: nopage @%.8lx, vma %p, %x\n",__FILE__,__LINE__,
		addr, vma, no_share));

	ASSERT((VM(start) <= addr) && (addr < VM(end)));

	if (! (graph && map->file)) {

		TRACE(printk("%s:%i: nopage error\n",__FILE__,__LINE__));
		return 1;
	}

	if (DEV.state & KGI_MAPPED_ON_DISPLAY) {

		register ulong pstart, mstart, mend, offset;

		current->min_flt++;

		switch(map->mmap) {

		case MAP_LINEAR_LINEAR:
			TRACE(printk("linear mapped linear mmio\n"));
			map->offset = offset = 0;
			mstart = VM(start);
			mend   = VM(end);
			pstart = 0;
			break;

		case MAP_LINEAR_PAGED:
			TRACE(printk("linear mapped paged mmio\n"));
			offset  = addr - VM(start);
			offset -= offset % map->r->psize;
			map->offset = offset;
			mstart  = VM(start) + offset;
			mend	= mstart + map->r->psize;
			pstart  = 0;
			ASSERT(mend <= VM(end));
			break;

		case MAP_PAGED_LINEAR:
			TRACE(printk("paged mapped linear mmio\n"));
			mstart = VM(start);
			mend   = VM(end);
			offset = 0;
			pstart = map->offset;
			break;

		case MAP_PAGED_PAGED:
			TRACE(printk("paged mapped paged mmio\n"));
			mstart = VM(start);
			mend   = VM(end);
			offset = map->offset;
			pstart = 0;
			break;
		default:
			INTERNALERROR;
			return 1;
		}

		TRACE(printk(KERN_ERR "%s:%i: nopage @%.8lx, remap %.8lx-%.8lx"
			" to '%s' %.8lx, offset %lx\n", __FILE__,__LINE__, addr,
			mstart, mend, map->r->name, map->r->pstart + pstart,
			offset));

		ENTER_CRITICAL;

			graph_unmap_mappings(*(map->other));
			TRACE(printk(KERN_DEBUG"%s:%i: Setting offset = %lx\n",\
				__FILE__,__LINE__,offset));

			if (map->r->set_offset) {

				map->r->set_offset(map->r, offset);

			} else {

				map->r->offset=offset;
			}

			TRACE(printk(KERN_DEBUG \
				"%s:%i: page range remap (%p) %lx-%lx, @%lx\n",\
				__FILE__,__LINE__, map->mm, mstart, mend,\
				map->r->pstart+pstart));
			fast_remap_page_range(map->mm, (map->mstart = mstart),
				(map->mend = mend), map->r->pstart + pstart,
				TO_PGPROT(map->r->prot));

		LEAVE_CRITICAL;

		TRACE(printk(KERN_DEBUG"%s:%i: nopage return OK!\n",\
			__FILE__,__LINE__));
		return 0;

	} else {

		/* !!!	do the neccessary background actions here.
		*/
		TRACE(printk("%s:%i: nopage in background!\n",\
			__FILE__,__LINE__));
		return 1;
	}
}

/*	graph_mmio_unmap() has to unmap a mmio mapping. As we do not allow
**	partial mapping of MMIO regions, we consequently have to deny partial
**	unmapping. This is ensured in mm/mmap.c. We only have to remove our
**	local mapping state info because the pagetables are fixed in mm/mmap.c.
*/
static void graph_mmio_unmap(struct vm_area_struct *vma, unsigned long start,
	size_t len)
{
	struct kgi_mmio_mapping *map = VM(private);
	struct kgi_graphic *graph = (void *) kgidev[map->devid];

	TRACE(printk("%s:%i: mmio_unmap() vma %p, %.8lx, %i\n",\
		__FILE__,__LINE__, vma, start, len));

	if (graph) {

		/*	Device is still opened by at least one process. We
		**	only remove the specified mapping.
		*/
		struct kgi_mmio_mapping *prev = *(map->other);
		ASSERT(prev);

		if (prev == map) {

			*(map->other) = map->next;

		} else {

			while (prev->next != map) {

				prev = prev->next;
			}
			prev->next = map->next;
		}
	}

	ASSERT(mmapcnt[map->devid - MAX_NR_CONSOLES]);
	mmapcnt[map->devid - MAX_NR_CONSOLES]--;
	TRACE(memset(map, 0, sizeof(*map)));
	TRACE(VM(private) = NULL);
	kgi_free_small(map);
}

static struct vm_operations_struct graph_mmio_vmops = {

	NULL,                   /* open		*/
	NULL,			/* close	*/
	graph_mmio_unmap,	/* unmap	*/
	NULL,                   /* protect	*/
	NULL,                   /* sync		*/
	NULL,                   /* advise	*/
	graph_mmio_nopage,	/* nopage	*/
	NULL,                   /* wppage	*/
	NULL,                   /* swapout	*/
	NULL,                   /* swapin	*/
};

/*	Ping Pong Buffering
*/

static unsigned long graph_accel_nopage(struct vm_area_struct *vma, 
	unsigned long addr, int no_share)
{
	unsigned long offset = addr - VM(start);
	register struct kgi_accel_mapping *map = VM(private);
	struct kgi_graphic *graph = (void *) kgidev[map->devid];
	ulong size = map->ping.size;
	ulong domapstart, unmapstart;
	struct kgi_accel_buffer *domap, *unmap;


	ASSERT(map);
	ASSERT(map->ping.size == map->pong.size);

	TRACE(printk("%s:%i: accel '%s' nopage @%.8lx\n",__FILE__,__LINE__,
		     map->accel->name,addr));

	if (! (map->file && (DEV.state & KGI_MAPPED_ON_DISPLAY))) {

		/* !!!	well, this should be done in a better way.
		*/
		return 1;
	}

	/*	figure out which buffer to unmap and which to map. Wait until
	**	it is processed before continuing.
	*/
	if (offset < size) {

		if (map->ping.state != ACCEL_IDLE) {

			map->accel->wait(map->accel);
		}
		unmap = &(map->pong);
		domap = &(map->ping);
		unmapstart = (domapstart = VM(start)) + size;

	} else {

		if (map->pong.state != ACCEL_IDLE) {

			map->accel->wait(map->accel);
		}
		unmap = &(map->ping);
		domap = &(map->pong);
		domapstart = (unmapstart = VM(start)) + size;
	}
	domap->state = ACCEL_IDLE;

	TRACE(printk("%s:%i: domap %p state %i, unmap %p state %i\n",
	        __FILE__,__LINE__, domap->buffer, domap->state, unmap->buffer,
		unmap->state));
	
	if (unmap->state == ACCEL_FILL) {

		TRACE(printk("%s:%i: unmap %.8lx size %lx from %.8lx\n",
			__FILE__,__LINE__, unmapstart, size,
			virt_to_phys(unmap->buffer)));
		fast_unmap_page_range(VM(mm), unmapstart, unmapstart + size);
		unmap->state = ACCEL_WAIT;
		(map->accel->execute)(map->accel, unmap);
	}

	TRACE(printk("%s:%i: domap %.8lx size %lx to %.8lx\n",
		__FILE__,__LINE__, domapstart, size,
		virt_to_phys(domap->buffer)));

	domap->state = ACCEL_FILL;
	fast_remap_page_range(VM(mm), domapstart, domapstart + size,
		virt_to_phys(domap->buffer), VM(page_prot));
		
	return 0;
}

static void graph_accel_unmap(struct vm_area_struct *vma, unsigned long start,
	size_t len)
{
	struct kgi_accel_mapping *map = VM(private);
	struct kgi_accel_mapping *prev = *(map->other);

	TRACE(printk("%s:%i: unmap accel '%s'\n", __FILE__,__LINE__,
		     map->accel->name));

	/*	make sure the accel has no references hanging around.
	*/
	if (map->ping.state == ACCEL_WAIT || map->ping.state == ACCEL_EXEC) {

		map->accel->wait(map->accel);
	}

	if (map->pong.state == ACCEL_WAIT || map->pong.state == ACCEL_EXEC) {

		map->accel->wait(map->accel);
	}

	/*	The device is still opened by at least one process. We only
	**	remove the specified mapping here.
	*/
	if (map->file) {

		ASSERT(prev);
		if (prev == map) {

			*(map->other) = map->next;

		} else {

			while (prev->next != map) {

				prev = prev->next;
			}
			prev->next = map->next;
		}
	}

	ASSERT(mmapcnt[map->devid - MAX_NR_CONSOLES]);
	mmapcnt[map->devid - MAX_NR_CONSOLES]--;
	kfree(map->pong.alloc);
	kfree(map->ping.alloc);
	TRACE(memset(map, 0, sizeof(*map)));
	TRACE(VM(private) = NULL);
	kgi_free_small(map);
}

static struct vm_operations_struct graph_accel_vmops = {

	NULL,                   /* open		*/
	NULL,			/* close	*/
	graph_accel_unmap,	/* unmap	*/
	NULL,                   /* protect	*/
	NULL,                   /* sync		*/
	NULL,                   /* advise	*/
	graph_accel_nopage,	/* nopage	*/
	NULL,                   /* wppage	*/
	NULL,                   /* swapout	*/
	NULL,                   /* swapin	*/
};

/*	GC mapping. Grrr ... why do I have to put such trivial thing here ...
 *	One could handle this in the mmap once upon a time ...
 */

static unsigned long graph_GC_nopage(struct vm_area_struct *vma, 
	unsigned long addr, int no_share)
{
	TRACE(printk("nopage %p !\n",VM(private)));
	return (unsigned long) VM(private);	/* This holds a pointer to the GC ... */
}

static struct vm_operations_struct graph_GC_vmops = {

	NULL,                   /* open		*/
	NULL,			/* close	*/
	NULL,			/* unmap	*/
	NULL,                   /* protect	*/
	NULL,                   /* sync		*/
	NULL,                   /* advise	*/
	graph_GC_nopage,	/* nopage	*/
	NULL,                   /* wppage	*/
	NULL,                   /* swapout	*/
	NULL,                   /* swapin	*/
};

/*	graph_mmap() does the neccessary stuff to add a mapping to the current
**	mmio area mappings. We assume:
**	* one mm_struct per process
**	* vm_area_structs served by VM_FASTHANDLER are not merged together
**	* file pointer can be used to identify mappings on release.
*/
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,56)
static int graph_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct inode *inode = file->f_dentry->d_inode;
#else
static int graph_mmap(struct inode *inode, struct file *file,
	struct vm_area_struct *vma)
{
#endif
	struct kgi_graphic *graph = file->private_data;
	size_t size = VM(end) - VM(start);
	IS_CRITICAL;

	TRACE(printk(KERN_ERR "%s:%i: mmap(): dev %i, %.8lx-%.8lx, "
		"offset %lx, size %lx\n",__FILE__,__LINE__,\
		DEV.id, VM(start), VM(end), VM(offset), VM(end)-VM(start)));

	if ((VM(offset) & MMAP_TYPE) == MMAP_TYPE_MMIO) {

		struct kgi_mmio_region *r;
		struct kgi_mmio_mapping *map;

		if (VM(offset) & MMAP_PER_REGION_INDEX) {

			TRACE(printk(KERN_ERR"%s:%i: mmio per-index (0x%.4lx)"\
				"mapping\n",__FILE__,__LINE__,\
				VM(offset) & MMAP_INDEX));

			/*	mmio mapping per region.	*/
			VM(offset) &= MMAP_INDEX;
			r = (VM(offset) < MAX_NR_MMIO_REGIONS)
				? MMIO[VM(offset)] : NULL;
		} else {

			int i;
			TRACE(printk(KERN_ERR"%s:%i: mmio per-type (0x%.4lx)"\
				"maping\n",__FILE__,__LINE__,\
				VM(offset) & MMAP_INDEX));

			/*	mmio mapping per type.	*/
			VM(offset) &= MMAP_INDEX;

			for (i = 0; i < MAX_NR_MMIO_REGIONS; i++) {

				if (MMIO[i] && (MMIO[i]->type & MMAP_INDEX) ==
					VM(offset)) {

					break;
				}
			}
			r = (i < MAX_NR_MMIO_REGIONS) ? MMIO[i] : NULL;
			VM(offset) = i;
		}

		if (r==NULL) {

			TRACE(printk(KERN_ERR"%s:%i: No region for %lx\n",\
				__FILE__,__LINE__,VM(offset)));
			return -ENXIO;

		} else {

			TRACE(printk(KERN_ERR"%s:%i: mapping with \"%s\"\n",\
				__FILE__,__LINE__,r->name));
		}

		ASSERT(r ? ((r->size & ~PAGE_MASK) == 0) : 1);

		if ((VM(flags)&0xF) & ~(r->prot)) {

			TRACE(printk(KERN_ERR"%s:%i: mmio map perm(%.4x) "\
				"!= user perm(%.4x)\n",__FILE__,__LINE__,\
				r->prot,VM(flags)));
			return -EPERM;
		}

		TRACE(printk(KERN_DEBUG"%s:%i: map mmio '%s' @%.8lx size %lx\n",
			__FILE__,__LINE__, r->name, r->pstart, r->size));

		if ((r->size % PAGE_SIZE) || (size > r->size)) {

			TRACE(printk(KERN_ERR"%s:%i: map failed\n",
				__FILE__,__LINE__));
			return -EINVAL;
		}

		/*	We have to keep a copy of all mappings, thus we must
		**	ensure that there is no remapping (e.g. VC-switch)
		**	possible when updating the mapping lists.
		*/
		if (! (map = kgi_malloc_small(sizeof(*map), GFP_USER))) {

			TRACE(printk(KERN_ERR"%s:%i: mmap failed\n",\
				__FILE__,__LINE__));
			return -ENOMEM;
		}

		map->devid  = DEV.id;
		map->r      = r;
		map->mm     = VM(mm);
		map->mstart = 0;
		map->mend   = 0;
		map->offset = 0;
		map->file   = file;

		/*	now figure out how to map it. We have four possible
		**	cases: a linear or paged region; mapped linear or paged.
		*/
		if (r->size == r->psize) {

			/*	linear region	*/
			map->mmap = (size == r->size)
				? MAP_LINEAR_LINEAR : MAP_PAGED_LINEAR;

#if 0
			if (r->size % size) {

				TRACE(printk(KERN_ERR"%s:%i: mmap failed!\n",
					__FILE__,__LINE__));
				kgi_free_small(map);
				return -EINVAL;
			}
#endif
		} else {

			/*	paged region	*/
			map->mmap = (size == r->size)
				? MAP_LINEAR_PAGED : MAP_PAGED_PAGED;

			if ((size < r->size) && (size != r->psize)) {

				TRACE(printk(KERN_ERR"%s:%i: mmap failed!\n",
					__FILE__,__LINE__));
				kgi_free_small(map);
				return -EINVAL;
			}
		}

		if (zeromap_page_range(VM(start), VM(end) - VM(start),
			TO_PGPROT(r->prot))) {

			TRACE(printk(KERN_ERR"%s:%i: mmap failed\n",\
				__FILE__,__LINE__));
			kgi_free_small(map);
			return -ENOMEM;
		}

		ENTER_CRITICAL;

			fast_unmap_page_range(VM(mm), VM(start), VM(end));
			map->other = &(MMIO_MAP[VM(offset)]);
			map->next = MMIO_MAP[VM(offset)];
			MMIO_MAP[VM(offset)] = map;

		#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,53)
			vma->vm_dentry = dget(file->f_dentry);
		#else
			VM(inode)  = inode;		/* attach to inode  */
		#endif
			VM(private)= map;		/* private reference*/
			VM(flags) |= VM_FASTHANDLER;	/* use the GGI-hack */
			VM(ops)    = &graph_mmio_vmops;
			inode->i_count++;

		LEAVE_CRITICAL;

		TRACE(printk(KERN_DEBUG"%s:%i: mapped '%s'\n",\
			__FILE__,__LINE__,r->name));

		mmapcnt[DEV.id - MAX_NR_CONSOLES]++;
		return EOK;
	}

	if ((VM(offset) & MMAP_TYPE) == MMAP_TYPE_ACCEL) {

		struct kgi_accel *accel;
		struct kgi_accel_mapping *map;

		if (VM(offset) & MMAP_PER_REGION_INDEX) {

			VM(offset) &= MMAP_INDEX;
			accel = (VM(offset) < MAX_NR_ACCEL_REGIONS)
				? ACCEL[VM(offset)] : NULL;
		} else {

			int i;
			VM(offset) &= MMAP_INDEX;

			for (i = 0; i < MAX_NR_ACCEL_REGIONS; i++) {

				if (ACCEL[i] && ACCEL[i]->type == VM(offset)) {

					break;
				}
			}
			accel = (i < MAX_NR_ACCEL_REGIONS)
				? ACCEL[i] : NULL;
		}

			
		if (accel==NULL) {

			TRACE(printk(KERN_ERR"%s:%i: No region for %lx\n",\
				__FILE__,__LINE__,VM(offset)));
			return -ENXIO;

		} else {

			TRACE(printk(KERN_ERR"%s:%i: mapping with \"%s\"\n",\
				__FILE__,__LINE__,accel->name));
		}

		if ((VM(flags)&0xF) & ~(VM_READ|VM_WRITE|VM_SHARED)) {

			TRACE(printk(KERN_ERR \
				"%s:%i: accel perm(%.4x) != user perm(%.4x)\n",\
				__FILE__,__LINE__, VM_READ|VM_WRITE|VM_SHARED,\
				VM(flags)));
			return -EPERM;
		}

		if (!accel || (size != 2*PAGE_SIZE)) {

			TRACE(printk("%s:%i: mmap failed\n",__FILE__,__LINE__));
			return -EINVAL;
		}

		if (! (map = kgi_malloc_small(sizeof(*map), GFP_USER))) {

			TRACE(printk("%s:%i: mmap failed\n",__FILE__,__LINE__));
			return -ENOMEM;
		}

		/*	As i386 linux lacks a method to allocate a continuous
		**	region of physical RAM (for DMA) as well as the
		**	capability to get the physical address of a vmalloc()ed
		**	buffer; we have to kmalloc() two pages here and
		**	use only one of them. Don't blame us for this, tell
		**	Linus.
		*/
		memset(map, 0, sizeof(*map));
		if (!  ((map->ping.alloc = kmalloc(2*PAGE_SIZE,
				GFP_DMA | GFP_USER)) &&
			(map->pong.alloc = kmalloc(2*PAGE_SIZE,
				GFP_DMA | GFP_USER))) ||
			zeromap_page_range(VM(start), VM(end) - VM(start), 
				VM(page_prot))) {

			TRACE(printk("%s:%i: mmap failed\n",__FILE__,__LINE__));
			kgi_free_huge(map->ping.buffer);
			kgi_free_huge(map->pong.buffer);
			kgi_free_small(map);
			return -ENOMEM;
		}
		memset(map->ping.alloc, 0, 2*PAGE_SIZE);
		memset(map->pong.alloc, 0, 2*PAGE_SIZE);
		map->ping.buffer = (void *)(((ulong)map->ping.alloc+ ~PAGE_MASK)
						& PAGE_MASK);
		map->pong.buffer = (void *)(((ulong)map->pong.alloc+ ~PAGE_MASK)
						& PAGE_MASK);
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,31) 
		mem_map_reserve(MAP_NR(__va(virt_to_phys(map->ping.buffer))));
		mem_map_reserve(MAP_NR(__va(virt_to_phys(map->pong.buffer))));
#endif
		TRACE(*((ggi_uint *) map->ping.buffer) = 0xDEADBEEF);
		TRACE(*((ggi_uint *) map->pong.buffer) = 0xBEEFDEAD);

		map->ping.size = map->pong.size = PAGE_SIZE;
		map->file = file;
		map->devid = DEV.id;
		map->accel = accel;

		ENTER_CRITICAL;

			fast_unmap_page_range(VM(mm), VM(start), VM(end));
			map->other = &(ACCEL_MAP[VM(offset)]);
			map->next = ACCEL_MAP[VM(offset)];
			ACCEL_MAP[VM(offset)] = map;	

		#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,53)
			vma->vm_dentry = dget(file->f_dentry);
		#else
			VM(inode)  = inode;		/* attach to inode  */
		#endif
			VM(private)= map;		/* private reference*/
			VM(flags) |= VM_FASTHANDLER;
			VM(ops) = &graph_accel_vmops; 
			inode->i_count++;
			
		LEAVE_CRITICAL;

		TRACE(printk("%s:%i: accel '%s' mapped\n",__FILE__,__LINE__,
			accel->name));
		mmapcnt[DEV.id - MAX_NR_CONSOLES]++;
		
		return EOK;
	}

	if ((VM(offset) & MMAP_TYPE) == MMAP_TYPE_GC) {

		if (size != PAGE_SIZE) {

			TRACE(printk("%s:%i: mmap failed\n",__FILE__,__LINE__));
			return -EINVAL;
		}

		ENTER_CRITICAL;

		TRACE(printk("Starting map of %x-%x to %p\n",VM(start), VM(end),GC));
		TRACE(*(int *)GC=0xDeadbeef);

		vma->vm_ops = &graph_GC_vmops;
		VM(private) = GC;

		#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,53)
			vma->vm_dentry = dget(file->f_dentry);
		#else
			VM(inode)  = inode;		/* attach to inode  */
			inode->i_count++;
		#endif
		LEAVE_CRITICAL;

		TRACE(printk("%s:%i: GC mapped\n",__FILE__,__LINE__));
		
		return EOK;
	}

	return -ENXIO;
}

/*
**	IOCTL services
*/

static int graph_driver_command(struct kgi_graphic *graph, ggi_uint cmd, ggi_uint arg)
{
	int err,i;
	struct kgi_mmio_region *mmio;
	struct kgi_accel *accel;

	switch(cmd) {

	case DRIVER_MMIO_OFFSET:
		/* !!! implement! */
		return -EINVAL;

	case DRIVER_TESTDRIVER:		/* check if driver is present	*/
		return DEVICEFILE_MAGIC;

	case DRIVER_TESTMODE:		/* check if mode is possible	*/
		IO_RES = IO_BUF;
		return kgi_check_mode(DEV.dpy, 
			(struct kgi_mode *) graph->io_buf, CMD_PROPOSE);

	case DRIVER_SETGRAPHMODE:	/* set mode for device	*/
		#define	newmode	((struct kgi_mode*) IO_BUF)

		if (mmapcnt[DEV.id - MAX_NR_CONSOLES]) {

			return -EBUSY;
		}

		if ((err = kgi_check_mode(DEV.dpy, newmode, CMD_PROPOSE))) {

			return err;
		}
		memcpy(&(DEV.mode), newmode, sizeof(DEV.mode));
		kgi_set_split(&DEV, DEV.mode.virt.y);
		IO_RES = &(DEV.mode);

		mmio = newmode->mmio;
		for (i = 0; i < MAX_NR_MMIO_REGIONS; i++) {

			MMIO[i] = mmio;
			mmio = mmio ? mmio->next : NULL;
			ASSERT(! MMIO_MAP[i]);

			#ifdef DEBUG

				if (MMIO[i]) {

					TRACE(printk(KERN_ERR \
					"%s:%i: mmio%i: '%s', @%.8lx,"\
					" psize %lx, total %lx\n",
						__FILE__,__LINE__, i,
						MMIO[i]->name, MMIO[i]->pstart,
						MMIO[i]->psize, MMIO[i]->size));
				}
			#endif
		}

		accel = newmode->accel;
		for (i = 0; i < MAX_NR_ACCEL_REGIONS; i++) {

			ACCEL[i] = accel;
			accel = accel ? accel->next : NULL;

			#ifdef DEBUG

				if (ACCEL[i]) {

					TRACE(printk(KERN_DEBUG \
						"%s:%i: accel%i : '%s'\n",
						__FILE__,__LINE__,
						i, ACCEL[i]->name));
				}
			#endif
		}

		for (i = 0; i < newmode->request.frames; i++) {

			ggi_clut *clut = kgi_default_clut(&(newmode->request));

			ASSERT(clut);

			DEV.org[i].x = DEV.org[i].y = 0;
			memcpy(&(DEV.lut[i]), clut, sizeof(ggi_clut));
		}
		kgi_set_pointer_window(&DEV,  0, DEV.mode.visible.x, 
			0, DEV.mode.visible.y);

		if (!(DEV.state & KGI_MAPPED_ON_DISPLAY)) {

			kgi_focus_device(DEV.id);

		} else {

			kgi_set_mode(DEV.dpy, IO_BUF);
		}
		return EOK;

	case DRIVER_GETGRAPHMODE:	/* return graphic mode		*/
		IO_CNT = sizeof(ggi_mode);
		IO_RES =  &(DEV.mode);
		return EOK;

	case DRIVER_GRAPHICS:		/* go to graphic mode		*/
		kgi_focus_device(DEV.id);
		return EOK;

	case DRIVER_TEXT:		/* go to text mode (console)	*/
		kgi_focus_device(DEV.id - MAX_NR_CONSOLES);
		return EOK;

#if 0
	case DRIVER_GETINFO:
	case DRIVER_SETINFO:
	case DRIVER_STRATEGY:

	case DRIVER_RAMDAEMON_INSTALL:
	case DRIVER_RAMDAEMON_ASK:
	case DRIVER_RAMDAEMON_CPY:
#endif
	default:
		TRACE(printk(KERN_INFO "%s:%i: Unknown command %x\n",\
			__FILE__,__LINE__, cmd));
		return -E(DRIVER, UNKNOWN);
	}
}

/*	ioctl() services. This is the primary command interface. Other 
**	driver-specific interfaces may be provided using mmio-regions or 
**	by giving direct access to registers where this is safe. 
**	Other OSes might even have other means.
*/
static int graph_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	struct kgi_graphic *graph = file->private_data;
	int foo;
	int result = 0;

	/*	Lock will be non-zero if there is another ioctl() call in
	**	progress but scheduled. Not sure if this could happen really,
	**	just to make sure.
	** NOTE	When hacking this procedure, make sure you clear the lock!
	*/
	ASSERT(! IO_LOCK);
	if (IO_LOCK) {

		printk(KERN_ERR "%s:%i: ioctl() while other in progress. "
			"please report to " MAINTAINER ".\n",
			__FILE__,__LINE__);
		return -EAGAIN; 
	}
	IO_LOCK = 1;

	IO_CNT = _IOC_SIZE(cmd);
	IO_RES = NULL;

	/*	Get the data from userspace.
	*/
	if (cmd & IOC_IN) {

		if ((foo = verify_area(VERIFY_READ, (void *) arg, IO_CNT))) {

		          result = foo;
		          goto unlock;
	        }

		copy_from_user(IO_BUF, (void *) arg, IO_CNT);
	}

	/*	Dispatch the requested command.
	*/
	foo = _IOC_TYPE(cmd);
	if (foo == IOC_DRIVER) {

		TRACE(printk(KERN_ERR "%s:%i: driver command %.8x\n",\
			__FILE__,__LINE__,cmd));
		result = graph_driver_command(graph, cmd, arg);

	} else {

		TRACE(printk(KERN_ERR "%s:%i: dispatched command %.8x\n",\
			__FILE__,__LINE__,cmd));
		result = ((foo < KGI_NR_SUBSYSTEMS) && (DEV.dpy->cmd[foo])) 
				? DEV.dpy->cmd[foo](graph, cmd, arg)
				: -E(DRIVER, UNKNOWN);
	}

	/*	write return values into userspace, but only if the call was
	**	successful.
	*/	
	if ((cmd & IOC_OUT) && (result == EOK)) {

		if ((foo = verify_area(VERIFY_WRITE, (void *) arg, IO_CNT))) {

			result = foo;
			goto unlock;
		}

		ASSERT(IO_RES);
	
		if (IO_RES && (0 < IO_CNT) && (IO_CNT < (IOCSIZE_MASK+1))) {
		
			copy_to_user((void *) arg, IO_RES, IO_CNT);

		} else {

			result = -E(DRIVER, UNKNOWN);
		}
	}

unlock:
	IO_LOCK = 0;
	TRACE(printk("%s:%i: ioctl finished.\n",__FILE__,__LINE__));
	return result;
}

/*
**	generic stuff
*/

static void graph_handle_event(struct kgi_device *dev, ggi_event *event)
{
	#define	graph	((struct kgi_graphic *) dev)
	struct ggi_event_buffer *buf = EVBUFS;

	DEBUG2("graph_handle_event for dev %i", dev->id);

	ASSERT(buf);
	ASSERT(event);
	ASSERT(dev);

	/*	!!! change this. Implement proper input event filtering! */
	while (buf) {

		ggiEventPut(buf, event);
		buf = buf->next;
	}
	wake_up(&(graph->waiting_for_event));

	#undef graph
}

/*	graph_reset() is called after a device is allocated and attached,
**	so it is not mapped yet. We just initialize the kgi_graphic local state.
*/
static int graph_reset(struct kgi_graphic *graph)
{
	struct kgi_mode	mode;
	ggi_uint i;
	struct kgi_mmio_region *mmio;

	DEV.save = graph_save;
	DEV.restore = graph_restore;

	if (!  (GC = get_free_page(GFP_ATOMIC)) ) {
		TRACE(printk(KERN_ERR"failed, could not allocate GC.\n"));
		return -ENOMEM;
	}
	TRACE(printk("allocated GC %p\n",GC));
	/* FIXME ! SECURITY ! Do we need to zap contents of GC here ? */

	if (IO_BUF == NULL) {

		IO_BUF = vmalloc(IO_BUF_SIZE);
		if (IO_BUF == NULL) 
			return -ENOMEM;

		IO_CNT = 0;
		IO_LOCK = 0;
	}

	TRACE(printk(KERN_DEBUG"Checking mode...\n"));
	mode.request = DEV.mode;
	if (kgi_check_mode(DEV.dpy, &mode, CMD_PROPOSE)) {

		INTERNALERROR;
	}

	mmio = mode.mmio;
	for (i = 0; i < MAX_NR_MMIO_REGIONS; i++) {
		MMIO[i] = mmio ? mmio : NULL;
		mmio    = mmio ? mmio->next : NULL;
	}
	if (mmio) {

		printk(KERN_ERR "display %i reported >= %i MMIO regions\n",
			DEV.dpy->id, MAX_NR_MMIO_REGIONS);
	}

	DEV.event_mask = emAll;
	DEV.handle_event = graph_handle_event;

	return EOK;
}

/*	graph_done() has to free memory allocated in graph_reset() and 
**	elsewhere. It has also to release and unmap the mappings for that
**	particular file.
*/
static void graph_done(struct kgi_graphic *graph)
{
	TRACE(printk(KERN_DEBUG"%s:%i: graph_done()\n",__FILE__,__LINE__));
	ASSERT(IO_BUF);	vfree(IO_BUF);
	TRACE(printk("*GC %x\n",*(int *)GC));
	if (GC) free_page(GC);
}

/*
**	generic operations
*/

/*	Open a graphic special file. We force all processes that made
**	references mmap() to invalidate these before we allow any reallocation.
*/
static int graph_open(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev)
			? MINOR(inode->i_rdev) 
			: ( (current && current->tty && current->tty->device)
			    ? MINOR(current->tty->device)
			    : 1 );
	struct kgi_graphic *graph;
	ggi_event_buffer *buf;

	TRACE(printk(KERN_DEBUG "%s:%i: graph_open (minor = %d)\n",\
		__FILE__,__LINE__,minor));

	minor--;

	if (minor < MAX_NR_CONSOLES) {

		graph = (void *) kgidev[minor + MAX_NR_CONSOLES];
		ASSERT(graph ? (DEV.type == KGI_GRAPHIC_DEV) : 1);

	} else {

		TRACE(printk(KERN_DEBUG "%s:%i: invalid minor %i!\n",\
			__FILE__,__LINE__, minor));
		return -ENXIO;
	}

	buf = kgi_malloc_small(sizeof(ggi_event_buffer), GFP_KERNEL);
	if (buf == NULL)
		return -ENOMEM;

	buf->buf = kgi_malloc_huge(PAGE_SIZE);
	if (buf->buf == NULL) {
		kgi_free_small(buf);
		return -ENOMEM;
	}
	ASSERT(buf);
	ASSERT(buf->buf);
	buf->bigbang = jiffies;
	buf->size = PAGE_SIZE;
	buf->read = buf->write = 0;
	buf->overflow = 0;
	buf->priv = file;

#if 0
	TRACE({
		ggi_event_buffer *test = EVBUFS;
		while (test) {

			ASSERT(test->priv != file);
			test = test->next;
		}
	});
#endif

	if (graph) {

		printk(KERN_ERR "%s:%i: multiple open not yet allowed!\n",\
			__FILE__,__LINE__);
		kgi_free_huge(buf->buf);
		kgi_free_small(buf);
		return -EBUSY;	/* !!! temporary! */
		/* !!! For now allow only one process per device */

	} else {

		TRACE(printk(KERN_ERR "%s:%i: allocating graph%i ... ",\
			__FILE__,__LINE__, minor));

		/*	Force to free all references before reopen is allowed.
		*/
		if (mmapcnt[minor]) {

			TRACE(printk(KERN_ERR"busy, access denied.\n"));
			kgi_free_huge(buf->buf);
			kgi_free_small(buf);
			return -EBUSY;
		}
		TRACE(printk(KERN_DEBUG"mmapcnt[%d]=%d\n",\
			minor, mmapcnt[minor]));

		if (! (graph = kgi_alloc_device(minor + MAX_NR_CONSOLES,
			GFP_USER))) {

			TRACE(printk(KERN_ERR"failed, no memory.\n"));
			kgi_free_huge(buf->buf);
			kgi_free_small(buf);
			return -ENOMEM;
		}
		TRACE(printk(KERN_DEBUG"kgi_alloc_device(%d,GFP_USER)=%p\n",\
			minor + MAX_NR_CONSOLES, graph));

		if (kgi_attach_device(&DEV)) {

			TRACE(printk(KERN_ERR"failed, could not attach.\n"));
			kgi_free_huge(buf->buf);
			kgi_free_small(buf);
			return -ENXIO;
		}
		TRACE(printk(KERN_DEBUG"kgi_attach_device(%p)=EOK\n",&DEV));

		if (graph_reset(graph)) {

			TRACE(printk(KERN_ERR"failed, could not reset.\n"));
			kgi_free_huge(buf->buf);
			kgi_free_small(buf);
			graph_done(graph);
			kgi_detach_device(&DEV);
			kgi_free_device(&DEV);
			return -ENXIO;
		}
		
		TRACE(printk(KERN_DEBUG"open on display %i\n", DEV.dpy->id));
	}
	TTY = current->tty;
	DEV.refcnt++;
	file->private_data = graph;
	buf->next = EVBUFS;
	EVBUFS = buf;

	return EOK;
}

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,31)
#define RELEASE_OK 0
#define RELEASE_ERR -1
static int graph_release(struct inode *inode, struct file *file)
#else
#define RELEASE_OK
#define RELEASE_ERR
static void graph_release(struct inode *inode, struct file *file)
#endif
{
	struct kgi_graphic *graph = file->private_data;
	struct ggi_event_buffer *buf;
	ggi_uint i;

	if (!graph) {

		return RELEASE_ERR;
	}

	/*	Invalidate mappings for this file. The vm_area_structs remain
	**	valid until the process explicitely unmaps them; if they are
	**	referenced but have no file we send SIGBUS.
	*/
	for (i = 0; i < MAX_NR_MMIO_REGIONS; i++) {

		struct kgi_mmio_mapping *map = MMIO_MAP[i];

		while (map) {

			if (map->file == file) {

				map->file = NULL;
				fast_unmap_page_range(map->mm, map->mstart,
					map->mend);
				map->mstart = map->mend = 0;
			}
			map = map->next;
		}
	}
	/*	The accel mappings don't touch the hardware directly, so
	**	we don't unmap them but catch invalid references at the next
	**	ping<->pong fault.
	*/
	for (i = 0; i < MAX_NR_ACCEL_REGIONS; i++) {

		struct kgi_accel_mapping *map = ACCEL_MAP[i];

		while (map) {

			if (map->file == file) {

				map->file = NULL;
			}
			map = map->next;
		}
	}

	buf = EVBUFS;
	if (buf && (buf->priv == file)) {

		EVBUFS = buf->next;
		kgi_free_huge(buf->buf);
		kgi_free_small(buf);

	} else {

		while (buf) {

			if (buf->next->priv == file) {

				ggi_event_buffer *tmp = buf->next;

				buf->next = tmp->next;

				DEBUG2("free event buffer for file %p", file);
				kgi_free_huge(tmp->buf);
				kgi_free_small(tmp);
			}
			buf = buf->next;
		}
	}

	/*	If this is the last reference to this device, free associated
	**	status and go back to the console. Well, if there is one... :-)
	** !!!	Improve this. We have to go back to the console the calling
	** !!!	process isrunning on, not the one that is calculated from DEV.id
	*/
	if (--(DEV.refcnt) == 1) {

 		TRACE(printk(KERN_DEBUG "%s:%i: freeing graph%i...\n", \
			__FILE__,__LINE__, DEV.id - MAX_NR_CONSOLES));

		ASSERT(! EVBUFS);

		if (DEV.state & KGI_MAPPED_ON_DISPLAY) {

			ASSERT(DEV.dpy->dev == &DEV);
			kgi_unmap_device(&DEV);

			if (kgidev[DEV.id - MAX_NR_CONSOLES]) {

				kgi_focus_device(DEV.id - MAX_NR_CONSOLES);
			}
		}
		graph_done(graph);
		kgi_detach_device(&DEV);
		kgi_free_device(&DEV);
	}

	return RELEASE_OK;
}
#undef RELEASE_OK
#undef RELEASE_ERR

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,26)
static int graph_select(struct inode *inode, struct file *file, int type,
	select_table *wait)
{
	struct kgi_graphic *graph = file->private_data;

	switch (type) {

	case SEL_IN:
		/* !!!	Implement input event filtering per file! */
		if (ggiEventAvailable(EVBUFS)) {

			return 1;
		}
		select_wait(&(graph->waiting_for_event), wait);
		return 0;

	case SEL_OUT:
	default:
		return 0;
	}
}
#else
static unsigned int graph_poll(struct file *file, poll_table *wait)
{
	struct kgi_graphic *graph = file->private_data;

	/* !!!	Implement input event filtering per file! */
	poll_wait(&(graph->waiting_for_event), wait);
	if (ggiEventAvailable(EVBUFS)) {
			return POLLIN | POLLRDNORM;
	}

	return 0;
}
#endif /* v2.1.26 */

#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,60)
static ssize_t graph_read(struct file * file, char *buf,
                          size_t cnt, loff_t *ppos)
{
        struct inode *inode = file->f_dentry->d_inode;
#elif LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
static long graph_read(struct inode *inode, struct file *file,
	char *buf, unsigned long cnt)
{
#else
static int graph_read(struct inode *inode, struct file *file,
	char *buf, int cnt)
{
#endif
	struct kgi_graphic *graph = file->private_data;
	ggi_event event;
	ggi_event_buffer *evbuf = EVBUFS;
	int written = 0;

	while ( evbuf && evbuf->priv != file ) {

		evbuf = evbuf->next;
	}
	ASSERT(evbuf);
	if (! evbuf) {

		return -ENXIO;
	}

	while (ggiEventAvailable(evbuf)) {

		ggiEventCopy(evbuf, &event);
		if (cnt < event.size) {

			return written;
		}
		ggiEventSkip(evbuf);
		copy_to_user(buf, &event, event.size);
		inode->i_atime = CURRENT_TIME;
		buf += event.size;
		written += event.size;
		cnt -= event.size;
	}
	return written;
}

static struct file_operations graph_fops =
{
	NULL,		/* lseek	*/
	graph_read,	/* read		*/
	NULL,		/* write	*/
	NULL,		/* readdir	*/
#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,26)
	graph_select,	/* select	*/
#else
	graph_poll,	/* poll		*/
#endif
	graph_ioctl,	/* ioctl	*/

	graph_mmap,	/* mmap		*/
	graph_open,	/* open		*/
	graph_release,	/* release	*/
	NULL,		/* fsync	*/
};

/*
**	Kernel Interface
*/
int dev_graphic_init(void)
{
	memset(mmapcnt, 0, sizeof(mmapcnt));

	if (register_chrdev(GRAPHDEV_MAJOR, GRAPHDEV_NAME, &graph_fops)) {

		printk(KERN_ERR "Can't register graphic devices.\n");
		return -EIO;
	}

	printk(KERN_NOTICE "graphic devices registered.\n");
	return EOK;
}
